#include "h.h"
#include "nsyslog.h"
#include "var.h"
#include "htcp.h"


size_t
buf_size(b)
	buflist_t *b;
{
	size_t sz = 0;

	while (b) {
		sz += b->bl_rem;
		b = b->bl_next;
	}
	return sz;
}


/*
 * concatenate number buffers of at least total size sz into one buffer.
 * free up buffers pulled into the new big buffer.
 */
buflist_t *
buf_pullup(buf, sz)
	buflist_t **buf;
	size_t sz;
{
	buflist_t *b, **bp, *bn;
	size_t avail;

	avail = buf_size(*buf);
	if (avail < sz)
		return NULL;
	b = buf_new(avail);
	if (b == NULL)
		return NULL;

	bp = buf;
	while ((bn = *bp) != NULL) {
		bcopy(bn->bl_ptr, b->bl_ptr, bn->bl_rem);
		b->bl_ptr += bn->bl_rem;
		*bp = bn->bl_next;
		buf_free(bn);
	}
	b->bl_rem = avail;
	b->bl_ptr = b->bl_buf;
	*buf = b;
	return b;
}


/*
 * add some buffers on to the end of a buffer list.  return a pointer to what
 * will be the new tail.
 */
buflist_t **
buf_add(tail, buf1, len1, buf2, len2)
	buflist_t **tail;
	u_char *buf1;
	size_t len1;
	u_char *buf2;
	size_t len2;
{
	buflist_t *bl, *b;

	bl = buf_new(len1 + len2);
	if (bl == NULL)
		return NULL;

	if (len1 > 0)
		bcopy(buf1, bl->bl_ptr, len1);
	if (len2 > 0)
		bcopy(buf2, bl->bl_ptr + len1, len2);
	bl->bl_rem = bl->bl_sz;
	bl->bl_buf[bl->bl_sz] = '\0';
	*tail = bl;
	return &bl->bl_next;
}


int
buf_get(bh, buf, bufsz)
	buflist_t **bh;
	char *buf;
	size_t bufsz;
{
	size_t i, cnt;
	buflist_t *b;
	char *s;

	b = *bh;
	if (bufsz > buf_size(b))
		return 0;		/* No. */

	for (cnt = 0, s = buf; (bufsz > 0); ) {
		i = MIN(b->bl_rem, bufsz);
		bcopy(b->bl_ptr, s, i);
		b->bl_rem -= i;
		cnt += i;
		s += i;
		bufsz -= i;
		if (b->bl_rem == 0) {
			*bh = b->bl_next;
			buf_free(b);
			b = *bh;
		} else
			b->bl_ptr += i;
	}

	return cnt;
}


/*
 * If there is not enough data for the complete message, just return, else
 * extract a single syslog message from the buffer.  extract prepended
 * information about the message into the tcplmsg pointer passed.
 */
int
buf_getmsg(bh, ptr, sslp, buf, bufsz)
	buflist_t **bh;
	void *ptr;
	void *sslp;
	char *buf;
	size_t bufsz;
{
	size_t sz, len, clen, mlen;
	htcpcmd_t *hc, hcm;
	buflist_t *b, **bp;
	tcplmsg_t *tlm;
	int cmd, op;
	ssl_t *ssl;
	char *s;

again:
	b = *bh;
	if (b == NULL)
		return 0;
	ssl = sslp;
	tlm = ptr;
	mlen = 0;
	len = 0;
	bp = bh;
	/*
	 * Only messages that have a tlm will have a hcm in from of them.
	 */
	if (tlm != NULL) {
		if (b->bl_rem < sizeof(hcm))
			return 0;
		/*
		 * check for unaligned access to htcpcmd_t components.
		 */
		if (((u_long)b->bl_ptr & 0x1) != 0) {
			bcopy(b->bl_ptr, (char *)&hcm, sizeof(hcm));
			hc = &hcm;
		} else
			hc = (htcpcmd_t *)b->bl_ptr;
		sz = sizeof(*hc);
		cmd = ntohs(hc->hth_cmd);
		op = ntohs(hc->hth_op);
		if (HTC_ISVARSZ(cmd))
			sz += op;
		cmd = HTC_CMD(cmd);

		/*
		 * Do we have enough for htcpcmd_t plus its args ?
		 */
		if (sz > buf_size(b))
			return 0;		/* No. */
		/*
		 * If we don't have enough contiguous data in the
		 * first buffer, attempt to pull it all up into
		 * one buffer.
		 */
		if (sz > b->bl_rem) {
			b = buf_pullup(bh, sz);
			if (b == NULL)
				return -1;
		}

		b->bl_rem -= sizeof(htcpcmd_t);
		b->bl_ptr += sizeof(htcpcmd_t);

#ifdef	HAVE_SSL_H
		if ((cmd == HTC_SSL) && (ssl != NULL)) {
			/*
			 * Need to call SSL_read to decrypt data.
			 */
			buflist_t *b2;
			int err;

			b2 = buf_new(op);
			err = SSL_read(ssl->ssl_ssl, b2->bl_ptr, op);
			*ssl->ssl_rtail = b2;
			ssl->ssl_rtail = &b2->bl_next;
			b2->bl_rem = err;
			goto again;
		}
#endif
		/*
		 * extract the tcplmsg struct first.  Don't assume it is all
		 * in the same buffer.
		 */
		s = (char *)tlm;
		clen = sizeof(*tlm);
		while ((b != NULL) && (clen > 0)) {
			sz = MIN(b->bl_rem, clen);
			bcopy(b->bl_ptr, s, sz);
			bufsz -= sz;
			b->bl_rem -= sz;
			clen -= sz;
			len += sz;
			if (b->bl_rem == 0) {
				*bp = b->bl_next;
				buf_free(b);
				b = *bp;
			} else
				b->bl_ptr += sz;
		}
		tlm->tl_len = ntohl(tlm->tl_len);
		tlm->tl_mlen = ntohl(tlm->tl_mlen);
		tlm->tl_hlen = ntohl(tlm->tl_hlen);
		tlm->tl_pri = ntohl(tlm->tl_pri);
		tlm->tl_tv = ntohl(tlm->tl_tv);
		mlen = tlm->tl_mlen;
	} else {
		/*
		 * Look for at least one line feed marking an end of message.
		 * Count how many bytes we go over to get there too.
		 */
		for (b = *bh; b != NULL; b = b->bl_next) {
			s = memchr(b->bl_ptr, '\n', b->bl_rem);
			if (s == NULL)
				mlen += b->bl_rem;
			else {
				s++;
				mlen += s - b->bl_ptr;
				break;
			}
		}
		if (s == NULL)	/* no LF */
			return 0;
	}

	/*
	 * If the buffer is bigger than this single message, then make it the
	 * same size as the message.
	 */
	clen = bufsz;
	if (clen > mlen)
		clen = mlen;

	/*
	 * Now extract the log message.  Again, don't assume it is all in the
	 * same buffer.
	 */
	s = buf;
	while ((b != NULL) && (clen > 0)) {
		sz = MIN(b->bl_rem, clen);
		bcopy(b->bl_ptr, s, sz);
		clen -= sz;
		b->bl_rem -= sz;
		len += sz;
		if (b->bl_rem == 0) {
			*bp = b->bl_next;
			buf_free(b);
			b = *bp;
		} else
			b->bl_ptr += sz;
	}

	/*
	 * If the message was bigger than the buffer, then eat up the rest of
	 * the message.
	 */
	if (mlen > bufsz) {
		mlen -= bufsz;
		while ((b != NULL) && (mlen > 0)) {
			sz = MIN(mlen, b->bl_rem);
			b->bl_rem -= sz;
			if (b->bl_rem == 0) {
				*bp = b->bl_next;
				buf_free(b);
				b = *bp;
			} else
				b->bl_ptr += sz;
		}
	}
	return len;
}


void
buf_freelist(b)
	buflist_t *b;
{
	buflist_t *b2;

	for (b2 = b; b2 != NULL; b2 = b) {
		b = b2->bl_next;
		buf_free(b2);
	}
}


void
buf_free(b)
	buflist_t *b;
{
	if (b->bl_base != b->bl_buf)
		free(b->bl_base);
	if (b->bl_src != NULL)
		free(b->bl_src);
	free(b);
}


buflist_t *
buf_new(len)
	size_t len;
{
	buflist_t *b;
	size_t alen = len;

	/*
	 * Round up size to help aligned allocation.
	 */
	if ((alen & 1) != 0)
		alen++;

	b = (buflist_t *)malloc(sizeof(*b) + alen);
	if (b == NULL)
		return NULL;
	b->bl_base = b->bl_buf;
	b->bl_ptr = b->bl_buf;
	b->bl_src = NULL;
	b->bl_slen = 0;
	b->bl_rem = 0;
	b->bl_sz = len;
	b->bl_next = NULL;
	b->bl_nh = NULL;
	return b;
}


/*
 * dummy read
 */
int
buf_discard(bh, len)
	buflist_t **bh;
	size_t len;
{
	size_t i, cnt;
	buflist_t *b;

	b = *bh;
	if (b == NULL)
		return 0;
	if (len >= buf_size(b)) {
		*bh = b->bl_next;
		buf_free(b);
		return len;
	}

	for (cnt = 0; (len > 0); ) {
		i = MIN(b->bl_rem, len);
		b->bl_rem -= i;
		cnt += i;
		len -= i;
		if (b->bl_rem == 0) {
			*bh = b->bl_next;
			buf_free(b);
			b = *bh;
		} else
			b->bl_ptr += i;
	}

	return cnt;
}
