#include "h.h"
#include "nsyslog.h"
#include "var.h"
#include "y.tab.h"

#if !defined(lint)
static const char rcsid[] = "$Id: cfg.c,v 1.31 1999/03/28 04:43:13 darrenr Exp $";
#endif

filetype_t	*fttable;
u_int	fttsize;
int	max_dfd;
int	max_sfd;
char	myhostname[MAXHOSTNAMELEN+1];
char	fullhostname[MAXHOSTNAMELEN+1];


char *
get_shorthostname()
{
	char *s;

	if (!*myhostname) {
		gethostname(myhostname, MAXHOSTNAMELEN);
		myhostname[MAXHOSTNAMELEN] = '\0';
		s = strchr(myhostname, '.');
		if (s != NULL)
			*s = '\0';
	}

	if (!*fullhostname) {
		gethostname(fullhostname, MAXHOSTNAMELEN);
		fullhostname[MAXHOSTNAMELEN] = '\0';
	}
	return myhostname;
}


char *
get_fullhostname()
{
	char *s;

	if (!*myhostname) {
		gethostname(myhostname, MAXHOSTNAMELEN);
		myhostname[MAXHOSTNAMELEN] = '\0';
		s = strchr(myhostname, '.');
		if (s != NULL)
			*s = '\0';
	}

	if (!*fullhostname) {
		gethostname(fullhostname, MAXHOSTNAMELEN);
		fullhostname[MAXHOSTNAMELEN] = '\0';
	}
	return fullhostname;
}


/*
 * initialize table size to match the maximum number of open files.
 */
void
init_config()
{
	fttsize = getdtablesize();
	fttable = (filetype_t *)malloc(fttsize * sizeof(*fttable));
	max_dfd = -1;
	max_sfd = -1;
}


void
cleanup_config()
{
	free(fttable);
}


int
setup_filetable(rfdp)
	fd_set *rfdp;
{
	int cnt = 0, fd, num = 0;
	void *ptr = NULL;
	sdlist_t *s;

	/*
	 * populate fttable with open message sources.  store a pointer to
	 * the source variable, the type of file and a pointer to its parent
	 * data structure.
	 */
	while ((fd = next_sfd(&ptr, &cnt)) != -2) {
		if (fd >= 0) {
			if (fttable[fd].ft_var != NULL)
				continue;
			s = (sdlist_t *)ptr;
			if (opts & OPT_VERBOSE)
				fprintf(stderr, "fd %d: type=%x %p source\n",
					fd, fttable[fd].ft_type);
			fttable[fd].ft_var = s->sd_src;
			num++;
			if (fd > max_sfd)
				max_sfd = fd;
		}
	}

	ptr = NULL;
	cnt = 0;

	/*
	 * mimic what we did above for sources but this time for destinations
	 */
	while ((fd = next_dfd(&ptr, &cnt)) != -2) {
		if (fd >= 0) {
			if (fttable[fd].ft_var != NULL)
				continue;
			s = (sdlist_t *)ptr;
			num++;
			fttable[fd].ft_var = s->sd_dst;
			if (fttable[fd].ft_type == DEST_TCP) {
				if (opts & OPT_VERBOSE)
					fprintf(stderr, "fd %d: type=%x dest\n",
						fd, fttable[fd].ft_type);
				FD_SET(fd, rfdp);
			}
			if (fd > max_dfd)
				max_dfd = fd;
		}
	}
	return MAX(max_dfd, max_sfd);
}


/*
 * read in the configuration file
 */
int
read_config(file, rfdp)
	char *file;
	fd_set *rfdp;
{
	FILE *fp;
	int i;

	init_lex();

	if (opts & OPT_OLDSYSLOG)
		return read_oldconfig(file, rfdp);

	fp = fopen(file, "r");
	if (!fp && !strcmp(file, NSYSLOG_CONF) && strcmp(file, SYSLOG_CONF)) {
		if (opts & (OPT_DEBUG|OPT_VERBOSE))
			fprintf(stderr, "Trying to open %s\n", file);
		file = cfgfile = SYSLOG_CONF;
		fp = fopen(file, "r");
	}
	if (!fp)
		return -1;

	if (opts & (OPT_DEBUG|OPT_VERBOSE))
		fprintf(stderr, "Reading configuration from %s\n", file);

	yyin = fp;
	while (!feof(fp))
                if (yyparse() != 0)
			break;
	if (!feof(fp))
		yyparse();	/* avoid error when we handle HUP signal */
	fclose(fp);

	/*
	 * Couldn't make _any_ sense of this config file!
	 */
	if (token == 0) {
		if (!(opts & OPT_OLDSYSLOG)) {
			if (opts & (OPT_DEBUG|OPT_VERBOSE))
				fprintf(stderr,
					"Retrying as old format (%s)\n", file);
			opts |= OPT_OLDSYSLOG;
			i = read_oldconfig(file, rfdp);
			opts &= ~OPT_OLDSYSLOG;
			return i;
		}
		return -2;
	}

	/*
	 * Redisplay the configuration file as we see it.
	 */
	if (opts & (OPT_DEBUG|OPT_VERBOSE))
		dump_all();

	return setup_filetable(rfdp);
}


/*
 * allocate a new filetype structure and initialize the type to the one
 * passed.
 */
struct filetype *
new_filetype(type)
	int type;
{
	filetype_t *ft;

	if (!(NEW(ft, NEW_FT)))
		return NULL;
	ft->ft_type = type;
	ft->ft_vt = type_find(type);
	return ft;
}


/*
 * close all open files, fsync destination ones first.
 */
void
close_all()
{
	int fd, cnt = 0;
	void *ptr = NULL;

	while ((fd = next_dfd(&ptr, &cnt)) != -2) {
		fsync(fd);
		close(fd);
	}

	while ((fd = next_sfd(&ptr, &cnt)) != -2)
		close(fd);
}

int
reload()
{
	int n;

	expire_all();
	bzero((char *)fttable, fttsize * sizeof(*fttable));
	max_dfd = -1;
	max_sfd = -1;
	n = load_config();
	expire_old();
	hash_initfile(options.op_hashtype);
	return n;
}
