#include "h.h"
#include "var.h"
#include "nsyslog.h"
#include "hash.h"

extern char *optarg;
extern	int	errno;


#if BSD < 199306
# ifdef	__STDC__
extern	const char	*const sys_errlist[];
# else
extern	char	*sys_errlist[];
# endif
#endif

void
#ifdef __STDC__
syserr(char *fmt, ...)
#else
syserr(fmt, va_alist)
char	*fmt;
va_dcl
#endif
{
	char buf[MAXLINE], *s, *t = NULL;
	const char *errm;
	va_list pvar;
 
	errm = sys_errlist[errno];
	s = strstr(fmt, "%m");
	if (s != NULL) {
		t = (char *)malloc(strlen(fmt) + strlen(errm) + 3);
		strncpy(t, fmt, s - fmt);
		t[s - fmt] = '\0';
		strcat(t, errm);
		strcat(t, s+2);
		fmt = t;
	}
#ifdef __STDC__
	va_start(pvar, fmt);
#else
	va_start(pvar);
#endif
	vfprintf(stderr, fmt, pvar);
	va_end(pvar);
	fputc('\n', stderr);
	if (t)
		free(t);
}


void
hash_flush()
{
	return;
}

main(argc, argv)
	int argc;
	char *argv[];
{
	u_char salt[HASH_SALT_SZ], pass[MAXPASS + 1], mac[MAC_SIZE];
	char *logdir = NULL, *hashfile = NULL, *passfile = NULL, *hdir = NULL;
	char *saltfile = NULL, *logfile = NULL, *h0file = NULL;
	char *algo = "SHA1", *dbuf = NULL, *s, line[MAXLINE+1];
	int i, c, fd, skip = 0, hfd, err, sz;
	u_long cnts[3];
	struct hashrec hrec, hmac;
	hashtype_t *h;
	htable_t *ht;
	FILE *fp;

	while ((c = getopt(argc, argv, "a:d:h:H:l:n:p:s:")) != -1)
		switch (c)
		{
		case 'a' :
			algo = optarg;
			break;
		case 'd' :
			logdir = optarg;
			break;
		case 'D' :
			hdir = optarg;
			break;
		case 'h' :
			/*
			 * file to check against
			 */
			hashfile = optarg;
			break;
		case 'H' :
			h0file = optarg;
			break;
		case 'n' :
			/*
			 * skip this many records before starting validation.
			 */
			skip = atoi(optarg);
			break;
		case 'p' :
			/*
			 */
			passfile = optarg;
			break;
		case 'l' :
			/*
			 * the contents of this file are checked
			 */ 
			logfile = optarg;
			break;
		case 's' :
			/*
			 * read the salt from this file.
			 */
			saltfile = optarg;
			break;
		}

	if (logfile == NULL) {
		syserr("No logfile specified to check");
		exit(1);
	}

	/*
	 * check the hash algorigthm specified to see if we recognise it.
	 */
	ht = hash_lookup(algo);
	if (ht == NULL) {
		syserr("unknown hash algorithm %s", optarg);
		exit(1);
	}
	h = hash_get(ht->ht_type);
	hash_init(h);
	dbuf = malloc(h->ht_dsize);

	/*
	 * To validate the first line in the hashfile, we need to know the
	 * input hash file (H0) used.  This could come from one of two places:
	 * 1. the user enters the original password which is combined with
	 *    the original salt or;
	 * 2. the user supplies a path to a file containing the original H0
	 *    file.
	 */
	if (h0file == NULL) {
		if (saltfile == NULL) {
			if (access(NSYSLOG_SALT, R_OK) == -1) {
				syserr("Need salt file with password");
				exit(1);
			}
			saltfile = NSYSLOG_SALT;
		}
		hash_readsalt(saltfile, salt, sizeof(salt));

		if (passfile == NULL) {
			/*
			 * read password/phrase in from stdin.
			 */
			printf("Enter the initial password.\n");
			fp = stdin;
		} else {
			/*
			 * read password/phrase in from a file.
			 */
			fp = fopen(passfile, "r");
			if (fp == NULL) {
				perror("fopen");
				exit(1);
			}
		}

		fflush(stdin);

		do {
			bzero(pass, sizeof(pass));
			if (fp == stdin)
				printf("Password:");
			if (!fgets(pass, sizeof(pass), fp))
				exit(0);
			if (strlen(pass) < MINPASS)
				continue;
		} while (pass[strlen(pass) - 1] != '\n');
		if (fp != stdin)
			fclose(fp);
	} else {
		/*
		 * read H0 for the logfile from a file.
		 */
		fd = hash_openH0("", h0file, O_RDONLY);
		if (fd >= 0) {
			err = read(fd, dbuf, h->ht_dsize);
			if (err != h->ht_dsize) {
				syserr("error reading H0 for %s: %m", h0file);
				err = -1;
			} else {
				err = read(fd, (char *)cnts, sizeof(cnts));
				if (err != sizeof(cnts)) {
					syserr("error reading line count: %m");
					err = -1;
				}
			}
			close(fd);
		}
	}
	fp = fopen(logfile, "r");
	if (fp == NULL) {
		syserr("Unable to open logfile %s: %m\n", logfile);
		exit(1);
	}
	hfd = hash_openfile(hdir, hashfile, logdir, logfile, h);
	if (hfd == 01) {
		syserr("cannot open hashfile:%m");
		exit(1);
	}
	err = read(hfd, (char *)&hrec, sizeof(hrec));
	if (err != sizeof(hrec)) {
		perror("read");
		exit(1);
	}
	if ((hrec.hr_lcnt != 0) || (hrec.hr_pri != LOG_START)) {
		fprintf(stderr, "Invalid start record for hash file\n");
		exit(1);
	}
	printf("Hash log started at %s", ctime(&hrec.hr_when));
	if (h0file == NULL) {
		sz = MIN(sizeof(pass), h->ht_dsize);
		/*
		 * hr_hcnt should contain the number of passes needed to do
		 * another `hash_nseconds'
		 */
		for (i = hrec.hr_hcnt - 1; i >= 0; i--) {
			hash_init(h);
			hash_update(h, pass, sizeof(pass));
			hash_final(h, dbuf);
			bcopy(dbuf, pass, sz);
		}
		hash_init(h);
		hash_update(h, pass, sizeof(pass));
		hash_update(h, salt, sizeof(salt));
		s = strrchr(logfile, '/');
		if (s != NULL)
			s++;
		else
			s = logfile;
		hash_update(h, s, strlen(s));
		hash_final(h, dbuf);
	}

	/*
	 * Validate the text-log file against the hash log file, line by
	 * line, record by record.
	 */
	for (i = 1; fgets(line, sizeof(line), fp); i++) {
		err = read(hfd, (char *)&hmac, sizeof(hmac));
		if (err != sizeof(hmac)) {
			perror("read");
			exit(1);
		}
		if (hmac.hr_pri == LOG_CHKPOINT) {
			if (bcmp(dbuf, hmac.hr_mac, MAC_SIZE))
				fprintf(stderr, "H0 Invalid\n");
			continue;
		}
		bcopy((char *)&hmac, (char *)&hrec, sizeof(hmac));
		bzero(hmac.hr_mac, sizeof(hmac.hr_mac));
		/*
		 * Calculate what the MAC should be.
		 */
		hash_init(h);
		hash_update(h, dbuf, h->ht_dsize);
		hash_update(h, line, strlen(line));
		hash_update(h, (char *)&hmac, sizeof(hmac));
		hash_final(h, hmac.hr_mac);
		if (bcmp(hrec.hr_mac, hmac.hr_mac, MAC_SIZE))
			fprintf(stderr, "Line %d doesn't match\n", i);
		else
			fprintf(stderr, "Line %d matches\r", i);
		/*
		 * Recreate H0 for the next line.
		 */
		hash_init(h);
		hash_update(h, dbuf, h->ht_dsize);
		hash_final(h, dbuf);
	}
	fputc('\n', stderr);
	fclose(fp);
	return 0;
}
