#include "h.h"
#include "nsyslog.h"
#include "var.h"
#include "y.tab.h"

#ifdef	HAVE_DOOR_H

#include <door.h>

typedef struct doorspec {
	struct doorspec *ds_next;
	int ds_cnt;
	int ds_fd;
	int ds_mode;
	int ds_pri;
	time_t ds_when;
	buflist_t *ds_rlist;
	buflist_t **ds_rtail;
	char *ds_path;
} doorspec_t;


void * door_vinit __P((int, arglist_t *));
void door_reopen __P((void *));
void door_print __P((void *));
int door_fd __P((void *));
int door_inccnt __P((void *, int));
void door_expire __P((void));
void door_expireold __P((void));

static	doorspec_t	*dslist = NULL;
vartype_t	doortype = {
	NULL,		/* next */
	NULL,		/* ptr */
	SRC_DOOR,	/* stype */
	0,		/* dtype */
	"door",		/* name */
	door_vinit,	/* init */
	NULL,		/* read */
	NULL,		/* readnext */
	NULL,		/* write */
	NULL,		/* link */
	NULL,		/* match */
	NULL,		/* accept */
	door_reopen,	/* reopen */
	NULL,		/* logmsg */
	door_print,	/* print */
	door_fd,	/* fd */
	door_inccnt,	/* inccnt */
	NULL,		/* drain */
	door_expire,	/* expire */
	door_expireold,	/* expireold */
	NULL,		/* more */
};


extern	int	do_raw;
extern	doorspec_t	*door_new __P((char *));


void *
door_vinit(type, args)
	int type;
	arglist_t *args;
{
	int proto;
	char *p1;

	if (args == NULL)
		return NULL;
	p1 = args->a_arg;
	if (args->a_next != NULL)
		return NULL;
	return door_new(p1);
}


doorspec_t *
door_find(parm)
	char *parm;
{
	doorspec_t *ds;

	for (ds = dslist; ds; ds = ds->ds_next)
		if (!strcmp(ds->ds_path, parm))
			return ds;
	return NULL;
}


/*
 * NOTE: In reality, Solaris (upto and including 7), use doors only for
 *       some sort of synchronisation.  This function, door_server(),
 *       could simply call door_return() and work as expected.  For now,
 *       at least, this is all that is recommended be done.  Allowing
 *       syslog messages to be sent through doors still needs to be
 *       evaluated in terms of message content and general security
 *       concerns.
 */
void door_server(cookie, argp, arg_size, dp, n_desc)
	void *cookie;
	char *argp;
	size_t arg_size;
	door_desc_t *dp;
	size_t n_desc;
{
#if 0
	logmsgtype_t lm;
	doorspec_t *ds;

	ds = cookie;

	bzero((char *)&lm, sizeof(lm));
	lm.lm_flags = FLG_USER;
	if (ds != NULL) {
		lm.lm_ft = fttable + ds->ds_fd;
		if (argp != NULL && arg_size > 0) {
			lm.lm_log = new_logmsg(argp, arg_size);
			lm.lm_when = time(NULL);
			logmsg(&lm);
		}
	}
#endif
	door_return(NULL, 0 , NULL, 0);
}


doorspec_t *
door_new(parm)
	char *parm;
{
	doorspec_t *ds;
	int fd;

	if ((ds = door_find(parm))) {
		ds->ds_mode &= ~VAR_OLD;
		return ds;
	}

	if (!(NEW(ds, NEW_DOOR)))
		return NULL;
	ds->ds_rtail = &ds->ds_rlist;

	if (do_raw) {
		ds->ds_mode |= do_raw;
		do_raw = 0;
	}
	ds->ds_next = dslist;
	ds->ds_path = strdup(parm);
	fd = open(ds->ds_path, 0666, O_CREAT|O_RDONLY);
	if (fd == -1)
		syserr("door creation failed [%s]:%m", ds->ds_path);
	else {
		fcntl(fd, F_SETFD, fcntl(fd, F_GETFD, 0) | FD_CLOEXEC);
		close(fd);
		fdetach(ds->ds_path);
		fd = door_create(door_server, &ds->ds_fd, 0);
		if (fd == -1)
			syserr("door creation failed [%s]:%m", ds->ds_path);
		else
			fattach(fd, ds->ds_path);
	}
	ds->ds_fd = fd;
	ds->ds_cnt = 0;
	dslist = ds;
	return ds;
}


void
door_free(ds)
	doorspec_t *ds;
{
	doorspec_t *s, **sp;
	buflist_t *b;

	if (ds->ds_cnt > 0)
		return;

	assert(ds->ds_cnt == 0);
	for (sp = &dslist; (s = *sp); )
		if (s == ds) {
			*sp = s->ds_next;
			break;
		} else
			sp = &s->ds_next;
	buf_freelist(ds->ds_rlist);
	door_revoke(ds->ds_fd);
	free(ds->ds_path);
	free(ds);
}


void
door_expire()
{
	doorspec_t	*ds, **dsp;

	for (dsp = &dslist; (ds = *dsp); ) {
		assert(ds->ds_cnt == 0);
		ds->ds_mode |= VAR_OLD;
		dsp = &ds->ds_next;
		if (opts & OPT_VERBOSE)
			fprintf(stderr, "keep door: %s (%d)\n",
				ds->ds_path, ds->ds_cnt);
	}
}


void
door_expireold()
{
	doorspec_t	*ds, **dsp;

	for (dsp = &dslist; (ds = *dsp); ) {
		if (!(ds->ds_mode & VAR_OLD)) {
			dsp = &ds->ds_next;
			continue;
		}
		if (opts & OPT_VERBOSE)
			fprintf(stderr, "close door: %s (%d)\n",
				ds->ds_path, ds->ds_cnt);
		dsp = &dslist;
		door_free(ds);
	}
}


void
door_drain(ptr)
	void *ptr;
{
	doorspec_t *ds;

	ds = ptr;
	return;
}


void
door_reopen(ptr)
	void *ptr;
{
	doorspec_t *ds;

	ds = ptr;
	return;
}


int
door_inccnt(ptr, inc)
	void *ptr;
	int inc;
{
	doorspec_t *ds;

	ds = ptr;
	ds->ds_cnt += inc;
	return ds->ds_cnt;
}


int
door_fd(ptr)
	void *ptr;
{
	doorspec_t *ds;

	ds = ptr;
	return ds->ds_fd;
}


void
door_print(ptr)
	void *ptr;
{
	doorspec_t *ds;

	ds = ptr;
	printf(" door %s ", ds->ds_path);
}

#endif	/* HAVE_DOOR_H */
