#include "h.h"
#include "nsyslog.h"
#include "var.h"
#include "hash.h"
#include "y.tab.h"


#if !defined(lint)
static const char rcsid[] = "$Id: fac.c,v 1.8 1999/05/10 18:18:42 darrenr Exp $";
#endif


typedef struct facspec {
	struct facspec *fs_next;
	int fs_cnt;
	int fs_fac;
	int fs_not;
} facspec_t;


#define	OK_FAC(a,b)	(((a) & LOG_FACMASK) == ((b) & LOG_FACMASK))


facspec_t	*faclist = NULL;
void fac_print __P((void *));
int fac_inccnt __P((void *, int ));
void fac_expire __P((void));
int fac_match __P((void *, logmsgtype_t *));

vartype_t	factype = {
	NULL,		/* next */
	NULL,		/* ptr */
	FIL_FAC,	/* stype */
	0,		/* dtype */
	"fac",		/* name */
	NULL,		/* init */
	NULL,		/* read */
	NULL,		/* readnext */
	NULL,		/* write */
	NULL,		/* link */
	fac_match,	/* match */
	NULL,		/* accept */
	NULL,		/* reopen */
	NULL,		/* logmsg */
	fac_print,	/* print */
	NULL,		/* fd */
	fac_inccnt,	/* inccnt */
	NULL,		/* drain */
	fac_expire,	/* expire */
	NULL,	/* expireold */
	NULL,		/* more */
};


/*
 * Letters per facility were added to help ensure compatibility with IRIX.
 */
table_t	facs[] = {
	{ "kern", LOG_KERN, 'A' },	{ "user", LOG_USER, 'B' },
	{ "mail", LOG_MAIL, 'C' },	{ "daemon", LOG_DAEMON, 'D' },
	{ "auth", LOG_AUTH, 'E' },	{ "syslog", LOG_SYSLOG, 'F' },
	{ "lpr", LOG_LPR, 'G' },	{ "news", LOG_NEWS, 'H' },
	{ "uucp", LOG_UUCP, 'I' },
#ifdef sun
	{ "cron2", LOG_CRON1, 'W' },
#else
	{ "cron", LOG_CRON1, 'J' },
#endif
	{ "ftp", LOG_FTP, 'U' },	{ "authpriv", LOG_AUTHPRIV, 'V' },
	{ "audit", LOG_AUDIT, 'K' },	{ "logalert", LOG_LFMT, 'L' },
#ifdef	sun
	{ "cron", LOG_CRON2, 'J' },
#else
	{ "cron2", LOG_CRON2, 'W' },
#endif
	{ "mark", LOG_MARK, 'X' },
	{ "local0", LOG_LOCAL0, 'M' },	{ "local1", LOG_LOCAL1, 'N' },
	{ "local2", LOG_LOCAL2, 'O' },	{ "local3", LOG_LOCAL3, 'P' },
	{ "local4", LOG_LOCAL4, 'Q' },	{ "local5", LOG_LOCAL5, 'R' },
	{ "local6", LOG_LOCAL6, 'S' },	{ "local7", LOG_LOCAL7, 'T' },
	{ NULL, 0 }
};


table_t *
fac_find(fac)
	int fac;
{
	int	i;

	for (i = 0; facs[i].name; i++)
		if (fac == facs[i].value)
			return facs + i;
	return NULL;
}


/*
 * map a facility number to its name
 */
char *
fac_toname(facpri)
	int facpri;
{
	int	i, fac;

	for (i = 0, fac = facpri & LOG_FACMASK; facs[i].name; i++)
		if (fac == facs[i].value)
			return facs[i].name;
	return NULL;
}


/*
 * map a facility name to its number
 */
int
fac_findname(name)
	char *name;
{
	int	i;

	for (i = 0; facs[i].name; i++)
		if (!strcmp(facs[i].name, name))
			return facs[i].value;
	return -1;
}


facspec_t *
fac_finduse(fac, not)
	int fac, not;
{
	facspec_t *fs;

	for (fs = faclist; fs; fs = fs->fs_next)
		if ((fs->fs_fac == fac) && (fs->fs_not == not))
			return fs;
	return NULL;
}


int
fac_ok(pri, fac)
	int pri;
	int fac;
{
	return OK_FAC(pri, fac);
}


int
fac_match(ptr, lmp)
	void *ptr;
	logmsgtype_t *lmp;
{
	facspec_t *fs;
	int ok;

	fs = ptr;
	ok = fac_ok(lmp->lm_pri, fs->fs_fac);
	ok ^= fs->fs_not;
	return !ok;
}


facspec_t *
fac_new(parm)
	char *parm;
{
	facspec_t *fs;
	int fac, not;

	not = get_not();

	if ((fac = fac_findname(parm)) < 0)
		return NULL;

	if ((fs = fac_finduse(fac, not)))
		return fs;

	if (!(NEW(fs, NEW_FACS)))
		return NULL;
	fs->fs_fac = fac;
	fs->fs_next = faclist;
	fs->fs_not = not;
	fs->fs_cnt = 0;
	faclist = fs;
	return fs;
}


facspec_t *
facval_new(fac)
	int fac;
{
	facspec_t *fs;
	int not;

	not = get_not();

	if ((fs = fac_finduse(fac, not)))
		return fs;

	if (!(NEW(fs, NEW_FACS)))
		return NULL;
	fs->fs_fac = fac;
	fs->fs_next = faclist;
	fs->fs_cnt = 0;
	fs->fs_not = not;
	faclist = fs;
	return fs;
}


void
fac_expire()
{
	facspec_t	*fcs;

	while ((fcs = faclist)) {
		assert(fcs->fs_cnt == 0);
		faclist = fcs->fs_next;
		free(fcs);
	}
}


int
fac_inccnt(ptr, inc)
	void *ptr;
	int inc;
{
	facspec_t *fs;

	fs = ptr;
	fs->fs_cnt += inc;
	return fs->fs_cnt;
}


void
fac_print(ptr)
	void *ptr;
{
	facspec_t *fs;

	fs = ptr;
	print_not(fs->fs_not);
	printf(" facility %s ", fac_toname(fs->fs_fac));
}
