#include "h.h"
#include "nsyslog.h"
#include "var.h"
#include "hash.h"
#include "y.tab.h"


#if !defined(lint)
static const char rcsid[] = "$Id: facpri.c,v 1.7 1999/05/10 18:18:43 darrenr Exp $";
#endif


typedef struct facprispec {
	struct facprispec *fp_next;
	int fp_cnt;
	int fp_facpri;
	int fp_comp;
	int fp_not;
} facprispec_t;


facprispec_t	*fplist = NULL;
void facpri_print __P((void *));
int facpri_inccnt __P((void *, int ));
void facpri_expire __P((void));
int facpri_match __P((void *, logmsgtype_t *));

vartype_t	facpritype = {
	NULL,		/* next */
	NULL,		/* ptr */
	FIL_FACPRI,	/* stype */
	0,		/* dtype */
	"facpri",		/* name */
	NULL,		/* init */
	NULL,		/* read */
	NULL,		/* readnext */
	NULL,		/* write */
	NULL,		/* link */
	facpri_match,	/* match */
	NULL,		/* accept */
	NULL,		/* reopen */
	NULL,		/* logmsg */
	facpri_print,	/* print */
	NULL,		/* fd */
	facpri_inccnt,	/* inccnt */
	NULL,		/* drain */
	facpri_expire,	/* expire */
	NULL,	/* expireold */
	NULL,		/* more */
};


facprispec_t *
facpri_find(facpri, not)
	int facpri;
	int not;
{
	facprispec_t *fp;

	for (fp = fplist; fp; fp = fp->fp_next)
		if (fp->fp_facpri == facpri && fp->fp_not == not)
			return fp;
	return NULL;
}


int
facpri_match(ptr, lmp)
	void *ptr;
	logmsgtype_t *lmp;
{
	facprispec_t *fp;
	int ok;

	fp = ptr;
	ok = fac_ok(fp->fp_facpri, lmp->lm_pri);
	ok ^= fp->fp_not;
	if (!ok)
		return !ok;
	ok = pri_matchvals(lmp->lm_pri, fp->fp_facpri, fp->fp_comp);
	ok ^= fp->fp_not;
	return !ok;
}


facprispec_t *
facpri_new(parm1, comp, parm2)
	char *parm1;
	int comp;
	char *parm2;
{
	int fac, pri, not;
	facprispec_t *fp;

	not = get_not();

	if ((fac = fac_findname(parm1)) < 0) {
		fprintf(stderr, "unknown facility %s\n", parm1);
		return NULL;
	}
	if ((pri = pri_findname(parm2)) < 0) {
		fprintf(stderr, "unknown priority %s\n", parm2);
		return NULL;
	}

	if ((fp = facpri_find(fac|pri))) {
		return fp;
	}

	if (!(NEW(fp, NEW_FACPRIS)))
		return NULL;
	fp->fp_facpri = fac|pri;
	fp->fp_next = fplist;
	fp->fp_not = not;
	fp->fp_comp = comp;
	fp->fp_cnt = 0;
	fplist = fp;
	return fp;
}


facprispec_t *
facprival_new(facpri, comp)
	int facpri, comp;
{
	facprispec_t *fp;
	int not;

	not = get_not();

	if ((fp = facpri_find(facpri, not)))
		return fp;

	if (!(NEW(fp, NEW_FACPRIS)))
		return NULL;
	fp->fp_facpri = facpri;
	fp->fp_next = fplist;
	fp->fp_cnt = 0;
	fp->fp_not = not;
	fp->fp_comp = comp;
	fplist = fp;
	return fp;
}


void
facpri_expire()
{
	facprispec_t	*fp;

	while ((fp = fplist)) {
		assert(fp->fp_cnt == 0);
		fplist = fp->fp_next;
		free(fp);
	}
}


void
facpri_print(ptr)
	void *ptr;
{
	facprispec_t *fp;

	fp = ptr;
	print_not(fp->fp_not);
	printf(" facpri %s, %s%s", fac_toname(fp->fp_facpri),
		comp_toname(fp->fp_comp), pri_toname(fp->fp_facpri));
}


int
facpri_inccnt(ptr, inc)
	void *ptr;
	int inc;
{
	facprispec_t *fs;

	fs = ptr;
	fs->fp_cnt += inc;
	return fs->fp_cnt;
}
