#include "h.h"
#include "nsyslog.h"
#include "var.h"
#include "hash.h"
#include "y.tab.h"


#if !defined(lint)
static const char rcsid[] = "$Id: file.c,v 1.15 1999/05/10 18:18:45 darrenr Exp $";
#endif

filespec_t * file_new __P((int, char *));
void * file_vinit __P((int, arglist_t *));
int file_recv __P((void *, logmsgtype_t *, tcplmsg_t *, char *, size_t));
void file_reopen __P((void *));
int file_logmsg __P((void *, int *, struct iovec *, int, logmsgtype_t *));
void file_print __P((void *));
int file_fd __P((void *));
int file_inccnt __P((void *, int));
void file_expire __P((void));
void file_expireold __P((void));

filespec_t	*fslist = NULL;
static	filespec_t	*cur_file = NULL;
vartype_t	filetype = {
	NULL,		/* next */
	NULL,		/* ptr */
	SRC_FILE,	/* stype */
	DEST_FILE,	/* dtype */
	"file",		/* name */
	file_vinit,	/* init */
	file_recv,	/* read */
	NULL,		/* readnext */
	NULL,		/* write */
	NULL,		/* link */
	NULL,	 	/* match */
	NULL,		/* accept */
	file_reopen,	/* reopen */
	file_logmsg,	/* logmsg */
	file_print,	/* print */
	file_fd,	/* fd */
	file_inccnt,	/* inccnt */
	NULL,		/* drain */
	file_expire,	/* expire */
	file_expireold,	/* expireold */
	NULL,		/* more */
};


filespec_t *
file_find(mode, parm)
	int mode;
	char *parm;
{
	filespec_t *fs;

	for (fs = fslist; fs; fs = fs->fs_next)
		if ((VAR_TYPE(fs->fs_mode) == mode) &&
		    !strcmp(fs->fs_path, parm))
			return fs;
	return NULL;
}


int
open_file(mode, path)
	int mode;
	char *path;
{
	struct stat sb;
	char *fullpath;
	int fd;

	if (*path == '/')
		fullpath = path;
	else {
		fullpath = (char *)malloc(strlen(path) +
					  strlen(options.op_logdir) + 2);
		strcpy(fullpath, options.op_logdir);
		strcat(fullpath, "/");
		strcat(fullpath, path);
	}

	if (mode == DEST_FILE) {
		if ((stat(fullpath, &sb) == 0) && S_ISCHR(sb.st_mode))
			alarm(2);
#ifdef	MAKE_LOG_FILES
		fd = open(fullpath, O_WRONLY|O_APPEND|O_CREAT, 0600);
#else
		fd = open(fullpath, O_WRONLY|O_APPEND);
#endif
		alarm(0);
	} else
		fd = open(fullpath, O_RDONLY);
	if (fd != -1)
		fcntl(fd, F_SETFD, fcntl(fd, F_GETFD, 0) | FD_CLOEXEC);

	if (fullpath != path)
		free(fullpath);
	return fd;
}


void *
file_vinit(type, args)
	int type;
	arglist_t *args;
{
	char *path;

	if (args == NULL)
		return NULL;
	if (args->a_next != NULL)
		return NULL;
	return file_new(type, args->a_arg);
}


filespec_t *
file_new(mode, parm)
	int mode;
	char *parm;
{
	struct stat sb;
	filespec_t *fs;

	if ((fs = file_find(mode, parm))) {
		fs->fs_mode &= ~VAR_OLD;
		ft_set(fs->fs_fd, fs, mode);
		return fs;
	}

	if (!(NEW(fs, NEW_FILES)))
		return NULL;
	cur_file = fs;
	fs->fs_mode = mode;
	fs->fs_path = strdup(parm);
	fs->fs_syncfreq = options.op_syncfreq;
	fs->fs_lastsync = time(NULL);
	fs->fs_markfreq = options.op_markfreq;
	fs->fs_next = fslist;
	fs->fs_fd = open_file(mode, parm);
	if ((mode == DEST_FILE) && (fs->fs_fd >= 0)) {
		if (fstat(fs->fs_fd, &sb))
			syserr("fstat(%d) failed: %m", fs->fs_fd);
		else {
			if (S_ISREG(sb.st_mode))
				fs->fs_mode |= VAR_HASH;
		}
	}
	if ((fs->fs_mode & VAR_HASH) && (options.op_hashtype != HASH_NONE) &&
	    FILE_DOHASH(fs)) {
		fs->fs_hptr = hash_find(options.op_hashtype);
		if (fs->fs_hptr != NULL)
			fs->fs_hfd = hash_openfile(options.op_hashdir, NULL,
						   options.op_logdir, parm,
						   fs->fs_hptr);
	} else
		fs->fs_hfd = -1;
	if (fs->fs_fd >= 0)
		ft_set(fs->fs_fd, fs, mode);
	fs->fs_cnt = 0;
	fslist = fs;
	return fs;
}


void
file_end()
{
	cur_file = NULL;
}


void
set_sync(syncfreq)
	int syncfreq;
{
	cur_file->fs_syncfreq = syncfreq;
}


void
set_mark(markfreq)
	int markfreq;
{
	cur_file->fs_markfreq = markfreq;
}


void
file_free(fs)
	filespec_t *fs;
{
	assert(fs->fs_cnt == 0);
	if (fs->fs_dbuf)
		free(fs->fs_dbuf);
	free(fs->fs_path);
	if (fs->fs_hfd >= 0)
		close(fs->fs_hfd);
	if (fs->fs_fd >= 0)
		close(fs->fs_fd);
	free(fs);
}


void
file_expire()
{
	filespec_t	*fs, **fsp;

	for (fsp = &fslist; (fs = *fsp); ) {
		if (VAR_TYPE(fs->fs_mode) == SRC_FILE) {
			fs->fs_mode |= VAR_OLD;
			fsp = &fs->fs_next;
			continue;
		}
		*fsp = fs->fs_next;
		file_free(fs);
	}
}


void
file_expireold()
{
	filespec_t	*fs, **fsp;

	for (fsp = &fslist; (fs = *fsp); ) {
		if (!(fs->fs_mode & VAR_OLD)) {
			fsp = &fs->fs_next;
			continue;
		}
		*fsp = fs->fs_next;
		file_free(fs);
	}
}


int
file_logmsg(ptr, fdp, iov, nio, lmp)
	void *ptr;
	int *fdp;
	struct iovec *iov;
	int nio;
	logmsgtype_t *lmp;
{
	filespec_t *fs;
	filetype_t *ft;
	time_t now;
	int fd, nw;

	fs = ptr;
	fd = fs->fs_fd;
	ft = fttable + fd;
	nw = 0;

	/*
	* If we have written a line to a file, check if we should
	* call fsync() on the fd yet.
	*/
	now = time(NULL);
	if (opts & OPT_VERBOSE)
		fprintf(stderr, "msg to file %s\n", fs->fs_path);
	if (fd >= 0)
		nw = writev(fd, iov, nio);
	if (nw > 0) {
		fs->fs_written++;
		if (((fs->fs_syncfreq != 0) &&
		     (fs->fs_written >= fs->fs_syncfreq)) ||
		    ((fs->fs_synctime != 0) &&
		     (now - fs->fs_lastsync > fs->fs_synctime)) ||
		    ((fs->fs_syncpri >= 0) &&
		     (fs->fs_syncpri <= LOG_PRI(lmp->lm_pri)))) {
			fsync(fd);
			fs->fs_lastsync = now;
			fs->fs_written = 0;
		}
		if ((fs->fs_hfd >= 0) && fs->fs_hptr)
			hash_addrec(lmp, ft, iov, nio);
	} else if (nw < 0)
		syserr("error writing to %s: %m", fs->fs_path);
	return nw;
}


void
file_reopen(ptr)
	void *ptr;
{
	filespec_t *fs;

	fs = ptr;
	fs->fs_fd = open_file(fs->fs_mode, fs->fs_path);
}


int
file_recv(ptr, lmp, tlm, buf, bufsz)
	void *ptr;
	logmsgtype_t *lmp;
	tcplmsg_t *tlm;
	char *buf;
	size_t bufsz;
{
	filespec_t *fs;

	fs = ptr;
	lmp->lm_src = NULL;
	lmp->lm_slen = 0;
	lmp->lm_flags = FLG_SYS;
	return read(fs->fs_fd, buf, bufsz);
}


int
file_inccnt(ptr, inc)
	void *ptr;
	int inc;
{
	filespec_t *fs;

	fs = ptr;
	fs->fs_cnt += inc;
	return fs->fs_cnt;
}


int
file_fd(ptr)
	void *ptr;
{
	filespec_t *fs;

	fs = ptr;
	return fs->fs_fd;
}


void
file_print(ptr)
	void *ptr;
{
	filespec_t *fs;

	fs = ptr;
	if (fs->fs_mode & VAR_INTERNAL)
		printf(" internal ");
	else
		printf(" file %s ", fs->fs_path);
}
