#include "h.h"
#include "nsyslog.h"
#include "var.h"
#include "hash.h"
#include "sha1.h"


static	void	hash_updatedbuf __P((hashtype_t *, char *));
static	int	hash_nseconds __P((int, hashtype_t *, char *, size_t));


/*
 * initialize hashing by reading in the password from a file.
 */
hashtype_t *
hash_initfile(type)
	int type;
{
	hashtype_t *hptr = NULL;
	struct hashrec hrec;
	int fd, err = 0;
	struct stat st;
	filespec_t *fs;
	netspec_t *ns;

	if (type == HASH_NONE)
		return;
	hptr = hash_find(type);
	if (hptr == NULL) {
		syserr("unknown hashing type");
		return NULL;
	}
	bzero((char *)&hrec, sizeof(hrec));
	hrec.hr_when = time(NULL);
	hrec.hr_pri = LOG_START;
	(*hptr->ht_create)(hptr);

	/*
	 * read in the 'H0's from the various files.
	 */
	for (fs = fslist; fs != NULL; fs = fs->fs_next) {
		if (!FILE_DOHASH(fs))
			continue;

		fs->fs_hptr = hptr;
		fd = hash_openH0(options.op_hashdir, fs->fs_path, O_RDONLY);
		if (fd >= 0) {
			fs->fs_dbuf = malloc(hptr->ht_dsize);
			err = read(fd, fs->fs_dbuf, hptr->ht_dsize);
			if (err != hptr->ht_dsize) {
				syserr("error reading H0 for %s", fs->fs_path);
				err = -1;
			} else {
				err = read(fd, (char *)fs->fs_cnts,
					   sizeof(fs->fs_cnts));
				if (err != sizeof(fs->fs_cnts)) {
					syserr("error reading line count (%s)",
						fs->fs_path);
					err = -1;
				}
			}
			if (err != -1) {
				if (fstat(fs->fs_hfd, &st) == 0) {
					if (st.st_size == 0) {
						fs->fs_lcnt = 0;
						fs->fs_fcnt++;
						hrec.hr_lcnt = 0;
						hrec.hr_hcnt = fs->fs_hcnt;
						hrec.hr_fcnt = fs->fs_fcnt;
						err = write(fs->fs_hfd, &hrec,
							    sizeof(hrec));
						if (err == -1)
							syserr("starting hash file for %s", fs->fs_path);
					}
				} else
					syserr("fstat(hfd=%d)", fs->fs_hfd);
			}
			close(fd);
		}
	}
	for (ns = nslist; ns != NULL; ns = ns->ns_next) {
		if (ns->ns_path == NULL)
			continue;
		if (ns->ns_type != DEST_TCP)
			continue;
		hptr = hash_find(ns->ns_hash);
		if (hptr == NULL)
			continue;
		fd = hash_openH0(options.op_hashdir, ns->ns_path, O_RDONLY);
		if (fd >= 0) {
			ns->ns_dbuf = malloc(hptr->ht_dsize);
			if (read(fd, ns->ns_dbuf, hptr->ht_dsize) !=
			    hptr->ht_dsize)
				syserr("error reading H0 for %s", fs->fs_path);
			else if (read(fd, (char *)&ns->ns_lcnt,
				      sizeof(ns->ns_lcnt)) !=
				 sizeof(ns->ns_lcnt))
				syserr("error reading line count for %s",
					ns->ns_path);
			close(fd);
		} else
			return NULL;
	}
	return hptr;
}


/*
 * ask for a password to initialize the hashing.
 */
void
hash_initpass(type)
	int type;
{
	char pass[MAXPASS + 1], salt[HASH_SALT_SZ];
	hashtype_t *hptr;
	hashrec_t hrec;
	filespec_t *fs;
	netspec_t *ns;
	htable_t *ht;
	char *s;
	int n;

	hptr = hash_find(type);
	if (hptr == NULL) {
		fprintf(stderr, "no hashing algorithm selected\n");
		return;
	}

	(*hptr->ht_create)(hptr);


	printf("Enter a password, between %d and %d characters long.\n",
		MINPASS, MAXPASS);

	do {
		printf("	 1---5----1----5----2----5----3----5----4\n");
		printf("Password:");
		fflush(stdout);
		bzero(pass, sizeof(pass));
		if (!fgets(pass, sizeof(pass), stdin))
			exit(0);
		if (strlen(pass) < MINPASS)
			continue;
	} while (pass[strlen(pass) - 1] != '\n');

	/*
	 * While there is one (only) password input, we need to generate an
	 * H0 for each file we store hashes for and make them all different
	 * yet able to be later recalculated.
	 */
	bzero(hrec.hr_mac, sizeof(hrec.hr_mac));
	hrec.hr_when = time(NULL);
	hrec.hr_pri = LOG_START;
	n = hash_nseconds(HASH_REHASHTIME, hptr, pass, sizeof(pass));
	hash_getsalt(salt, sizeof(salt));
	hash_savesalt(NULL, salt, sizeof(salt));

	for (fs = fslist; fs != NULL; fs = fs->fs_next) {
		s = strrchr(fs->fs_path, '/');
		if (s != NULL)
			s++;
		else
			s = fs->fs_path;
		fs->fs_dbuf = malloc(hptr->ht_dsize);
		hash_init(hptr);
		hash_update(hptr, pass, sizeof(pass));
		hash_update(hptr, salt, sizeof(salt));
		hash_update(hptr, s, strlen(s));
		hash_final(hptr, fs->fs_dbuf);
		fs->fs_hptr = hptr;
		fs->fs_hcnt = n + 1;
		fs->fs_lcnt = 0;
		fs->fs_fcnt = 0;
		/*
		 * Write initial record to file saying when it was started and
		 * how many hashes were done in n seconds.
		 */
		hrec.hr_hcnt = n;
		hrec.hr_lcnt = 0;
		hrec.hr_fcnt = 0;
		if (fs->fs_hfd == -1)
			fs->fs_hfd = hash_openH0(options.op_hashdir,
						 fs->fs_path, O_WRONLY);
		if (write(fs->fs_hfd, &hrec, sizeof(hrec)) != sizeof(hrec))
			perror("write");
	}

	for (ns = nslist; ns != NULL; ns = ns->ns_next)
		if (ns->ns_type == DEST_TCP) {
			ns->ns_dbuf = malloc(hptr->ht_dsize);
			hash_init(hptr);
			hash_update(hptr, pass, sizeof(pass));
			hash_update(hptr, salt, sizeof(salt));
			hash_update(hptr, ns->ns_path, strlen(ns->ns_path));
			hash_final(hptr, ns->ns_dbuf);
			ns->ns_hptr = hptr;
			ns->ns_lcnt = n + 1;
		}
	bzero(pass, sizeof(pass));
	hash_flush();
}


/*
 * for a given log message which has been saved to a file, calculate the
 * hash for it.
 */
void
hash_addrec(lmp, ft, iov, cnt)
	logmsgtype_t *lmp;
	filetype_t *ft;
	struct iovec *iov;
	int cnt;
{
	char *dbuf, *path, *mac;
	hashrec_t hrec, *hrp;
	hashtype_t *hptr;
	int i, hfd;

	if (ft->ft_type == DEST_FILE) {
		filespec_t *fs = ft->ft_ptr;

		hfd = fs->fs_hfd;
		hptr = fs->fs_hptr;
		dbuf = fs->fs_dbuf;
		path = fs->fs_path;
		hrp = &hrec;
		hrp->hr_hcnt = fs->fs_hcnt++;
		hrp->hr_lcnt = fs->fs_lcnt++;
		hrp->hr_fcnt = fs->fs_fcnt;
	} else if (ft->ft_type == DEST_TCP) {
		netspec_t *ns = ft->ft_ptr;

		hfd = -1;
		hptr = ns->ns_hptr;
		dbuf = ns->ns_dbuf;
		path = ns->ns_path;
		hrp = (hashrec_t *)ns->ns_hmac;
		hrp->hr_lcnt = ns->ns_lcnt++;
		hrp->hr_hcnt = hrp->hr_lcnt;
		hrp->hr_fcnt = 0;
	}
	if (hptr == NULL || hptr->ht_type == HASH_NONE)
		return;

	hrp->hr_pri = lmp->lm_pri;
	hrp->hr_when = lmp->lm_when;
	mac = hrp->hr_mac;
	bzero(mac, MAC_SIZE);

	hash_init(hptr);
	hash_update(hptr, dbuf, hptr->ht_dsize);
	for (i = 0; i < cnt; i++)
		hash_update(hptr, iov[i].iov_base, iov[i].iov_len);
	hash_update(hptr, (char *)hrp, sizeof(*hrp));
	hash_final(hptr, mac);

	if ((hfd >= 0) && (write(hfd, (char *)hrp, sizeof(*hrp)) < 0))
		syserr("error writing to hash file for %s: %m", path);

	hash_updatedbuf(hptr, dbuf);
}


void
hash_flush()
{
	filespec_t *fs;
	netspec_t *ns;

	for (fs = fslist; fs != NULL; fs = fs->fs_next)
		if (FILE_DOHASH(fs))
			hash_writeH0(fs->fs_hptr, fs->fs_path, fs->fs_dbuf,
				     fs->fs_cnts, sizeof(fs->fs_cnts));

	for (ns = nslist; ns != NULL; ns = ns->ns_next)
		if (ns->ns_type == DEST_TCP)
			hash_writeH0(ns->ns_hptr, ns->ns_path, ns->ns_dbuf,
				     &ns->ns_lcnt, sizeof(ns->ns_lcnt));
}


static int hash_stop;

static void
hash_timer()
{
	hash_stop = 1;
}


static int
hash_nseconds(n, hptr, buf, bufsz)
	int n;
	hashtype_t *hptr;
	char *buf;
	size_t bufsz;
{
	struct itimerval new;
	char *mac;
	int i = 0, csz;

	if (n <= 0)
		return 0;

	signal(SIGVTALRM, hash_timer);
	hash_stop = 0;
	new.it_interval.tv_sec = 0;
	new.it_interval.tv_usec = 0;
	new.it_value.tv_sec = n;
	new.it_value.tv_usec = 0;
	csz = MIN(bufsz, hptr->ht_dsize);
	mac = (char *)malloc(hptr->ht_dsize);
	if (!setitimer(ITIMER_VIRTUAL, &new, NULL)) {
		while (hash_stop == 0) {
			hash_init(hptr);
			hash_update(hptr, buf, bufsz);
			hash_final(hptr, mac);
			bcopy(mac, buf, csz);
			i++;
		}
	} else
		perror("setitimer");
	free(mac);
	return i;
}


static void
hash_updatedbuf(hptr, dbuf)
	hashtype_t *hptr;
	char *dbuf;
{
	static time_t last_H0 = 0;
	time_t now;

	if (hptr == NULL)
		return;

	hash_init(hptr);
	hash_update(hptr, dbuf, hptr->ht_dsize);
	hash_final(hptr, dbuf);

	if (last_H0 == 0) {
		hash_flush();
		last_H0 = time(NULL);
	} else {
		now = time(NULL);
		if (now - last_H0 > HASH_WRITE_INTERVAL) {
			last_H0 = now;
			hash_flush();
		}
	}
}
