#include "h.h"
#include "nsyslog.h"
#include "var.h"
#include "hash.h"
#include "crypt.h"
#include "htcp.h"

static char *htcp_secret = NULL;
static int htcp_slen = 0;

extern	fd_set	mwfd;

/*
 * This code handles TCP connections for syslogd.  This includes the state
 * machine for the initial negoiation phase and all the IO wrappers required
 * for async TCP communications.
 */


/*
 * array of valid commands for each state
 */
static int
htcs_valid[HTCS_NUM] = {
	0,
	HTC_HASH_FN,		/* HTCS_HASH_FN */
	HTC_ACK|HTC_HASH_FN,	/* HTCS_HASH_ACK */
	HTC_CRYPT_FN,		/* HTCS_CRYPT_FN */
	HTC_ACK|HTC_CRYPT_FN,	/* HTCS_CRYPT_ACK */
	HTC_NONCE,		/* HTCS_NONCE_S */
	HTC_NONCE,		/* HTCS_NONCE_D */
	HTC_HASH,		/* HTCS_HASH_S */
	HTC_HASH,		/* HTCS_HASH_D */
	HTC_SSL,
	HTC_SSL,
	0
};


/*
 * read the syslog secret file and store its contents into the global
 * buffer.
 */
static int
htcp_getsecret(ns)
	netspec_t *ns;
{
	struct stat sb;
	char *file;
	int fd;

	if (ns->ns_secret && ns->ns_slen)
		return 0;

	file = (char *)malloc(strlen(NSYSLOG_SECRET) +
			      strlen(inet_ntoa(ns->ns_sin.sin_addr)) + 2);
	strcpy(file, NSYSLOG_SECRET);
	strcat(file, ".");
	strcat(file, inet_ntoa(ns->ns_sin.sin_addr));

	fd = open(file, O_RDONLY);
	if (fd == -1) {
		syserr("can't open syslog secret file(%s):%m", file);
		free(file);
		return -1;
	}
	if (fstat(fd, &sb)) {
		syserr("can't fstat syslog secret fd(%d):%m", fd);
		free(file);
		close(fd);
		return -1;
	}

	ns->ns_slen = sb.st_size;
	if (!ns->ns_secret)
		ns->ns_secret = (char *)malloc(sb.st_size);
	if (read(fd, ns->ns_secret, sb.st_size) != sb.st_size) {
		syserr("read error: syslog secret fd(%d):%m", fd);
		free(file);
		close(fd);
		return -1;
	}
	if (opts & OPT_VERBOSE)
		fprintf(stderr, "got secret for %s: %d bytes\n",
			inet_ntoa(ns->ns_sin.sin_addr), ns->ns_slen);
	free(file);
	close(fd);
	return 0;
}


/*
 * Setup netspec_t structure for sending log messages via a TCP connection
 */
int
htcp_initdst(ns)
	netspec_t *ns;
{
	hashtype_t *hptr;

	if (htcp_getsecret(ns) != 0)
		return -1;

	hptr = ns->ns_hptr;
	if (hptr == NULL) {
		hptr = hash_get(options.op_hashtype);
		ns->ns_hptr = hptr;
	}
	if (ns->ns_hptr == NULL && options.op_hashtype != HASH_NONE)
		return -1;

	fttable[ns->ns_fd].ft_type = ns->ns_type;
	fttable[ns->ns_fd].ft_ptr = ns;
	fttable[ns->ns_fd].ft_vt = type_find(ns->ns_type);

	if (opts & OPT_VERBOSE)
		fprintf(stderr, "new connection to %s,%d\n",
			inet_ntoa(ns->ns_sin.sin_addr),
			ntohs(ns->ns_sin.sin_port));
	ns->ns_hash = (hptr == NULL) ? HASH_NONE : hptr->ht_type;
	/*
	 * as we're going to be sending the HASH_FN command first, we will then
	 * be in the HASH_ACK state as we're expecting that next
	 */
	ns->ns_state = HTCS_HASH_ACK;
	return htcp_send_cmd(ns, HTC_HASH_FN, ns->ns_hash);
}


int
htcp_sendmsg(ns, lmp, iov, niov)
	netspec_t *ns;
	logmsgtype_t *lmp;
	struct iovec *iov;
	int niov;
{
	struct sockaddr_in sin;
	hashtype_t *hptr;
	tcplmsg_t *tlm;
	int len, i, j;
	htcpcmd_t hc;
	char *s;
		
	for (len = 0, i = niov - 1; i >= 0; i--)
		len += iov[i].iov_len;

	tlm = (tcplmsg_t *)malloc(len + sizeof(*tlm));
	tlm->tl_v = 1;
	tlm->tl_zero[0] = 0;
	tlm->tl_zero[1] = 0;
	tlm->tl_zero[2] = 0;
	tlm->tl_len = htonl(len + sizeof(*tlm));
	tlm->tl_mlen = htonl(len);
	hptr = ns->ns_hptr;
	if (hptr->ht_dsize > 0)
		tlm->tl_hlen = htonl(sizeof(hashrec_t));
	else
		tlm->tl_hlen = 0;
	tlm->tl_pri = htonl(lmp->lm_pri);
	tlm->tl_tv = htonl(lmp->lm_when);

	i = sizeof(sin);
	switch (VAR_TYPE(lmp->lm_ft->ft_type))
	{
	case SRC_INET :
	case SRC_TCP :
		/*
		 * Messages sourced via UDP and TCP will already have had the
		 * lm_src struct filled in.
		 */
		tlm->tl_src = ((struct sockaddr_in *)lmp->lm_src)->sin_addr;
		break;
	default :
		/*
		 * This message had a local source, so supply the local IP
		 * addres for this message (use the local end of the connection
		 * for obtaining the IP#).
		 */
		if (getsockname(ns->ns_fd, (struct sockaddr *)&sin, &i) == 0)
			tlm->tl_src = sin.sin_addr;
		break;
	}

	s = (char *)tlm + sizeof(*tlm);
	hc.hth_cmd = HTC_MSG|HTC_VARLEN;
	hc.hth_op = len;

	for (i = 0; i < niov; i++) {
		bcopy(iov[i].iov_base, s, iov[i].iov_len);
		s += iov[i].iov_len;
	}
	i = htcp_send(ns, &hc, (char *)tlm, len + sizeof(*tlm));

	/*
	 * Now send the hash.
	 */
	if (hptr->ht_dsize > 0)
		j = htcp_send(ns, NULL, ns->ns_hmac, sizeof(hashrec_t));
	else
		j = 0;
	if (j == -1)
		i = -1;
	else
		i += j;
	free(tlm);
	return i;
}


/*
 * send a command plus data buffer (if any) to the remote connection.
 * If we can't sent it all out at once, then just queue it up for another
 * time.
 */
int
htcp_send(ns, cmdp, buf, len)
	netspec_t *ns;
	htcpcmd_t *cmdp;
	u_char *buf;
	size_t len;
{
	int error, cmd;

	if (cmdp)
		cmd = HTC_CMD(cmdp->hth_cmd);
	else
		cmd = 0;
	/*
	 * Don't even try send a message if we know the connection state
	 * hasn't yet arrived at the correct level.  We could possibly queue
	 * it up in yet another buffer list(!)...
	 */
	if ((cmdp == NULL || cmd == HTC_MSG) && (net_ready(ns) == 0)) {
		if (opts & OPT_VERBOSE)
			fprintf(stderr,
				"connection not ready, dumping msg\n");
		return 0;
	}

	if (cmdp != NULL) {
		cmdp->hth_cmd = htons(cmdp->hth_cmd);
		cmdp->hth_op = htons(cmdp->hth_op);
	}

#ifdef	HAVE_SSL_H
	if ((ns->ns_ssl != NULL) && (ns->ns_state == HTCS_MSG) &&
	    (cmd != HTC_SSL)) {
		ssl_t *ssl = ns->ns_ssl;
		error = SSL_write(ssl->ssl_ssl, (char *)cmdp, sizeof(cmdp));
		if ((error >= 0) && (buf != NULL) && (len != 0))
			error = SSL_write(ssl->ssl_ssl, buf, len);
		if (opts & OPT_VERBOSE) {
			fputs(ERR_error_string(ERR_get_error(), NULL), stderr);
			fputc('\n', stderr);
		}
	} else
#endif
		error = net_send(ns, (char *)cmdp, sizeof(*cmdp), buf, len);

	if (error == -1)
		return -1;

	if (error > 0) {
		/*
		 * We're currently in a non-sending state, so preserve that
		 * state and enter the "sending" state.
		 */
		if (ns->ns_state != HTCS_SENDING)
			ns->ns_nstate = ns->ns_state;
		ns->ns_state = HTCS_SENDING;
		FD_SET(ns->ns_fd, &mwfd);
		return 1;
	}
	FD_CLR(ns->ns_fd, &mwfd);
	return 0;
}


/*
 * wrapping function to allow commands plus operands to be easily sent
 */
int
htcp_send_cmd(ns, cmd, op)
	netspec_t *ns;
	u_int cmd;
	u_int op;
{
	htcpcmd_t hcmd;

	hcmd.hth_cmd = cmd;
	hcmd.hth_op = op;
	return htcp_send(ns, &hcmd, NULL, 0);
}


/*
 * wrapping function to allow commands plus operands with a buffer to be
 * easily sent
 */
int
htcp_send_excmd(ns, cmd, op, buf, len)
	netspec_t *ns;
	u_int cmd;
	u_int op;
{
	htcpcmd_t hcmd;

	hcmd.hth_cmd = cmd;
	hcmd.hth_op = op;
	return htcp_send(ns, &hcmd, buf, len);
}


int
htcp_ssltalk(ns, op, sz)
	netspec_t *ns;
	u_int op;
	size_t sz;
{
#ifdef	HAVE_SSL_H
	SSL *s;
#endif

	if (ns->ns_state == HTCS_SSL_C) {
#ifdef	HAVE_SSL_H
		if (ns->ns_ssl == NULL)
			net_sslinit(ns, 0);
		if (ns->ns_ssl == NULL) {
#endif
			if (htcp_send_cmd(ns, HTC_ERROR|HTC_SSL, 0) == 0)
				ns->ns_state = HTCS_MSG;
			else
				ns->ns_nstate = HTCS_MSG;
			return -1;
#ifdef	HAVE_SSL_H
		}
		s = ns->ns_ssl->ssl_ssl;
		SSL_connect(s);
		if (opts & OPT_VERBOSE) {
			fputs(ERR_error_string(ERR_get_error(), NULL), stderr);
			fputc('\n', stderr);
		}
		if (s->state == SSL_ST_OK) {
			ns->ns_state = HTCS_MSG;
			if (opts & OPT_VERBOSE) {
				fprintf(stderr, "SSL CONNECT OK\n");
				fprintf(stderr, "Ciphers: %s\n",
					SSL_get_cipher(s));
			}
		}
#endif
	} else if (ns->ns_state == HTCS_SSL_A) {
#ifdef	HAVE_SSL_H
		if (ns->ns_ssl == NULL)
			net_sslinit(ns, 1);
		if (ns->ns_ssl == NULL) {
#endif
			if (htcp_send_cmd(ns, HTC_ERROR|HTC_SSL, 0) == 0)
				ns->ns_state = HTCS_MSG;
			else
				ns->ns_nstate = HTCS_MSG;
			return -1;
#ifdef HAVE_SSL_H
		}
		s = ns->ns_ssl->ssl_ssl;
		SSL_accept(s);
		if (opts & OPT_VERBOSE) {
			fputs(ERR_error_string(ERR_get_error(), NULL), stderr);
			fputc('\n', stderr);
		}
		if (s->state == SSL_ST_OK) {
			ns->ns_state = HTCS_MSG;
			if (opts & OPT_VERBOSE) {
				fprintf(stderr, "SSL ACCEPT OK\n");
				fprintf(stderr, "Ciphers: %s\n",
					SSL_get_cipher(s));
			}
		}
#endif
	}
#ifdef HAVE_SSL_H
	if (opts & OPT_VERBOSE)
		fprintf(stderr, "ns_state %d s_state %x sz %u/%u op %u\n",
			ns->ns_state, s->state, sz,
			buf_size(ns->ns_rlist), op);
	return 0;
#endif
}


/*
 * Returns:
 * -1 - error, free up netspec_t structure
 * -2 - not enough data in buffers to read and take action
 *  0 - action taken.
 */
int
htcp_action(ns)
	netspec_t *ns;
{
	htcpcmd_t *hc, hcm;
	hashtype_t *hptr;
	char *s, *digest;
	u_short op, cmd;
	int n, err = 0;
	buflist_t *b;
	size_t sz;

	if ((ns->ns_state != HTCS_SENDING)
#ifdef	HAVE_SSL_H
	    && ((ns->ns_ssl == NULL) || (ns->ns_ssl->ssl_r == 0))
#endif
	   ) {
		b = ns->ns_rlist;
		if (b == NULL)
			return -2;
		if (b->bl_rem < sizeof(hcm))
			return -2;
		/*
		 * check for unaligned access to htcpcmd_t components.
		 */
		if (((u_long)b->bl_ptr & 0x1) != 0) {
			bcopy(b->bl_ptr, (char *)&hcm, sizeof(hcm));
			hc = &hcm;
		} else
			hc = (htcpcmd_t *)b->bl_ptr;
		sz = sizeof(*hc);
		cmd = ntohs(hc->hth_cmd);
		op = ntohs(hc->hth_op);
		if (HTC_ISVARSZ(cmd))
			sz += op;
		cmd = HTC_CMD(cmd);

		/*
		 * Do we have enough for htcpcmd_t plus its args ?
		 */
		if (sz > buf_size(b))
			return -2;		/* No. */
		/*
		 * If we don't have enough contiguous data in the
		 * first buffer, attempt to pull it all up into
		 * one buffer.
		 */
		if (sz > b->bl_rem) {
			b = buf_pullup(&ns->ns_rlist, sz);
			if (b == NULL)
				return -1;
			ns->ns_rlist = b;
			ns->ns_rtail = &b->bl_next;
			while (*ns->ns_rtail != NULL)
				ns->ns_rtail = &(*ns->ns_rtail)->bl_next;
		}

		if ((cmd & HTC_ERROR) == 0)
			if (htcs_valid[ns->ns_state] != 0 &&
			    htcs_valid[ns->ns_state] != cmd) {
				syserr("invalid state %d[%d,%x] from [%s].\n",
					ns->ns_state, htcs_valid[ns->ns_state],
					cmd, inet_ntoa(ns->ns_sin.sin_addr));
				return -1;
			}
	} else {
		sz = 0;
		op = 0;
		cmd = 0;
	}

	switch (ns->ns_state)
	{
	case HTCS_SENDING :
		/*
		 * Attempt to send what's in the first buffer out and if we
		 * succeed in sending the entire buffer, then free it up and
		 * see if we can send another one.
		 */
		while ((b = ns->ns_wlist) != NULL) {
			n = send(ns->ns_fd, b->bl_ptr, b->bl_rem, 0);
			if (n == -1) {
				if (errno == EAGAIN || errno == EWOULDBLOCK ||
				    errno == ENOBUFS || errno == EINTR)
					n = 0;
				return n;
			}
			b->bl_rem -= n;
			if (b->bl_rem == 0) {
				/*
				 * Is this buffer the last in the list of
				 * current buffers ?
				 */
				if (b->bl_next == NULL) {
					ns->ns_wtail = &ns->ns_wlist;
					ns->ns_wlist = NULL;
				} else
					ns->ns_wlist = b->bl_next;
				free(b);
			} else {
				/*
				 * We've written as much as we can, which might
				 * even be zero bytes, but not all of what we
				 * wanted to write, so make this the last time
				 * through the loop.
				 */
				b->bl_ptr += n;
				break;
			}
		}

		if (ns->ns_wlist == NULL) {
			ns->ns_state = ns->ns_nstate;
			if (ns->ns_state == HTCS_SSL_C)
				htcp_ssltalk(ns, 0, 0);
		}
		return 0;
	/*
	 * If we accepted a connection from another then we come here first
	 * as we expect a hash command first up.
	 */
	case HTCS_HASH_FN :
		hptr = hash_find(op);
		if (op == HTCH_NONE) {
			err = htcp_send_cmd(ns, HTC_ACK|HTC_HASH_FN, HTCH_NONE);
			if (err == -1)
				break;
			ns->ns_hash = HASH_NONE;
			if (opts & OPT_VERBOSE)
				fprintf(stderr, "hashing NULL for %s\n",
					inet_ntoa(ns->ns_sin.sin_addr));
		} else if (hptr == NULL) {
			if (opts & OPT_VERBOSE)
				fprintf(stderr, "unknown hash op %d for %s\n",
					op, inet_ntoa(ns->ns_sin.sin_addr));
			err = htcp_send_cmd(ns, HTC_ERROR|HTC_HASH_FN, op);
			if (err == -1)
				break;
		} else {
			err = htcp_send_cmd(ns, HTC_ACK|HTC_HASH_FN, op);
			if (err == -1)
				break;
			ns->ns_hash = op;
			ns->ns_hptr = hash_get(op);
			hptr = ns->ns_hptr;
			if (htcp_getsecret(ns) != 0) {
				err = -1;
				break;
			}
			if (hptr != NULL)
				ns->ns_dbuf = malloc(hptr->ht_dsize);
			ns->ns_hfd = net_openlog(ns, 0);
			if (ns->ns_hfd >= 0) {
				ns->ns_lfd = net_openlog(ns, 1);
				/*
				 * If one is open, the other *must* be open.
				 * If only one opens, close the other.
				 */
				if (ns->ns_lfd == -1) {
					close(ns->ns_hfd);
					ns->ns_hfd = -1;
				}
			}
			if (ns->ns_hfd == -1) {
				syserr("couldn't open host log files");
				err = -1;
				break;
			}
			if (opts & OPT_VERBOSE)
				fprintf(stderr, "hash set to %s/%d for %s\n",
					hptr->ht_name, ns->ns_hash,
					inet_ntoa(ns->ns_sin.sin_addr));
		}
		if (err == 1)
			ns->ns_nstate = HTCS_CRYPT_FN;
		else
			ns->ns_state = HTCS_CRYPT_FN;
		break;
	/*
	 * The side making the connection comes here first as it sends a hash
	 * command and is then expecting an ACK.
	 */
	case HTCS_HASH_ACK :
		if (cmd & HTC_ERROR) {
			if (opts & OPT_VERBOSE)
				fprintf(stderr, "hash error for %d from %s\n",
					op, inet_ntoa(ns->ns_sin.sin_addr));
			hash_free(ns->ns_hptr);
			ns->ns_hptr = hash_getnext(ns->ns_hash);
			if (ns->ns_hptr == NULL)
				ns->ns_hash = HASH_NONE;
			else {
				hptr = ns->ns_hptr;
				ns->ns_hash = hptr->ht_type;
			}
			if (opts & OPT_VERBOSE)
				fprintf(stderr, "next hash is %s/%d for %s\n",
					hptr->ht_name, ns->ns_hash,
					inet_ntoa(ns->ns_sin.sin_addr));
			err = htcp_send_cmd(ns, HTC_HASH_FN, ns->ns_hash);
			if (err == -1)
				break;
		} else {
			hptr = hash_find(op);
			ns->ns_hptr = hash_get(op);
			if (opts & OPT_VERBOSE)
				fprintf(stderr, "hash agreed to %s/%d w/ %s\n",
					hptr->ht_name, ns->ns_hash,
					inet_ntoa(ns->ns_sin.sin_addr));
			ns->ns_crypt = CRYPT_NONE;
			err = htcp_send_cmd(ns, HTC_CRYPT_FN, CRYPT_NONE);
			if (err == -1)
				break;

			if (err == 1)
				ns->ns_nstate = HTCS_CRYPT_ACK;
			else
				ns->ns_state = HTCS_CRYPT_ACK;
			if (opts & OPT_VERBOSE)
				fprintf(stderr, "suggest no-encrypt with %s\n",
					inet_ntoa(ns->ns_sin.sin_addr));
		}
		break;
	case HTCS_CRYPT_FN :
		if (op != CRYPT_NONE) {
			if (opts & OPT_VERBOSE)
				fprintf(stderr, "unknown crypto %d for %s\n",
					op, inet_ntoa(ns->ns_sin.sin_addr));
			err = htcp_send_cmd(ns, HTC_ERROR|HTC_CRYPT_FN, op);
			if (err == -1)
				break;
		} else {
			if (opts & OPT_VERBOSE)
				fprintf(stderr, "ack null-crypto with %s\n",
					inet_ntoa(ns->ns_sin.sin_addr));
			err = htcp_send_cmd(ns, HTC_ACK|HTC_CRYPT_FN, op);
			if (err == -1)
				break;
			ns->ns_crypt = op;

			if (ns->ns_hash != HASH_NONE) {
				ns->ns_state = HTCS_NONCE_S;
				INIT_NONCE(&ns->ns_local);
				hash_getsalt(ns->ns_nonce, ns->ns_nlen);
				err = htcp_send_excmd(ns, HTC_NONCE|HTC_VARLEN,
						      ns->ns_nlen,
						      ns->ns_nonce,
						      ns->ns_nlen);
				if (err == 1)
					ns->ns_nstate = HTCS_NONCE_S;
				else
					ns->ns_state = HTCS_NONCE_S;
			} else {
				if (err == 1)
					ns->ns_nstate = HTCS_SSL_A;
				else
					ns->ns_state = HTCS_SSL_A;
			}
		}
		break;
	case HTCS_CRYPT_ACK :
		if (cmd & HTC_ERROR) {
#ifdef notyet
/*
 * When more than one encryption algorithm has been added...
 *
			if (another algorithm) {
				select algorithm;
				send new crypto info.;
				break;
			}
 */
#endif
			if (opts & OPT_VERBOSE)
				fprintf(stderr, "revert to no crypto for %s\n",
					inet_ntoa(ns->ns_sin.sin_addr));
			/*
			 * XXX - Currently this isn't technically required, as
			 * it is all that there is, but it'll need to be here
			 * here eventually.
			 */
			ns->ns_crypt = CRYPT_NONE;
			err = htcp_send_cmd(ns, HTC_CRYPT_FN, CRYPT_NONE);
			if (err == -1)
				break;
		} else {
			if (opts & OPT_VERBOSE)
				fprintf(stderr, "crypto agreed to %d with %s\n",
					op, inet_ntoa(ns->ns_sin.sin_addr));
			ns->ns_crypt = op;
			if (ns->ns_hash != HASH_NONE) {
				ns->ns_state = HTCS_NONCE_D;
				INIT_NONCE(&ns->ns_local);
				hash_getsalt(ns->ns_nonce, ns->ns_nlen);
				err = htcp_send_excmd(ns, HTC_NONCE|HTC_VARLEN,
						      ns->ns_nlen, ns->ns_nonce,
						      ns->ns_nlen);
				ns->ns_nstate = HTCS_NONCE_D;
			} else {
				if (err == 1)
					ns->ns_nstate = HTCS_SSL_C;
				else {
					ns->ns_state = HTCS_SSL_C;
					htcp_ssltalk(ns, 0, 0);
				}
			}
		}
		break;
	case HTCS_NONCE_S :
	case HTCS_NONCE_D :
		if (ns->ns_state == HTCS_NONCE_S)
			n = HTCS_HASH_S;
		else
			n = HTCS_HASH_D;
		ns->ns_rnlen = sz - sizeof(htcpcmd_t);
		ns->ns_rnonce = (char *)malloc(ns->ns_rnlen);
		bcopy(b->bl_ptr + sizeof(htcpcmd_t), ns->ns_rnonce,
		      ns->ns_rnlen);
		hptr = ns->ns_hptr; 
		hash_init(hptr);
		hash_update(hptr, ns->ns_nonce, ns->ns_nlen);
		hash_update(hptr, ns->ns_rnonce, ns->ns_rnlen);
		hash_update(hptr, ns->ns_secret, ns->ns_slen);
		digest = hash_final(hptr, NULL);

		err = htcp_send_excmd(ns, HTC_HASH|HTC_VARLEN, hptr->ht_dsize,
				      digest, hptr->ht_dsize);
		free(digest);
		if (err == -1)
			break;
		if (err == 1)
			ns->ns_nstate = n;
		else
			ns->ns_state = n;
		break;
	case HTCS_HASH_S :
	case HTCS_HASH_D :
		hptr = ns->ns_hptr; 
		if (hptr->ht_dsize != sz - sizeof(htcpcmd_t)) {
			err = -1;
			break;
		}
		hash_init(hptr);
		hash_update(hptr, ns->ns_rnonce, ns->ns_rnlen);
		hash_update(hptr, ns->ns_nonce, ns->ns_nlen);
		hash_update(hptr, ns->ns_secret, ns->ns_slen);
		digest = hash_final(hptr, NULL);

		if (bcmp(digest, b->bl_ptr + sizeof(htcpcmd_t),
			 hptr->ht_dsize) != 0) {
			syserr("error matching hash from %s, closing",
				inet_ntoa(ns->ns_sin.sin_addr));
			err = -1;
		} else {
			if (opts & OPT_VERBOSE)
				fprintf(stderr, "hashes match\n");
			free(ns->ns_nonce);
			ns->ns_nonce = NULL;
			free(ns->ns_rnonce);
			ns->ns_rnonce = NULL;
			if (ns->ns_state == HTCS_HASH_S) {
				ns->ns_state = HTCS_SSL_A;
			} else if (ns->ns_state == HTCS_HASH_D) {
				buf_discard(&ns->ns_rlist, sz);
				ns->ns_state = HTCS_SSL_C;
				if (ns->ns_rlist == NULL)
					ns->ns_rtail = &ns->ns_rlist;
				sz = 0;
				htcp_ssltalk(ns, 0, 0);
			}
		}
		free(digest);
		break;
	case HTCS_SSL_C :
	case HTCS_SSL_A :
		if (cmd & HTC_ERROR) {
			if (opts & OPT_VERBOSE)
				fprintf(stderr, "ssl disabled\n");
			net_sslfree(ns);
			ns->ns_state = HTCS_MSG;
		} else if (options.op_ssl == 0) {
			if (opts & OPT_VERBOSE)
				fprintf(stderr, "ssl disabled\n");
			err = htcp_send_cmd(ns, HTC_ERROR|HTC_SSL, 0);
			ns->ns_state = HTCS_MSG;
		} else {
			if (htcp_ssltalk(ns, (u_int)op, sz) == 0)
				sz = 0;
			else if (opts & OPT_DEBUG)
				fprintf(stderr, "left from ssltalk sz=%d\n",
					sz);
		}
		break;
	case HTCS_MSG :
		/*
		 * Don't want to eat up message header here, but in
		 * net_read().
		 */
#ifdef	HAVE_SSL_H
		if (ns->ns_ssl)
			ns->ns_ssl->ssl_r = op;
#endif
		if (opts & OPT_DEBUG)
			fprintf(stderr, "HTCS_MSG: sz=%d/%d\n", sz, op);
		sz = 0;
		break;
	}

	if (sz > 0) {
		buf_discard(&ns->ns_rlist, sz);
		if (ns->ns_rlist == NULL)
			ns->ns_rtail = &ns->ns_rlist;
	}
	if (err != -1)
		return 0;
	return -1;
}
