%{
/*
 * Copyright (C) 1997 by Darren Reed.
 *
 * $Id: lex.l,v 1.36 1999/05/10 18:20:20 darrenr Exp $
 */
#include "h.h"
#include "var.h"
#include "y.tab.h"


#if !defined(lint)
static const char rcsid[] = "$Id: lex.l,v 1.36 1999/05/10 18:20:20 darrenr Exp $";
#endif

#ifndef	__P
# ifdef __STDC__
#  define	__P(x)	x
# else
#  define	__P(x)	()
# endif
#endif

extern int opts, yydebug;

int lineNum, next, state, token;

int	next_state __P((int, int));
int	next_item __P((int));
int	save_token __P((int));
int	yylex __P((void));

struct	wordtab	{
	char	*word;
	int	state;
	int	next;
};

struct	wordtab	words[] = {
	{ "source",	IL_SOURCE,		IL_SRCNAME },
	{ "file",	IL_FILE,		IL_TOKEN },

	{ "filter",	IL_FILTER,		IL_FILNAME },
	{ "facility",	IL_FACILITY,		IL_TOKEN },
	{ "priority",	IL_PRIORITY,		IL_TOKEN },
	{ "facpri",	IL_FACPRI,		IL_TOKEN },
	{ "netmask",	IL_NETMASK,		IL_TOKEN },
	{ "match",	IL_MATCH,		IL_TOKEN },
	{ "match-file",	IL_MATCHFILE,		IL_TOKEN },
	{ "match-prog",	IL_MATCHPROG,		IL_TOKEN },

	{ "destination",IL_DESTINATION,		IL_DESTNAME },
	{ "user",	IL_USER,		IL_TOKEN },
	{ "sync",	IL_SYNC,		IL_TOKEN },
	{ "sync-pri",	IL_SYNCPRIORITY,	IL_TOKEN },
	{ "sync-time",	IL_SYNCTIME,		IL_TOKEN },
	{ "raw",	IL_RAW,			-1 },
	{ "command",	IL_COMMAND,		-1 },

	{ "internal",	IL_INTERNAL,		-1 },
	{ "options",	IL_OPTIONS,		-1 },
	{ "mark",	IL_MARK,		IL_TOKEN },
	{ "hash",	IL_HASH,		-1 },
	{ "logdir",	IL_LOGDIR,		IL_TOKEN },
	{ "dir",	IL_DIRECTORY,		IL_TOKEN },
	{ "directory",	IL_DIRECTORY,		IL_TOKEN },
	{ "algorithm",	IL_HASHALGO,		IL_TOKEN },
	{ "show-repeated", IL_REPEATED,		IL_TOKEN },
	{ "package-confs", IL_PACKAGES,		IL_TOKEN },
	{ "show-facpri", IL_SHOWFACPRI,		IL_TOKEN },
	{ "ssl",	IL_SSL,			-1 },
	{ "certificate",IL_CERTIFICATE,		IL_TOKEN },
	{ "cipher",	IL_CIPHER,		IL_TOKEN },

	{ "not",	IL_NOT,			-1 },
	{ "!",		IL_NOT,			-1 },
	{ "log",	IL_LOG,			-1 },
	{ "=",		IL_EQ,			-1 },
	{ "!=",		IL_NEQ,			-1 },
	{ "<",		IL_LT,			-1 },
	{ ">",		IL_GT,			-1 },
	{ ">=",		IL_GE,			-1 },
	{ "<=",		IL_LE,			-1 },
	{ "\377",	0,			0 },
	{ NULL, 0, 0 }
};
%}
white	[ \t\r]+
%%
{white}	;
\n	{ lineNum++; }
\{	{ return next_state('{', -1); }
\}	{ return next_state('}', -1); }
\|	{ return next_state('|', -1); }
\;	{ return next_state(';', -1); }
\,	{ return next_state(',', -1); }
#[^\n]*	;
[^ \{\}\n\t\;|,]*	{ return next_item(IL_TOKEN); }
\"[^\"]*\"	{ return next_item(IL_TOKEN); }
%%
int
save_token(type)
	int type;
{
	int	i;

	if (yydebug)
		fprintf(stderr, "save_token(%d) [%s]\n", type, yytext);
	yylval.str = strdup(yytext);
	if ((type == IL_SRCNAME) || (type == IL_FILNAME) ||
	    (type == IL_DESTNAME)) {
		if ((i = new_var(yylval.str, type)) != 0)
			return i;
	}
	return type;
}


int
next_item(nstate)
	int nstate;
{
	struct	wordtab	*wt;
	var_t	*v;
	int	i;

	if (yydebug)
		fprintf(stderr, "text=[%s] id=%d next=%d\n",
			yytext, nstate, next);

	if (next == IL_TOKEN) {
		next = -1;
		token++;
		return save_token(IL_TOKEN);
	}

	if (!isdigit(*yytext)) {
		for (wt = words; wt->word; wt++)
			if (!strcasecmp(wt->word, yytext)) {
				token++;
				return next_state(wt->state, wt->next);
			}
	}

	if (next == IL_SRCNAME || next == IL_FILNAME || next == IL_DESTNAME) {
		nstate = next;
		next = -1;
		i = save_token(nstate);
		if (i == IL_SRCNAME || i == IL_FILNAME || i == IL_DESTNAME)
			token++;
		return i;
	}

	if ((nstate == IL_TOKEN) && (v = get_var(yytext))) {
		yylval.str = yytext;
		token++;
		return v->v_type;
	}

	if (yydebug)
		fprintf(stderr,
			"unknown keyword=[%s] => converting to IL_TOKEN\n",
			yytext);
	next = -1;
	if (token > 0)
		token++;
	return save_token(IL_TOKEN);
}


int
next_state(nstate, fornext)
	int nstate, fornext;
{
	if (yydebug)
		fprintf(stderr, "next_state(nstate=%d,fornext=%d), next=%d\n",
			nstate, fornext, next);
	if (fornext != -1)
		next = fornext;
	return nstate;
}


void
init_lex()
{
	lineNum = 0;	
	next = -1;	
	state = 0;	
	token = 0;
}
