#include "h.h"
#include "nsyslog.h"
#include "var.h"
#include "hash.h"
#include "y.tab.h"


#if !defined(lint)
static const char rcsid[] = "$Id: log.c,v 1.51 1999/04/22 16:53:23 darrenr Exp $";
#endif

static	void	write_mark	__P((filetype_t *));
static	int	do_msg		__P((filetype_t *, logmsgtype_t *));
static	void	write_msg	__P((filetype_t *, logmsgtype_t *, \
				     struct iovec *, int));
static	int	chkfilter	__P((logmsgtype_t *, var_t *));
static	void	add_date	__P((logmsgtype_t *, time_t *));
static	void	add_host	__P((logmsgtype_t *));
static	void	distrib_msg	__P((var_t *, logmsgtype_t *));
static	int	prep_msg	__P((logmsgtype_t *, struct iovec *, int));

static	int	log_id = 0;

extern	filespec_t	*internal;


logmsg_t *
new_logmsg(msg, mlen)
	char *msg;
	int mlen;
{
	logmsg_t *lg;

	lg = (logmsg_t *)calloc(1, sizeof(logmsg_t));
	lg->lg_ref = 1;
	lg->lg_id = log_id++;
	if (mlen == 0)
		lg->lg_line = strdup(msg);
	else {
		lg->lg_line = (char *)malloc(mlen + 1);
		strcpy(lg->lg_line, msg);
	}
	return lg;
}


void
free_logmsg(lgp)
	logmsg_t **lgp;
{
	logmsg_t *lg = *lgp;

	lg->lg_ref--;
	assert(lg->lg_ref >= 0);
	if (!lg->lg_ref) {
		free(lg->lg_host);
		free(lg->lg_line);
		if (lg->lg_prog)
			free(lg->lg_prog);
		free(lg);
	}
	*lgp = NULL;
}


void
checksync()
{
	filespec_t *fs;
	filetype_t *ft;
	sdlist_t *sd;
	time_t now;
	var_t *d;
	int i;

	if (opts & OPT_VERBOSE) {
		now = time(NULL);
		fprintf(stderr, "checksync at %s", ctime(&now));
	}

	for (sd = sdlist; sd; sd = sd->sd_next) {
		d = sd->sd_dst;

		for (i = 0; i < d->v_num; i++) {
			ft = d->v_list + i;

			if (ft->ft_type != DEST_FILE)
				continue;
			fs = ft->ft_ptr;
			if ((fs->fs_fd >= 0) && (fs->fs_lastsync < now)) {
				fs->fs_lastsync = now;
				fsync(fs->fs_fd);
			}
		}
	}
}


static void
sendmark()
{
	logmsgtype_t lm;
	filespec_t *f;

	if ((internal != NULL) && (internal->fs_cnt > 0)) {
		bzero((char *)&lm, sizeof(lm));
		if ((f = file_find(SRC_FILE, "/dev/null")))
			lm.lm_ft = fttable + f->fs_fd;
		else
			return;
		lm.lm_log = new_logmsg("-- MARK --", 0);
		lm.lm_when = time(NULL);
		lm.lm_flags = FLG_SYS|FLG_MARK|FLG_HAVEPRI;
		lm.lm_len = strlen(lm.lm_line);
		lm.lm_pri = LOG_INFO|LOG_MARK;
		logmsg(&lm);
	}
}


void
checkmarks()
{
	filetype_t *ft;
	sdlist_t *sd;
	time_t now;
	var_t *d;
	int i;

	if (opts & OPT_VERBOSE) {
		now = time(NULL);
		fprintf(stderr, "checkmarks at %s", ctime(&now));
	}
	sendmark();

	for (sd = sdlist; sd; sd = sd->sd_next) {
		d = sd->sd_dst;

		for (i = 0; i < d->v_num; i++) {
			ft = d->v_list + i;

			if (ft->ft_mcount > 1) {
				write_mark(ft);
				ft->ft_mcount = 0;
			}
		}
	}
}


static void
write_mark(ft)
	filetype_t *ft;
{
	struct iovec iov[MAX_NIOV];
	logmsgtype_t lm;
	char msg[128];
	logmsg_t log;
	time_t now;
	int nio;

	now = time(NULL);
	bzero((char *)&lm, sizeof(lm));
	bcopy((char *)ft->ft_log, (char *)&log, sizeof(log));
	lm.lm_log = &log;

	snprintf(msg, sizeof(msg), "last message repeated %d times, from ",
		 ft->ft_mcount);
	if (sizeof(msg) - strlen(msg) > 53) {
		strncat(msg, ctime(&ft->ft_firsttime), 24);
		strcat(msg, " to ");
		strncat(msg, ctime(&ft->ft_lasttime), 24);
		strcat(msg, "\n");
	}

	log.lg_ref = 1;
	lm.lm_when = ft->ft_lasttime;
	lm.lm_line = strdup(msg);
	lm.lm_len = strlen(msg);
	lm.lm_pri = ft->ft_pri;
	add_host(&lm);
	add_date(&lm, &now);
	nio = prep_msg(&lm, iov, 6);
	write_msg(ft, &lm, iov, nio);
	free(lm.lm_line);
	free(lm.lm_host);
	free(lm.lm_time);
}


static int
do_msg(ft, lmp)
	filetype_t *ft;
	logmsgtype_t *lmp;
{
	if (ft->ft_log && (ft->ft_logid == lmp->lm_logid))
		return 2;	/* already sent this message here! */
	if (ft->ft_log && ft->ft_line && options.op_shrepeat) {
		if ((ft->ft_pri != lmp->lm_pri) ||
		    (ft->ft_len != lmp->lm_len) ||
		    ((0 == ft->ft_prog) != (0 == lmp->lm_prog)) ||
		    (ft->ft_prog && lmp->lm_prog &&
		     strcmp(ft->ft_prog, lmp->lm_prog)) ||
		    strcmp(ft->ft_line, lmp->lm_line) ||
		    strcmp(ft->ft_host, lmp->lm_host)) {
			if (ft->ft_mcount > 1) {
				write_mark(ft);
				ft->ft_mcount = 0;
			}
			free_logmsg(&ft->ft_log);
		} else {
			if (!ft->ft_firsttime)
				ft->ft_firsttime = ft->ft_lasttime;
			ft->ft_mcount++;
			ft->ft_lasttime = time(NULL);
			return 1;
		}
	}
	ft->ft_lasttime = time(NULL);
	ft->ft_log = lmp->lm_log;
	ft->ft_log->lg_ref++;
	ft->ft_pri = lmp->lm_pri;
	ft->ft_len = lmp->lm_len;
	ft->ft_mcount = 1;
	return 0;
}


static int
prep_msg(lmp, iov, maxiov)
	logmsgtype_t *lmp;
	struct iovec *iov;
	int maxiov;
{
	char pri[18];
	int nio = 0;
	table_t *t;

	if (maxiov == 0)
		return 0;

	/*
	 * build up an iovec, ordered correctly for a log message.
	 */
	iov[nio].iov_base = lmp->lm_time;
	iov[nio].iov_len = strlen(lmp->lm_time);
	nio++;
	if (nio == maxiov)
		return nio;
	iov[nio].iov_base = " ";
	iov[nio].iov_len = 1;
	nio++;
	if (nio == maxiov)
		return nio;
	if (options.op_shfacpri) {
		if ((t = pri_find(LOG_PRI(lmp->lm_pri))))
			pri[0] = t->letter;
		else
			pri[0] = '?';

		if ((t = fac_find((lmp->lm_pri & LOG_FACMASK) >> 3)))
			pri[1] = t->letter;
		else
			pri[1] = '?';
		pri[2] = ':';
		pri[3] = '\0';
		iov[nio].iov_base = pri;
		iov[nio].iov_len = 3;
		nio++;
		if (nio == maxiov)
			return nio;
	}
	if ((lmp->lm_flags & FLG_HTCP) == 0) {
		iov[nio].iov_base = lmp->lm_host;
		iov[nio].iov_len = strlen(lmp->lm_host);
		nio++;
	}
	if (nio == maxiov)
		return nio;
	if ((lmp->lm_flags & FLG_HTCP) == 0) {
		iov[nio].iov_base = " ";
		iov[nio].iov_len = 1;
		nio++;
	}
	if (nio == maxiov)
		return nio;
	iov[nio].iov_base = lmp->lm_line;
	iov[nio].iov_len = lmp->lm_len;
	nio++;
	return nio;
}


static void
distrib_msg(var, lmp)
	var_t *var;
	logmsgtype_t *lmp;
{
	struct iovec iov[MAX_NIOV];
	filetype_t *ft;
	int i, nio;

	nio = prep_msg(lmp, iov, 6);

	for (i = var->v_num - 1; i >= 0; i--) {
		ft = var->v_list + i;
		switch (do_msg(ft, lmp))
		{
		case 1 :
			if (opts & OPT_VERBOSE)
				fprintf(stderr, "duplicate message detected - \
suppressing logging\n");
			break;
		case 0 :
			write_msg(ft, lmp, iov, nio);
			break;
		case 2 :
			if (opts & OPT_VERBOSE)
				fprintf(stderr, "duplicate destination\n");
		default :
			break;
		}
	}
}


/*
 * write_msg - actually output the message to a destination.
 */
static void
write_msg(ft, lmp, iov, nio)
	filetype_t *ft;
	logmsgtype_t *lmp;
	struct iovec *iov;
	int nio;
{
	void *p = ft->ft_ptr;
	int nw = 0, fd;

	if (opts & OPT_VERBOSE)
		fprintf(stderr, "%s.%s ", fac_toname(lmp->lm_pri),
			pri_toname(lmp->lm_pri));

	nw = (* ft->ft_vt->vt_logmsg)(p, &fd, iov, nio, lmp);

	/*
	 * If there was an error opening the file, reopen it.
	 */
	if (nw < 0 && fd >= 0)
		ft->ft_closed = time(NULL);

	if ((nw == 0 && fd == -1) &&
	    (time(NULL) - ft->ft_closed > NSYSLOG_REOPEN))
		(* ft->ft_vt->vt_reopen)(p);
}


/*
 * chkfilter - check if a message matches a given list of filters.
 *	     return -1 if it doesn't match
 *	     return 0 if it does match
 */
static int
chkfilter(lmp, v)
	logmsgtype_t *lmp;
	var_t *v;
{
	filetype_t *ft;
	int i, ok = 0;

	for (i = 0, ft = v->v_list; i < v->v_num; i++, ft++) {
		ok = (* ft->ft_vt->vt_match)(ft->ft_ptr, lmp);
		if (ok == 0)
			return 0;
	}
	return 1;
}


/*
 * scan the list of log rules which join sources, filters and destinations
 * for rules which match this source.
 */
void
writelogmsg(lmp)
	logmsgtype_t *lmp;
{
	fillist_t *fa, *fo;
	sdlist_t *s;
	var_t *d;
	int i, j;

	for (s = sdlist; s; s = s->sd_next)
		if (var_ismember(s->sd_src, lmp->lm_ft) != 0) {
			if (s->sd_fil) {
				/*
				 * Check filters which are to be applied in
				 * sequence and check each chain of these.
				 */
				for (fo = s->sd_fil; fo; fo = fo->fl_or) {
					for (fa = fo; fa; fa = fa->fl_and) {
						j = chkfilter(lmp, fa->fl_fil);
						if ((j && !fa->fl_comp) ||
						    (!j && fa->fl_comp))
							break;
					}
					if ((!fa && !fo->fl_comp) ||
					    (fa && fo->fl_comp))
						break;
				}
				if (!fo)
					continue;
			}

			/*
			 * passed the filtering, so write it out to all the
			 * various destinations.
			 */
			if (opts & OPT_VERBOSE)
				fprintf(stderr, "j=%d: %s | %s -> %s\n",
					j, s->sd_src->v_name,
					(s->sd_fil == NULL) ? "" :
					s->sd_fil->fl_fil->v_name,
					s->sd_dst->v_name);
			distrib_msg(s->sd_dst, lmp);
		}
}


/*
 * add the hostname of the source into the message.  This goes between the
 * date-time stamp and the rest of the message so it may need proper
 * insertion.
 */
static void
add_host(lmp)
	logmsgtype_t *lmp;
{
	struct sockaddr_in *sin;
	struct hostent *hp;
	sockaddr_t *sk;
	char *hname;

	/*
	 * If the source is 'inet, then do the usual lookups of IP address.
	 */
	sk = (sockaddr_t *)lmp->lm_src;
	if (sk && sk->sa_family == AF_INET) {
		sin = (struct sockaddr_in *)sk;
		hp = gethostbyaddr((char *)&sin->sin_addr.s_addr,
				   sizeof(struct in_addr), sin->sin_family);
		if (!hp)
			hname = inet_ntoa(sin->sin_addr);
		else
			hname = hp->h_name;

	} else
		hname = get_shorthostname();

	lmp->lm_host = strdup(hname);
}


/*
 * If a log message doesn't have a reocgnisable time-date stamp at the front,
 * insert one.
 */
static void
add_date(lmp, timep)
	logmsgtype_t *lmp;
	time_t *timep;
{
	struct tm *tm;
	char *newline;
	char ts[17];

	if (timep == NULL)
		timep = &lmp->lm_when;

	if (!(tm = localtime(timep))) {
		snprintf(ts, sizeof(ts), "%u", (u_int) lmp->lm_when);
	} else {
		strftime(ts, sizeof(ts), "%h %e %H:%M:%S", tm);
		ts[sizeof(ts) - 1] = '\0';
	}

	if (!(newline = (char *)malloc(strlen(ts) + 1)))
		return;
	strcpy(newline, ts);
	lmp->lm_time = newline;
	return;
}


void
logmsg(lmp)
	logmsgtype_t *lmp;
{
	char c, *s, *t, *line, *oline;
	int omask, len;
	u_int pri = 0;
	sigset_t sigset, oldset;

	if (!(line = (char *)malloc(MAXLINE + 2)))
		return;

	s = lmp->lm_line;
	oline = s;
	len = lmp->lm_len;
	if (lmp->lm_flags & FLG_HAVEPRI)
		pri = lmp->lm_pri;
	else {
		if (*s == '<' && isdigit(*(s+1))) {
			s++;
			len--;
			for (; len && isdigit(*s); s++, len--)
				pri = 10 * pri + (*s - '0');
			if (len && *s == '>')
				s++, len--;
			if (pri >= (LOG_NFACILITIES << 3))
				pri = LOG_USER|LOG_NOTICE;
		} else
			pri = LOG_USER|LOG_NOTICE;
	}

	if (!(lmp->lm_flags & FLG_SYS) && (pri & LOG_PRIMASK) == LOG_KERN) {
		pri &= ~LOG_KERN;
		pri |= LOG_USER;
	}

	for (t = line; len && (c = *s++ & 0x7f) != '\0' && c != '\n' &&
	     t - line < MAXLINE; len--) {
		if (iscntrl(c)) {
			*t++ = '^';
			*t++ = c | 0x40;
		} else
			*t++ = c;
	}
	*t++ = '\n';
	*t = '\0';
	free(oline);
	lmp->lm_line = line;
	lmp->lm_len = strlen(line);
	lmp->lm_pri = pri;


	if (opts & OPT_VERBOSE)
		printf("logmsg: pri %o=%s.%s, flags %x, from %p=%d, msg [%s]\n",
			lmp->lm_pri, fac_toname(lmp->lm_pri),
			pri_toname(lmp->lm_pri), lmp->lm_flags, lmp->lm_src,
			lmp->lm_slen, line);

	sigemptyset(&sigset);
	sigaddset(&sigset, SIGHUP);
	sigaddset(&sigset, SIGALRM);
#if 0
	omask = sigblock(sigmask(SIGHUP)|sigmask(SIGALRM));
#endif
	sigprocmask(SIG_BLOCK, &sigset, &oldset);

	if (lmp->lm_len < 16 || line[3] != ' ' || line[6] != ' ' ||
	    line[9] != ':' || line[12] != ':' || line[15] != ' ')
		lmp->lm_flags |= FLG_ADDDATE;

	/*
	 * Hostname is already in messages from TCP connections
	 */
	if ((lmp->lm_flags & FLG_HTCP) == 0)
		add_host(lmp);
	if (lmp->lm_flags & FLG_ADDDATE)
		add_date(lmp, NULL);
	else {
		lmp->lm_time = (char *)malloc(16);
		strncpy(lmp->lm_time, line, 16);
		lmp->lm_time[15] = '\0';
		t = strdup(line + 16);
		free(line);
		line = NULL;
		lmp->lm_line = t;
		lmp->lm_len = strlen(t);
	}

	for (t = lmp->lm_line; *t && *t != ':' && *t != '['; t++)
		;
	if (*t) {
		s = strchr(lmp->lm_line, ' ');
		/*
		 * TCP messages already have the hostname in them, get it out.
		 */
		if ((lmp->lm_flags & FLG_HTCP) && (s < t)) {
			lmp->lm_host = (char *)malloc(s - lmp->lm_line + 1);
			strncpy(lmp->lm_host, lmp->lm_line, s - lmp->lm_line);
			lmp->lm_host[s - lmp->lm_line] = '\0';
			s++;
		} else
			s = lmp->lm_line;
		lmp->lm_prog = (char *)malloc(t - s + 1);
		strncpy(lmp->lm_prog, s, t - s + 1);
		lmp->lm_prog[t - s] = '\0';
	} else
		lmp->lm_prog = NULL;
	writelogmsg(lmp);
#if 0
	sigsetmask(omask);
#endif
	sigprocmask(SIG_SETMASK, &oldset, NULL);

	free_logmsg(&lmp->lm_log);
	free(lmp->lm_time);
	lmp->lm_time = NULL;
	return;
}
