#include "h.h"
#include "nsyslog.h"
#include "var.h"
#include "hash.h"


#if !defined(lint)
static const char rcsid[] = "$Id: main.c,v 1.56 1999/05/10 18:21:05 darrenr Exp $";
#endif

extern	char	*optarg;
extern	int	errno;
#ifdef	HAVE_SSL_H
BIO	*bio_err = NULL;
LHASH	*config = NULL;
#endif

fd_set	mrfd, mwfd;				/* master fd_set flags */
int	do_reopen = 0,
#ifdef	OpenBSD
	opts = OPT_OLDNOUDP,
#else
	opts = 0,
#endif
	dofork = 1,
	started = 0;
char	*progname = NULL;
char	*cfgfile = NULL,
	*pidfile = NSYSLOG_PID;
int	allow_severity = LOG_INFO;	/* for libwrap */
int	deny_severity = LOG_WARNING;	/* for libwrap */
char	*cfgfiles[] = { NSYSLOG_CONF, SYSLOG_CONF, NULL };
filespec_t	*internal = NULL;


static	void	fakelog		__P((char *));
static	void	startup_msg	__P((void));
static	void	hup_msg		__P((void));
static	void	end_msg		__P((int));
static	void	handle_hup	__P((int));
static	void	handle_term	__P((int));
static	void	handle_int	__P((int));
static	int	read_fd		__P((int, int));
static	int	check_sources	__P((int, time_t));

#if BSD < 199306
# ifdef	__STDC__
extern	const char	*const sys_errlist[];
# else
extern	char	*sys_errlist[];
# endif
#endif

void
#ifdef __STDC__
syserr(char *fmt, ...)
#else
syserr(fmt, va_alist)
char	*fmt;
va_dcl
#endif
{
	char buf[MAXLINE], *s, *t = NULL;
	const char *errm;
	va_list pvar;
 
	errm = sys_errlist[errno];
	s = strstr(fmt, "%m");
	if (s != NULL) {
		t = (char *)malloc(strlen(fmt) + strlen(errm) + 3);
		strncpy(t, fmt, s - fmt);
		t[s - fmt] = '\0';
		strcat(t, errm);
		strcat(t, s+2);
		fmt = t;
	}
#ifdef __STDC__
	va_start(pvar, fmt);
#else
	va_start(pvar);
#endif
	if (started != 0) {
		vsnprintf(buf, sizeof(buf), fmt, pvar);
		logerror(buf);
	} else {
		vfprintf(stderr, fmt, pvar);
		fputc('\n', stderr);
	}
	va_end(pvar);
	if (t)
		free(t);
}


void
init_internal()
{
	static filetype_t ft;
	static int called = 0;
	filespec_t *f;

	if ((opts & OPT_OLDSYSLOG) != 0) {
		new_var("syslog", VAR_SRC);
		new_obj("syslog");
	}
	f = file_new(SRC_FILE, "/dev/null");
	if (f != NULL) {
		internal = f;
		f->fs_mode |= VAR_INTERNAL;
		if (opts & OPT_OLDSYSLOG)
			(void) add_obj(SRC_FILE, f);
		if (f->fs_fd != 0) {
			close(0);
			dup2(f->fs_fd, 0);
			if (called == 1)
				bcopy(&ft, fttable + f->fs_fd,
				      sizeof(*fttable));
			else
				bcopy(fttable + f->fs_fd, fttable,
				      sizeof(*fttable));
			close(f->fs_fd);
			bzero(fttable + f->fs_fd, sizeof(*fttable));
			f->fs_fd = 0;
			called = 1;
			bcopy(fttable, &ft, sizeof(ft));
		}
	}
}


static void
fakelog(line)
	char *line;
{
	logmsgtype_t lm;
	filespec_t *f;

	if ((internal != NULL) && (internal->fs_cnt > 0)) {
		bzero((char *)&lm, sizeof(lm));
		if ((f = file_find(SRC_FILE, "/dev/null")))
			lm.lm_ft = fttable + f->fs_fd;
		else
			return;
		lm.lm_log = new_logmsg(line, 0);
		lm.lm_when = time(NULL);
		lm.lm_flags = FLG_ALL;
		lm.lm_len = strlen(lm.lm_line);
		logmsg(&lm);
	}
}


static void
handle_alrm(sig)
	int sig;
{
	return;
}


static void
handle_hup(sig)
	int sig;
{
	do_reopen = 1;
}


static void
end_msg(sig)
	int sig;
{
	char buf[MAXLINE];

	snprintf(buf, sizeof(buf), "<%d>%s[%d]: going down on signal %d\n",
		 LOG_SYSLOG|LOG_NOTICE, progname, getpid(), sig);
	fakelog(buf);
	close_all();
}


static void
handle_int(sig)
	int sig;
{
	do_reopen = 2;
}


static void
exit_int()
{
	end_msg(SIGINT);
	if (options.op_hashtype != HASH_NONE)
		hash_flush();
	expire_all();
	expire_old();
	cleanup_config();
	exit(SIGINT);
}


static void
handle_term(sig)
	int sig;
{
	do_reopen = 3;
}


static void
exit_term()
{
	end_msg(SIGTERM);
	if (options.op_hashtype != HASH_NONE)
		hash_flush();
	expire_all();
	expire_old();
	cleanup_config();
	exit(SIGTERM);
}


static void
startup_msg()
{
	char buf[MAXLINE];

	snprintf(buf, sizeof(buf), "<%d>%s[%d]: started\n",
		 LOG_SYSLOG|LOG_INFO, progname, getpid());
	fakelog(buf);
}


static void
hup_msg()
{
	char buf[MAXLINE];

	snprintf(buf, sizeof(buf), "<%d>%s[%d]: received HUP - reloading %s\n",
		 LOG_SYSLOG|LOG_INFO, progname, getpid(), cfgfile);
	fakelog(buf);
}


void
logerror(msg)
	char *msg;
{
	char *buf;

	if (opts & OPT_VERBOSE)
		fprintf(stderr, "error: <%d>%s\n", LOG_SYSLOG|LOG_ERR, msg);
	buf = (char *)malloc(strlen(msg) + 10);
	snprintf(buf, strlen(msg) + 10, "<%d>%s", LOG_SYSLOG|LOG_ERR, msg);
	fakelog(buf);
	free(buf);
}


void
async_fd(fd, onoff)
	int fd, onoff;
{
	int res;

	res = fcntl(fd, F_GETFL, 0);
	if (res == -1) {
		syserr("fcntl(%d, F_GETFL) failed: %m", fd);
		return;
	}

	if (onoff == 1) {
#if defined(O_NONBLOCK) || defined(O_NDELAY)
# ifdef O_NONBLOCK
		res |= O_NONBLOCK;
# else
#  ifdef        O_NDELAY
		res |= O_NDELAY;
#  endif
# endif
	} else {
# ifdef O_NONBLOCK
		res &= ~O_NONBLOCK;
# else
#  ifdef        O_NDELAY
		res &= ~O_NDELAY;
#  endif
# endif
#endif
	}

	if (fcntl(fd, F_SETFL, res) == -1)
		syserr("fcntl(%d, F_SETFL, %#x) failed: %m", fd, res);
	return;
}


void
ft_set(fd, ptr, type)
	int fd;
	void *ptr;
	int type;
{
	fttable[fd].ft_var = NULL;	/* set in setup_filetable */
	fttable[fd].ft_ptr = ptr;
	fttable[fd].ft_type = type;
	if (type == SRC_TCP)
		fttable[fd].ft_vt = type_find(SRC_INET_LISTEN);
	else
		fttable[fd].ft_vt = type_find(type);
	switch (VAR_TYPE(type))
	{
	case SRC_INET_LISTEN :
	case SRC_UNIX_LISTEN :
	case SRC_STREAM :
	case SRC_UNIX :
	case SRC_INET :
	case SRC_TCP :
		FD_SET(fd, &mrfd);
		break;
	default :
		break;
	}
}


static void
write_fd(fd)
	int fd;
{
	filetype_t *ft = fttable + fd;

	/*
	 * we can have blocked io to recover on network connections, pipes,
	 * tty's (unplugged serial ports with flow control), etc.
	 */
	(*ft->ft_vt->vt_drain)(ft->ft_ptr);
}


static int
read_fd(fd, flag)
	int fd, flag;
{
	filetype_t *ft = fttable + fd;
	union un_sockaddr un_sock;
	int len = 0, nfd, ret;
	int (*rf) __P((void *, logmsgtype_t *, tcplmsg_t *, char *, size_t));
	char buf[MAXLINE];
	logmsgtype_t lm;
	vartype_t *vt;
	tcplmsg_t tlm;

	if (ft->ft_ptr == NULL)
		return 0;
	vt = ft->ft_vt;
	rf = (flag) ? vt->vt_readnext : vt->vt_read;
	if ((flag != 0) && (VAR_TYPE(ft->ft_type) != SRC_TCP) &&
	    (VAR_TYPE(ft->ft_type) != DEST_TCP) || !rf)
		return 0;
	bzero((char *)&lm, sizeof(lm));
	lm.lm_slen = sizeof(un_sock);
	lm.lm_src = &un_sock;
	lm.lm_ft = ft;
	ret = 0;

	if ((ft->ft_type & VAR_LISTEN) != 0) {
		if (flag != 0)
			return 0;
		nfd = (*vt->vt_accept)(fd);
		if (nfd != -1) {
			FD_SET(nfd, &mrfd);
			ft = fttable + nfd;
			ft->ft_parent = fttable[fd].ft_var;
			ft->ft_mcount = 0;
			(*vt->vt_link)(fttable[fd].ft_ptr, ft->ft_ptr);
		}
		return 0;
	} else if (VAR_MODE(ft->ft_type) == SRC_NONE) {
		len = (*rf)(ft->ft_ptr, &lm, &tlm, buf, sizeof(buf));
		if ((len >= 0) && (vt->vt_more != NULL) &&
		    (*vt->vt_more)(ft->ft_ptr))
			ret = 1;
		if (len <= 0)
			return ret;
	} else if (VAR_MODE(ft->ft_type) == DEST_NONE) {
		len = (*rf)(ft->ft_ptr, &lm, &tlm, buf, sizeof(buf));
		if ((len >= 0) && (vt->vt_more != NULL) &&
		    (*vt->vt_more)(ft->ft_ptr))
			ret = 1;
		if (len == -1)
			FD_CLR(fd, &mrfd);
		return ret;
	}
	if (len >= sizeof(buf))
		len = sizeof(buf) - 1;
	buf[len] = '\0';
	lm.lm_log = new_logmsg(buf, 0);
	if (lm.lm_lpri != 0)
		lm.lm_pri = lm.lm_lpri;
	if (lm.lm_lwhen != 0)
		lm.lm_when = lm.lm_lwhen;
	lm.lm_when = time(NULL);
	lm.lm_len = len;
	logmsg(&lm);
	return ret;
}


static int
check_sources(nfds, delay)
	int nfds;
	time_t delay;
{
	struct timeval tv;
	fd_set rfd, wfd;
	int i, j = 0, n;

	bcopy((char *)&mrfd, (char *)&rfd, sizeof(rfd));
	bcopy((char *)&mwfd, (char *)&wfd, sizeof(wfd));
	tv.tv_usec = 0;
	tv.tv_sec = (delay < 0) ? -delay : delay;

	if (opts & OPT_VERBOSE)
		fprintf(stderr, "listening on %d fds\n", nfds);

	n = select(nfds + 1, &rfd, &wfd, NULL, &tv);
	if ((n == 0) && (delay < 0)) {
		for (i = nfds; (i >= 0); i--)
			if (FD_ISSET(i, &mrfd))
				j |= read_fd(i, 1);
	}

	if (n == -1) {
		if (errno == EINTR)
			return 0;
		syserr("select failed:%m");
		return -1;
	}
	for (i = 0; (n > 0); i++) {
		if (FD_ISSET(i, &wfd)) {
			write_fd(i);
			n--;
		}
		if (FD_ISSET(i, &rfd)) {
			j |= read_fd(i, 0);
			n--;
		}
	}
	return j;
}


int
load_config()
{
	regex_t buf;
	struct dirent *di;
	char *s;
	DIR *d;
	int n;

	bzero((char *)&buf, sizeof(buf));
	if ((opts & OPT_ASKPASSONLY) == 0)
		init_internal();
	if (opts & OPT_OLDSYSLOG)
		n = read_oldconfig(cfgfile, &mrfd);
	else
		n = read_config(cfgfile, &mrfd);

	if ((n <= 0)  && !(opts & OPT_ASKPASSONLY)) {
		fprintf(stderr, "no open sources, exiting...\n");
		exit(1);
	}

	if (options.op_pkgconfs) {
#ifdef	HAVE_REGCOMP
		(void) regcomp(&buf, ".*\\.syslog", REGCOMP_FLAGS);
#else
		buf = (char *)malloc(12);
		(void) compile(".*\\.syslog", buf, buf + 12, '\0');
#endif
		if ((d = opendir(options.op_pkgconfs))) {
			while ((di = readdir(d))) {
#ifdef	HAVE_REGCOMP
				if (!regexec(&buf, di->d_name, 0, NULL, 0))
#else
				if (step(di->d_name, buf))
#endif
				{
					s = malloc(strlen(di->d_name) + 2 + 
						   strlen(options.op_pkgconfs));
					sprintf(s, "%s/%s", di->d_name,
						options.op_pkgconfs);
					(void) read_config(s, &mrfd);
					free(s);
				}
			}
			closedir(d);
		}
#ifdef	HAVE_REGCOMP
		regfree(&buf);
#else
		free(buf);
#endif
	}
	return n;
}


void
setup_dir()
{
	struct stat sb;

	if (!options.op_sysdir)
		options.op_sysdir = strdup(NSYSLOG_DIR);
	if (stat(options.op_sysdir, &sb)) {
		if (mkdir(options.op_sysdir, 0700))
			syserr("mkdir(%s):%m", options.op_sysdir);
	}
}


#ifdef	HAVE_SSL_H
# define	OPENSSL_CONF	"openssl.cnf"
ssl_init()
{
	char *s, *t = NULL, *cert;
	long errline;

	if (opts & (OPT_VERBOSE|OPT_DEBUG))
		BIO_set_fp(bio_err,stderr,BIO_NOCLOSE|BIO_FP_TEXT);
	CRYPTO_mem_ctrl(CRYPTO_MEM_CHECK_ON);
	SSLeay_add_all_algorithms();
	SSL_library_init();
	SSL_load_error_strings();

	s = getenv("OPENSSL_CONF");
	if (s == NULL)
		s = getenv("SSLEAY_CONF");
	options.op_sslcipher = getenv("SSL_CIPHER");
	if (s == NULL) {
		cert = X509_get_default_cert_area();
		t = (char *)malloc(strlen(cert) + 6 + strlen(OPENSSL_CONF));
		sprintf(t, "%s/lib/%s", cert, OPENSSL_CONF);
		s = t;
	}
	config = CONF_load(config, s, &errline);
	if (config == NULL)
		ERR_clear_error();
	if (t)
		free(t);
}
#endif

int
main(argc, argv)
	int argc;
	char *argv[];
{
	time_t	nextcheck, nextmark, nextsync;
	int killflag = 0, reloadflag = 0;
	struct sigaction act;
	int i, n, ch;
	time_t now;
	char **s;
	FILE *fp;

	tzset();
	if (!(progname = strrchr(argv[0], '/')))
		progname = argv[0];
	else
		progname++;

	bzero((char *)&act, sizeof(act));

	act.sa_flags = 0;
	sigemptyset(&act.sa_mask);
	sigaddset(&act.sa_mask, SIGPIPE);
	sigaddset(&act.sa_mask, SIGINT);
	sigaddset(&act.sa_mask, SIGHUP);
	sigaddset(&act.sa_mask, SIGTERM);

	act.sa_handler = SIG_IGN;
	sigaction(SIGPIPE, &act, NULL);

	act.sa_handler = handle_hup;
	sigaction(SIGHUP, &act, NULL);

	act.sa_handler = handle_term;
	sigaction(SIGTERM, &act, NULL);

	act.sa_handler = handle_int;
	sigaction(SIGINT, &act, NULL);

	act.sa_handler = handle_alrm;
	sigaction(SIGALRM, &act, NULL);

	for (s = cfgfiles; cfgfile = *s; s++)
		if (access(cfgfile, 0) == 0)
			break;

	while ((ch = getopt(argc, argv, "AC:df:kl:m:Op:P:rs:uvV")) != -1)
		switch (ch)
		{
		case 'A' :
			opts |= OPT_ASKPASSONLY;
			break;
		case 'C' :
			options.op_chroot = strdup(optarg);
			break;
		case 'd' :
			opts |= OPT_DEBUG;
			dofork--;
			break;
		case 'f' :
			cfgfile = optarg;
			break;
		case 'k' :
			killflag = 1;
			break;
#ifdef	bsdi
		case 'l' :
			opts |= OPT_OLDSYSLOG|OPT_OLDNOUDP;
			break;
#endif
		case 'm' :
			options.op_markfreq = atoi(optarg);
			break;
		/*
		 * backward compatible with development made by others.
		 */
#ifndef bsdi
		case 'l' :
#endif
		case 'p' :
			opts |= OPT_OLDSYSLOG;
			new_var("syslog", VAR_SRC);
			new_obj("syslog");
			(void) add_obj(SRC_FILE, file_new(SRC_UNIX, optarg));
			break;
		case 'O' :
			opts |= OPT_OLDSYSLOG;
			break;
		case 'P' :
			pidfile = optarg;
			break;
		case 'r' :
			reloadflag = 1;
			break;
		case 's' :
			options.op_syncfreq = atoi(optarg);
			break;
		case 'u' :
			opts &= ~OPT_OLDNOUDP;
			opts |= OPT_OLDSYSLOG;
			break;
		case 'v' :
			opts |= OPT_VERBOSE;
			dofork--;
			break;
		case 'V' :
			printf("Version %s\n", NSYSLOG_VERSION);
			exit(0);
		}

	/*
	 * chroot set on the command line..
	 */
	if (options.op_chroot) {
		if (chdir(options.op_chroot))
			syserr("chdir(%s):%m", options.op_chroot);
		if (chroot(options.op_chroot))
			syserr("chdir(%s):%m", options.op_chroot);
	}

	if (killflag || reloadflag) {
		if (!(fp = fopen(pidfile, "r"))) {
			fprintf(stderr, "can't open pidfile\n");
			exit(1);
		}
		ch = fscanf(fp, "%d", &n);
		fclose(fp);
		if (ch == 1) {
			if (kill(n, killflag ? SIGTERM : SIGHUP)) {
				fprintf(stderr, "can't kill pid %d\n", n);
				exit(1);
			}
			exit(0);
		}
		exit(1);
	}

	if (dofork < -1)
		yydebug = 1;

	init_config();

#ifdef	HAVE_SSL_H
	ssl_init();
#endif

	n = load_config();

	setup_dir();

	if (opts & OPT_ASKPASSONLY) {
		hash_initpass(options.op_hashtype);
		exit(0);
	} else if (options.op_hashtype) {
		if (!hash_initfile(options.op_hashtype)) {
			fprintf(stderr, "failed to initialize hashing\n");
			exit(1);
		}
	}

	if (dofork > 0) {
#if BSD >= 199306
		daemon(1, 0);
#else
		if (fork() > 0)
			exit(0);
		setsid();
		close(0);
		close(1);
		/* fd #2 is already something else */
#endif
	}

	started = 1;

	if ((fp = fopen(pidfile, "w"))) {
		fprintf(fp, "%d\n", getpid());
		fflush(fp);
		fclose(fp);
	}
	startup_msg();
	now = time(NULL);

	if (options.op_synctime > 0)
		nextsync = now + options.op_synctime;
	else
		nextsync = 0;
	if (options.op_markfreq > 0)
		nextmark = now + options.op_markfreq;
	else
		nextmark = 0;
	if (nextmark > 0 && nextsync > 0)
		nextcheck = MIN(nextmark, nextsync);
	else if (nextmark > 0)
		nextcheck = nextmark;
	else if (nextsync > 0)
		nextcheck = nextsync;
	else
		nextcheck += 60;

	/*
	 * if set in config file and not on the command line.
	 */
	if (options.op_chroot) {
		if (chdir(options.op_chroot))
			syserr("chdir(%s):%m", options.op_chroot);
		if (chroot(options.op_chroot))
			syserr("chdir(%s):%m", options.op_chroot);
	}

	while ((i = check_sources(n, MIN(60, nextcheck - now))) >= 0) {
		if (do_reopen == 1) {
			do_reopen = 0;
			hup_msg();
			n = reload();
		} else if (do_reopen == 2)
			exit_int();
		else if (do_reopen == 3)
			exit_term();

		now = time(NULL);
		if ((nextmark > 0) && (now >= nextmark) &&
		    (options.op_markfreq > 0)) {
			checkmarks();
			nextmark = time(NULL) + options.op_markfreq;
		}
		if ((nextsync > 0) && (now >= nextsync) &&
		    (options.op_syncfreq > 0)) {
			checksync();
			nextsync = time(NULL) + options.op_synctime;
		}

		if ((nextmark > 0) && (nextsync > 0))
			nextcheck = MIN(nextmark, nextsync);
		else if (nextmark > 0)
			nextcheck = nextmark;
		else if (nextsync > 0)
			nextcheck = nextsync;
		else
			nextcheck += 60;
		if (i != 0)
			nextcheck = now - 1;
	}
	handle_term(SIGTERM);
	/* NOTREACHED */
}
