#include "h.h"
#include "nsyslog.h"
#include "var.h"
#include "hash.h"
#include "y.tab.h"


#if !defined(lint)
static const char rcsid[] = "$Id: matchfile.c,v 1.8 1999/05/10 18:18:46 darrenr Exp $";
#endif


typedef struct matchfilespec {
	struct matchfilespec *mf_next;
	int mf_cnt;
	char *mf_file;
	int mf_nlines;
	regex_t **mf_bufs;
int mf_not;
} matchfilespec_t;

#define	MATCHSZ	512


int matchfile_inccnt __P((void *, int));
void matchfile_print __P((void *));
void matchfile_expire __P((void));
int matchfile_match __P((void *, logmsgtype_t *));

static	matchfilespec_t	*mfslist = NULL;
vartype_t	matchfiletype = {
	NULL,		/* next */
	NULL,		/* ptr */
	FIL_MATCHFILE,	/* stype */
	0,		/* dtype */
	"match-file",	/* name */
	NULL,		/* init */
	NULL,		/* read */
	NULL,		/* readnext */
	NULL,		/* write */
	NULL,		/* link */
	NULL,		/* match */
	NULL,		/* accept */
	NULL,		/* reopen */
	NULL,		/* logmsg */
	matchfile_print,/* print */
	NULL,		/* fd */
	matchfile_inccnt,/* inccnt */
	NULL,		/* drain */
	matchfile_expire,/* expire */
	NULL,	/* expireold */
	NULL,		/* more */
};


matchfilespec_t *
matchfile_find(parm)
char *parm;
{
	matchfilespec_t *mf;

	for (mf = mfslist; mf; mf = mf->mf_next)
		if (!strcmp(mf->mf_file, parm))
			return mf;
	return NULL;
}


int
matchfile_match(ptr, lmp)
	void *ptr;
	logmsgtype_t *lmp;
{
	matchfilespec_t *mf;
	int i, ok;

	mf = ptr;
	ok = mf->mf_not;
	for (i = 0; i < mf->mf_nlines; i++)
#ifdef  HAVE_REGCOMP 
		if (!regexec(mf->mf_bufs[i], lmp->lm_line, 0, NULL, 0))
#else
		if (!step(lmp->lm_line, mf->mf_bufs[i]))
#endif  
			return !ok;
	return ok;
}


matchfilespec_t *
matchfile_new(parm)
	char *parm;
{
	char line[MATCHSZ + 1];
	matchfilespec_t *mf;
	int len, i, not;
	FILE *fp;

	not = get_not();

	if ((mf = matchfile_find(parm)))
		return mf;

	if (!(NEW(mf, NEW_MATCHFS)))
		return NULL;

	if (!(fp = fopen(parm, "r"))) {
		if (opts & OPT_DEBUG)
			fprintf(stderr, "couldn't open file: %s\n", parm);
		else
			syserr("opening file %s: %m\n", parm);
		free(mf);
		return NULL;
	}
	for (i = 0; (fgets(line, sizeof(line) - 1, fp)); i++)
		;
	rewind(fp);
	mf->mf_nlines = i;
	mf->mf_bufs = (void *)malloc(i * sizeof(*mf->mf_bufs));

	for (i = 0; (fgets(line, sizeof(line) - 1, fp)); i++) {
		line[sizeof(line) - 1] = '\0';
#ifdef	HAVE_REGCOMP
		(void) regcomp(mf->mf_bufs[i], parm, REGCOMP_FLAGS);
#else
		len = strlen(line) * 2 + 1024;
		mf->mf_bufs[i] = (regex_t *)calloc(1, len);
		(void) compile(line, mf->mf_bufs[i], mf->mf_bufs[i] + len,'\n');
#endif
	}
	fclose(fp);
	mf->mf_file = strdup(parm);
	mf->mf_next = mfslist;
	mf->mf_not = not;
	mf->mf_cnt = 0;
	mfslist = mf;
	return mf;
}


void
matchfile_expire()
{
	matchfilespec_t	*mfs;
	int	i;

	while ((mfs = mfslist)) {
		assert(mfs->mf_cnt == 0);
		mfslist = mfs->mf_next;

		for (i = 0; i < mfs->mf_nlines; i++)
#ifdef	HAVE_REGCOMP
			regfree(mfs->mf_bufs[i]);
#else
			free(mfs->mf_bufs[i]);
#endif
		free(mfs->mf_bufs);
		free(mfs->mf_file);
		free(mfs);
	}

}


int
matchfile_inccnt(ptr, inc)
	void *ptr;
	int inc;
{
	matchfilespec_t *mf;

	mf = ptr;
	mf->mf_cnt += inc;
	return mf->mf_cnt;
}


void
matchfile_print(ptr)
	void *ptr;
{
	matchfilespec_t *mf;

	mf = ptr;
	print_not(mf->mf_not);
	printf(" match-file %s ", mf->mf_file);
}
