#include "h.h"
#include "nsyslog.h"
#include "var.h"
#include "hash.h"
#include "htcp.h"
#ifdef	HAVE_TCPD_H
#include <tcpd.h>
#endif


#if !defined(lint)
static const char rcsid[] = "$Id: net.c,v 1.20 1999/05/10 18:25:21 darrenr Exp $";
#endif

extern	char	*progname;


void	*net_vinit __P((int, arglist_t *));
int	net_read __P((void *, logmsgtype_t *, tcplmsg_t *, char *, size_t));
int	net_readnext __P((void *, logmsgtype_t *, tcplmsg_t *,char *, size_t));
int	net_send __P((void *, char *, size_t, char *, size_t));
void	net_link __P((void *, void *));
int	net_logmsg __P((void *, int *, struct iovec *, int, logmsgtype_t *));
void	net_reopen __P((void *));
void	net_print __P((void *));
int	net_fd __P((void *));
int	net_inccnt __P((void *, int));
void	net_drain __P((void *));
void	net_expire __P((void));
void	net_expireold __P((void));
int	net_initsrc __P((int));
int	net_more __P((void *));

netspec_t	*net_find __P((int, int, struct in_addr, u_int));


netspec_t	*nslist = NULL;
vartype_t	udptype = {
	NULL,		/* next */
	NULL,		/* ptr */
	SRC_INET,	/* stype */
	DEST_INET,	/* dtype */
	"udp",		/* name */
	net_vinit,	/* init */
	net_read,	/* read */
	NULL,		/* readnext */
	net_send,	/* write */
	NULL,		/* link */
	NULL,		/* match */
	NULL,		/* accept */
	net_reopen,	/* reopen */
	net_logmsg,	/* logmsg */
	net_print,	/* print */
	net_fd,		/* fd */
	net_inccnt,	/* inccnt */
	NULL,		/* drain */
	net_expire,	/* expire */
	net_expireold,	/* expireold */
	NULL,		/* more */
};
vartype_t	tcptype = {
	NULL,		/* next */
	NULL,		/* ptr */
	TCP_LISTEN,	/* stype */
	DEST_TCP,	/* dtype */
	"tcp",		/* name */
	net_vinit,	/* init */
	net_read,	/* read */
	net_readnext,	/* readnext */
	net_send,	/* write */
	net_link,	/* link */
	NULL,		/* match */
	net_initsrc,	/* accept */
	net_reopen,	/* reopen */
	net_logmsg,	/* logmsg */
	net_print,	/* print */
	net_fd,		/* fd */
	net_inccnt,	/* inccnt */
	net_drain,	/* drain */
	NULL,		/* expire */
	NULL,		/* expireold */
	net_more,	/* more */
};


extern	int	do_raw;
extern	fd_set	mrfd, mwfd;


/*
 * create a new network socket based on args present.  at least once arg must
 * be passed (the IP address) whilst the second (optional) is a port number.
 */
void *
net_vinit(type, args)
	int type; 
	arglist_t *args;
{
	int proto, base;
	char *p1, *p2;

	if (args == NULL) 
		return NULL;

	base = VAR_TYPE(type);
	p1 = args->a_arg;
	args = args->a_next;
	if (args == NULL)
		return NULL;
	p2 = args->a_arg;

	if (base == SRC_INET || base == DEST_INET)
		proto = NET_UDP;
	else if (base == SRC_INET_LISTEN || base == DEST_TCP)
		proto = NET_TCP;
	return net_new(type, proto, p1, p2);
}


void
net_link(ptr1, ptr2)
	void *ptr1;
	void *ptr2;
{
	netspec_t *nsp = ptr1;		/* parent */
	netspec_t *nsc = ptr2;		/* child */

	nsc->ns_parent = nsp;
}


void
net_delete(ptr)
	void *ptr;
{
	netspec_t *ns = ptr;

	register netspec_t *n, **np;

	for (np = &nslist; (n = *np); np = &n->ns_next)
		if (n == ns) {
			*np = n->ns_next;
			break;
		}
}


netspec_t *
net_find(type, proto, ip, porti)
	int type;
	int proto;
	struct in_addr ip;
	u_int porti;
{
	netspec_t *ns;
	u_short port = porti & 0xffff;

	for (ns = nslist; ns; ns = ns->ns_next)
		if ((ns->ns_type == type) && (ns->ns_proto == proto) &&
		    (ns->ns_sin.sin_port == port) &&
		    !bcmp(&ns->ns_sin.sin_addr, &ip, sizeof(ip)))
			return ns;
	return NULL;
}


netspec_t *
net_new(type, proto, parm1, parm2)
	int type;
	int proto;
	char *parm1;
	char *parm2;
{
	struct in_addr ip;
	netspec_t *ns;
	u_short port;
	int fd;

	if (parm1 != NULL)
		ip = getipv4addr(parm1);
	if (parm2 != NULL)
		port = getportnum(NULL, parm2);
	if (parm1 != NULL && parm2 != NULL) {
		if ((ns = net_find(type, proto, ip, (u_int)port))) {
			/*
			 * If it's old and we're reusing it, then we need to
			 * reopen the hash and text log files.
			 */
			if ((ns->ns_mode & VAR_OLD) == VAR_OLD) {
				if (ns->ns_hfd >= 0)
					ns->ns_hfd = net_openlog(ns, 0);
				if ((ns->ns_hfd >= 0) && (ns->ns_lfd >= 0)) {
					ns->ns_lfd = net_openlog(ns, 1);
					if (ns->ns_lfd == -1) {
						close(ns->ns_hfd);
						ns->ns_hfd = -1;
					}
				}
				if (ns->ns_hfd == -1)
					syserr("couldn't open host log files");
				ns->ns_mode &= ~VAR_OLD;
			}
			if (ns->ns_fd >= 0)
				ft_set(ns->ns_fd, ns, ns->ns_type);
			return ns;
		}
	}

	if (!(NEW(ns, NEW_NETS)))
		return NULL;
	ns->ns_cnt = 0;
	ns->ns_hfd = -1;
	ns->ns_lfd = -1;
	ns->ns_lcnt = 0;
#ifdef HAVE_SSL_H
	ns->ns_ssl = NULL;
#endif
	ns->ns_rtail = &ns->ns_rlist;
	ns->ns_wtail = &ns->ns_wlist;
	ns->ns_type = type;
	ns->ns_proto = proto;
	ns->ns_path = NULL;
	ns->ns_hptr = hash_find(HASH_NONE);
	ns->ns_sin.sin_family = AF_INET;
	if (parm1 != NULL)
		ns->ns_sin.sin_addr = ip;
	if (parm2 != NULL)
		ns->ns_sin.sin_port = port;
	if (proto != NET_CLONE)
		fd = new_socket(type, proto, (sockaddr_t *)&ns->ns_sin,
				       sizeof(ns->ns_sin));
	else
		fd = -1;

	if (fd >= 0) {
		async_fd(fd, 1);
		(void) fcntl(fd, F_SETFD, fcntl(fd, F_GETFD, 0) | FD_CLOEXEC);
		ft_set(fd, ns, ns->ns_type);
	}
	ns->ns_fd = fd;

	if ((proto == NET_TCP) && (type == DEST_TCP)) {
		if (htcp_initdst(ns) == -1) {
			syserr("failed to make TCP connection to %s,%s",
			       parm1, parm2);
			ns->ns_fd = -1;
		}
		ns->ns_path = strdup(inet_ntoa(ns->ns_sin.sin_addr));
		ns->ns_hmac = (char *)malloc(sizeof(hashrec_t));
	}

	if (ns != NULL) {
		if (do_raw) {
			ns->ns_mode |= do_raw;
			do_raw = 0;
		}
		ns->ns_next = nslist;
		nslist = ns;
	}
	return ns;
}


void
net_sslfree(ns)
	netspec_t *ns;
{
#ifdef	HAVE_SSL_H
	ssl_t *ssl;

	ssl = ns->ns_ssl;
	if (ssl != NULL)
		if (ssl != NULL) {
			if (ssl->ssl_ctx)
				SSL_CTX_free(ssl->ssl_ctx);
			if (ssl->ssl_ssl)
				SSL_free(ssl->ssl_ssl);
			ns->ns_ssl = NULL;
		}
#endif
}


void
net_free(ns)
	netspec_t *ns;
{
	netspec_t *n, **np;
	int fd;

	fd = ns->ns_fd;
	if (ns->ns_cnt > 0)
		return;
	net_sslfree(ns);
	/*
	 * Get rid of any child connections that have been brought in by this
	 * listening socket.
	 */
	if (VAR_TYPE(ns->ns_type) == SRC_INET_LISTEN) {
		for (np = &nslist; (n = *np); )
			if (n->ns_parent == ns) {
				*np = n->ns_next;
				net_free(n);
			} else
				np = &n->ns_next;
	}
	/*
	 * If it's not in use any more, then clear it all away.
	 */
	if (ns->ns_hmac != NULL)
		free(ns->ns_hmac);
	if (ns->ns_path != NULL)
		free(ns->ns_path);
	if (ns->ns_hfd != -1)
		(void) close(ns->ns_hfd);
	if (ns->ns_lfd != -1)
		(void) close(ns->ns_lfd);
	if (fd >= 0) {
		(void) close(fd);
		FD_CLR(fd, &mrfd);
		FD_CLR(fd, &mwfd);
		fttable[fd].ft_closed = time(NULL);
		fttable[fd].ft_ptr = NULL;
	}
	if (ns->ns_nonce != NULL) {
		free(ns->ns_nonce);
		ns->ns_nonce = NULL;
	}
	if (ns->ns_rnonce != NULL) {
		free(ns->ns_rnonce);
		ns->ns_rnonce = NULL;
	}
	/*
	 * free up buffers in use only if it's not in use any more.
	 */
	buf_freelist(ns->ns_rlist);
	buf_freelist(ns->ns_wlist);
	net_delete(ns);
	free(ns);
}


void
net_expire()
{
	netspec_t **nsp, *ns;
	int type;

	for (nsp = &nslist; (ns = *nsp); ) {
		type = VAR_TYPE(ns->ns_type);
		if (type == SRC_INET || type == SRC_INET_LISTEN ||
		    type == SRC_TCP || type == DEST_TCP) {
			if (opts & OPT_VERBOSE)
				fprintf(stderr, "keep net: %s,%d (%d) %d\n",
					inet_ntoa(ns->ns_sin.sin_addr),
					ntohs(ns->ns_sin.sin_port), ns->ns_cnt,
					ns->ns_fd);
			if (VAR_TYPE(ns->ns_type) != SRC_TCP)
				ns->ns_mode |= VAR_OLD;
			if (ns->ns_hfd >= 0)
				close(ns->ns_hfd);
			if (ns->ns_lfd >= 0)
				close(ns->ns_lfd);
			nsp = &ns->ns_next;
			continue;
		}
		if (opts & OPT_VERBOSE)
			fprintf(stderr, "close net: %s,%d (%d)\n",
				inet_ntoa(ns->ns_sin.sin_addr),
				ntohs(ns->ns_sin.sin_port), ns->ns_cnt);
		nsp = &nslist;
		net_free(ns);
	}
}


void
net_expireold()
{
	netspec_t **nsp, *ns;

	for (nsp = &nslist; (ns = *nsp); ) {
		if (!(ns->ns_mode & VAR_OLD)) {
			nsp = &ns->ns_next;
			continue;
		}
		if (opts & OPT_VERBOSE)
			fprintf(stderr, "close net: %s,%d (%d)\n",
				inet_ntoa(ns->ns_sin.sin_addr),
				ntohs(ns->ns_sin.sin_port), ns->ns_cnt);
		nsp = &nslist;
		net_free(ns);
	}
}


int
new_socket(type, proto, sin, slen)
	int type, proto;
	sockaddr_t *sin;
	int slen;
{
	int fd = -1, opt = 1, res, nonb = 0, af, pf, base;
	sockaddr_t sins;

	base = VAR_TYPE(type);
	switch (base)
	{
	case SRC_INET_LISTEN :
	case DEST_INET :
	case SRC_INET :
	case DEST_TCP :
		af = AF_INET;
		if (proto == NET_TCP)
			pf = SOCK_STREAM;
		else if (proto == NET_UDP)
			pf = SOCK_DGRAM;
		else {
			logerror("unknown network protocol");
			fd = -2;
		}
		break;
	case SRC_UNIX_LISTEN :
	case DEST_UNIX :
	case SRC_UNIX :
		af = AF_UNIX;
		pf = proto;
		break;
	default :
		logerror("unknown network type");
		fd = -2;
		break;
	}

	if (fd != -2)
		fd = socket(af, pf, 0);

	if (fd == -1)
		syserr("socket: %m");
	else {
		if (af == AF_INET)
			setsockopt(fd, SOL_SOCKET, SO_REUSEADDR, (char *)&opt,
				   sizeof(opt));
		if (VAR_MODE(type) == SRC_NONE) {
			if (base == SRC_UNIX || base == SRC_UNIX_LISTEN)
				(void) unlink(((sun_t *)sin)->sun_path);
			if (bind(fd, sin, slen)) {
				syserr("bind failed: %m");
				close(fd);
				fd = -1;
			}
			if (base == SRC_UNIX || base == SRC_UNIX_LISTEN) {
				if (chmod(((sun_t *)sin)->sun_path, 0666) != 0)
					syserr("chmod on (%s) failed: %m",
						((sun_t *)sin)->sun_path);
			}
			if ((type & VAR_LISTEN) != 0) {
				if (listen(fd, 1) != 0)
					syserr("listen failed: %m");
				else if (opts & OPT_VERBOSE)
					fprintf(stderr, "listening on fd %d\n",
						fd);
			}
		} else if (VAR_MODE(type) == DEST_NONE) {
			if (connect(fd, sin, slen) == -1) {
				if (errno != EINPROGRESS) {
					syserr("connect failed: %m");
					close(fd);
					fd = -1;
				}
			}
		}
	}
	return fd;
}


int
net_ready(ns)
	netspec_t *ns;
{
	if ((ns->ns_state >= HTCS_MSG) || (ns->ns_nstate >= HTCS_MSG))
		return 1;
	if (opts & OPT_VERBOSE)
		fprintf(stderr, "network socket not ready\n");
	return 0;
}


int
net_more(ptr)
	void *ptr;
{
	netspec_t *ns;

	ns = ptr;
	return buf_size(ns->ns_rlist);
}


void
net_drain(ptr)
	void *ptr;
{
	netspec_t *ns;
	int fd;

	ns = ptr;
	fd = ns->ns_fd;

	if (htcp_action(ns) == -1) {
		if (VAR_TYPE(ns->ns_type) == SRC_TCP)
			net_free(ns);
		else {
			close(fd);
			FD_CLR(fd, &mrfd);
			FD_CLR(fd, &mwfd);
			fttable[fd].ft_closed = time(NULL);
			ns->ns_fd = -1;
		}
	}

	if (ns->ns_wlist == NULL)
		FD_CLR(fd, &mwfd);
}


int
net_logmsg(ptr, fdp, iov, nio, lmp)
	void *ptr;
	int *fdp;
	struct iovec *iov;
	int nio;
	logmsgtype_t *lmp;
{
	struct iovec liov[MAX_NIOV];
	int mode, fd, nw;
	filetype_t *ft;
	netspec_t *ns;
	char pri[18];

	ns = ptr;
	mode = ns->ns_mode;
	fd = ns->ns_fd;
	ft = fttable + fd;
	*fdp = fd;
	nw = 0;

	if (opts & OPT_VERBOSE)
		fprintf(stderr, "%smsg to host %s\n",
			(mode & VAR_RAW) ? "raw " : "",
			inet_ntoa(ns->ns_sin.sin_addr));

	/*
	 * If a destiation is marked as "raw", then send it with the
	 * priority information intact and without the hostname. 
	 */
	if (!(mode & VAR_RAW)) {
		if (fd >= 0) {
			if (ns->ns_proto == NET_TCP) {
				/*
				 * add it before calling sendmsg so
				 * that we can send the correct hash.
				 */
				if (net_ready(ns) != 0) {
					hash_addrec(lmp, ft, iov, nio);
					htcp_sendmsg(ns, lmp, iov, nio);
				}
			} else
				nw = writev(fd, iov, nio);
		}
	} else {
		bcopy((char *)iov, (char *)liov, sizeof(liov));
		liov[nio - 3].iov_len = liov[0].iov_len;
		liov[nio - 3].iov_base = liov[0].iov_base;
		snprintf(pri, sizeof(pri) - 1, "<%d>",
			 lmp->lm_lpri & (LOG_FACMASK|LOG_PRIMASK));
		liov[nio - 4].iov_base = pri;
		liov[nio - 4].iov_len = strlen(pri);
		if (fd >= 0) {
			if (ns->ns_proto == NET_TCP) {
				/*
				 * add it before calling sendmsg so
				 * that we can send the correct hash.
				 */
				if (net_ready(ns) != 0) {
					hash_addrec(lmp, ft, liov, 4);
					htcp_sendmsg(ns, lmp,
						     liov + (nio - 4),
						     4);
				}
			} else
				nw = writev(fd, liov + (nio - 4), 4);
		}
	}
	if (nw < 0)
		syserr("error writing to %s: %m",
		       inet_ntoa(ns->ns_sin.sin_addr));
	return nw;
}


void
net_reopen(ptr)
	void *ptr;
{
	netspec_t *ns;

	ns = ptr;
	ns->ns_fd = new_socket(ns->ns_type, ns->ns_proto,
			       (sockaddr_t *)&ns->ns_sin, sizeof(ns->ns_sin));
}


int
net_recv(ptr, sin, slen)
	void *ptr;
	struct sockaddr *sin;
	int *slen;
{
	buflist_t *bl, *bn, **b;
	int i, len, error;
	netspec_t *ns;
	char *buf;

	ns = ptr;
	if (ioctl(ns->ns_fd, FIONREAD, &len) == -1) {
		syserr("ioctl(%d[%s],FIONREAD) error, closing:%m", ns->ns_fd,
			inet_ntoa(ns->ns_sin.sin_addr));
		return -1;
	}
	if (len == 0) {
		syserr("EOF from fd %d[%s], closing", ns->ns_fd,
			inet_ntoa(ns->ns_sin.sin_addr));
		return -1;
	}
	if (len < 0) {
		syserr("ioctl error from fd %d[%s], closing:%m", ns->ns_fd,
			inet_ntoa(ns->ns_sin.sin_addr));
		return -1;
	}

	bl = buf_new(len);
	if (bl == NULL) {
		syserr("failed to allocate read buffer, closing %s",
			inet_ntoa(ns->ns_sin.sin_addr));
		return -1;
	}
	*ns->ns_rtail = bl;
	ns->ns_rtail = &bl->bl_next;

	error = recv(ns->ns_fd, bl->bl_ptr, len, 0);
	if (error == -1 && (errno == EAGAIN || errno == EWOULDBLOCK ||
			    errno == ENOBUFS || errno == EINTR))
		return 0;
	if (error <= 0)
		return -1;
	bl->bl_rem += error;
	return error;
}


int
net_read(ptr, lmp, tlm, buf, bufsz)
	void *ptr;
	logmsgtype_t *lmp;
	tcplmsg_t *tlm;
	char *buf;
	size_t bufsz;
{
	buflist_t **bh;
	netspec_t *ns;
	int len, dlen;
	ssl_t *ssl;
	char *s;

	ns = ptr;
	if (ns->ns_proto == NET_TCP) {
		len = net_recv(ns, &lmp->lm_src, &lmp->lm_slen);
		if ((len <= 0) && (ns->ns_rlist == NULL)) {
			if ((len == -1) && (ns->ns_type == SRC_TCP))
				net_free(ns);
			return len;
		}

		len = net_readnext(ptr, lmp, tlm, buf, bufsz);
	} else if (ns->ns_proto == NET_UDP) {
		struct sockaddr_in *sinp;

		len = recvfrom(ns->ns_fd, buf, bufsz, 0, lmp->lm_src,
			       &lmp->lm_slen);
		if (len < 0)
			syserr("recvfrom(%d,%s,%d):%m", ns->ns_fd,
				inet_ntoa(sinp->sin_addr), lmp->lm_slen);
		lmp->lm_flags = FLG_USER|FLG_SYS;
	}
	return len;
}


int
net_readnext(ptr, lmp, tlm, buf, bufsz)
	void *ptr;
	logmsgtype_t *lmp;
	tcplmsg_t *tlm;
	char *buf;
	size_t bufsz;
{
	buflist_t **bh;
	netspec_t *ns;
	int len, dlen;
	ssl_t *ssl;
	char *s;

	ns = ptr;
	len = 0;
	if (ns->ns_proto == NET_TCP) {
		switch (htcp_action(ns))
		{
		case -1 :
			if ((len <= 0) && (ns->ns_type == SRC_TCP))
				net_free(ns);
			return -1;
		case -2 :
			return 0;
		default :
			break;
		}

		if (ns->ns_state == HTCS_MSG) {
			ssl = ns->ns_ssl;
			bh = &ns->ns_rlist;
			dlen = buf_getmsg(bh, tlm, ssl, buf, bufsz);
			if (ns->ns_rlist == NULL)
				ns->ns_rtail = bh;
			if ((ns->ns_rlist == NULL) && (ssl != NULL) &&
			    (ssl->ssl_rlist != NULL)) {
				bh = &ssl->ssl_rlist;
				dlen = buf_getmsg(bh, tlm, NULL, buf, bufsz);
				if (ssl->ssl_rlist == NULL)
					ssl->ssl_rtail = bh;
			}
			if (dlen == 0)
				return 0;
			if (opts & OPT_DEBUG) {
				assert(dlen == tlm->tl_len);
				assert(tlm->tl_hlen == 0 ||
				       tlm->tl_hlen == sizeof(hashrec_t));
			}
			len = dlen - sizeof(*tlm);
			lmp->lm_flags = FLG_HTCP|FLG_HAVEPRI;
			lmp->lm_src = &ns->ns_sin;
			lmp->lm_slen = sizeof(ns->ns_sin);
			/*      
			 * Because the TCP format brings with it its own
			 * details regarding source, time, etc, set those
			 * now that the logmsg struct has been allocated.   
			 */     
			lmp->lm_lpri = tlm->tl_pri;
			lmp->lm_lwhen = tlm->tl_tv;
			/*
			 * If there appears to be some hash information that
			 * is also a part of this message, check to see if it
			 * would make up a hashrec_t.  If it doesn't match in
			 * size then discard it.
			 */
			if ((tlm->tl_hlen > 0) &&
			    (tlm->tl_hlen != sizeof(hashrec_t))) {
				buf_discard(bh, tlm->tl_hlen);
				if (opts & OPT_VERBOSE)
					fprintf(stderr, "discarding %d\n",
						tlm->tl_hlen);
			} else if (tlm->tl_hlen == sizeof(hashrec_t)) {
				dlen = buf_get(bh, ns->ns_hmac, tlm->tl_hlen);
				if (opts & OPT_DEBUG)
					fprintf(stderr,
						"ns_hmac len %d/%d fd %d/%d\n",
						tlm->tl_hlen, dlen, ns->ns_lfd,
						ns->ns_hfd);
				if ((dlen == tlm->tl_hlen) &&
				    (ns->ns_hfd >= 0)) {
					(void) write(ns->ns_hfd, ns->ns_hmac,
						     sizeof(hashrec_t));
					if (opts & OPT_VERBOSE)
						fprintf(stderr,
							"mac info to %s\n",
							ns->ns_path);
				}
				if ((len > 0) && (ns->ns_lfd >= 0)) {
					(void) write(ns->ns_lfd, buf,
						     tlm->tl_mlen);
					if (opts & OPT_VERBOSE)
						fprintf(stderr,
							"msg saved to %s\n",
							ns->ns_path);
				}
			}
			if (ns->ns_rlist == NULL)
				ns->ns_rtail = &ns->ns_rlist;
			if ((ssl != NULL) && (ssl->ssl_rlist == NULL))
				ssl->ssl_rtail = &ssl->ssl_rlist;
		} else
			len = 0;
	}
	return len;
}


int
net_send(ptr, buf1, len1, buf2, len2)
	void *ptr;
	char *buf1;
	size_t len1;
	char *buf2;
	size_t len2;
{
	netspec_t *ns;
	size_t tosend;
	int error;

	if (buf1 == NULL || len1 == 0) {
		len1 = 0;
		buf1 = NULL;
	}
	if (buf2 == NULL || len2 == 0) {
		len2 = 0;
		buf2 = NULL;
	}
	ns = ptr;
	tosend = len1 + len2;

	/*
	* queue present, so ensure that this send() looks like it has sent
	* nothing, causing it to all be added to the queue later.
	*/
	if (buf1 != NULL && ns->ns_wlist == NULL)
		error = send(ns->ns_fd, buf1, len1, 0);
	else
		error = 0;

	if (error == -1) {
		if (errno == EAGAIN || errno == EWOULDBLOCK ||
		    errno == ENOBUFS || errno == EINTR)
			error = 0;
		else if (errno == ENOTCONN) {
			error = connect(ns->ns_fd,
					(struct sockaddr *)&ns->ns_sin,
					sizeof(ns->ns_sin));
			if (error == -1 && (errno == EINPROGRESS ||
			    errno == EALREADY))
				error = 0;
		}
	}
	if (error == -1)
		return -1;
	tosend -= error;

	if ((error == len1) && (len2 != 0) && (buf2 != NULL)) {
		error = send(ns->ns_fd, buf2, len2, 0);
		if (error == -1 && (errno == EAGAIN || errno == EWOULDBLOCK ||
		    errno == ENOBUFS || errno == EINTR))
			error = 0;
		if (error == -1)
			return -1;
		tosend -= error;
	}

	/*
	 * Didn't send it all, so put what remains into buffers.
	 */
	if (tosend > 0) {
		if (tosend > len2) {
			buf1 += len1 - (tosend - len2);
			len1 = (tosend - len2);
		} else {
			len1 = tosend;
			buf1 = buf2 + (len2 - tosend);
			buf2 = NULL;
			len2 = 0;
		}
		ns->ns_wtail = buf_add(ns->ns_wtail, buf1, len1, buf2, len2);
	}
	return tosend;
}


int
net_initsrc(lfd)
	int lfd;
{
	int len, slen, fd;
	netspec_t *ns;

	ns = net_new(SRC_TCP, NET_CLONE, NULL, NULL);
	if (ns == NULL)
		return -1;
	slen = sizeof(ns->ns_sin);
	fd = accept(lfd, (struct sockaddr *)&ns->ns_sin, &slen);
	if (fd == -1) {
		net_free(ns);
		return -1;
	}
#if defined(HAVE_LIBWRAP) && defined(HAVE_TCPD_H)
	{
		struct request_info request;

		request_init(&request, RQ_DAEMON, progname, RQ_FILE, fd, 0);
		if (!hosts_access(&request)) {
			net_free(ns);
			return -1;
		}
	}
#endif

	async_fd(fd, 1);
	ns->ns_fd = fd;
	ns->ns_proto = NET_TCP;
	ft_set(fd, ns, ns->ns_type);
	ns->ns_hmac = (char *)malloc(sizeof(hashrec_t));
	ns->ns_path = strdup(inet_ntoa(ns->ns_sin.sin_addr));
	(void) fcntl(fd, F_SETFD, fcntl(fd, F_GETFD, 0) | FD_CLOEXEC);

	/*
	 * we start in the HTCS_HASH_FN state as we're expecting a HASH_FN
	 * command first up.
	 */
	ns->ns_state = HTCS_HASH_FN;
	if (opts & OPT_VERBOSE)
		fprintf(stderr, "new connection from %s,%d\n",
			ns->ns_path, ntohs(ns->ns_sin.sin_port));
	return fd;
}


int
net_inccnt(ptr, inc)
	void *ptr;
	int inc;
{
	netspec_t *ns;

	ns = ptr;
	ns->ns_cnt += inc;
	return ns->ns_cnt;
}


int
net_fd(ptr)
	void *ptr;
{
	netspec_t *ns;

	ns = ptr;
	return ns->ns_fd;
}


void
net_print(ptr)
	void *ptr;
{
	netspec_t *ns;

	ns = ptr;
	if (ns->ns_proto == NET_TCP)
		printf(" tcp");
	else if (ns->ns_proto == NET_UDP)
		printf(" udp");
	else
		printf(" <%d>", ns->ns_proto);
	printf(" %s,%d ", inet_ntoa(ns->ns_sin.sin_addr),
		ns->ns_sin.sin_port);
}


int
net_openlog(ns, which)
	netspec_t *ns;
	int which;
{
	hashtype_t *hptr;
	char *path, *ext;
	int fd;

	if (which == 0) {
		hptr = ns->ns_hptr;
		ext = hptr->ht_tab->ht_ext;
		path = (char *)malloc(strlen(ext) + strlen(ns->ns_path) + 1);
		sprintf(path, "%s%s", ns->ns_path, ext);
	} else if (which == 1)
		path = ns->ns_path;
	fd = open_file(DEST_FILE, path);
	if (path != ns->ns_path)
		free(path);
	return fd;
}


#ifdef	HAVE_SSL_H
static int snet_write __P((BIO *h,char *buf,int num));
static int snet_read __P((BIO *h,char *buf,int size));
static long snet_ctrl __P((BIO *h,int cmd,long arg1,char *arg2));
static int snet_new __P((BIO *h));
static int snet_free __P((BIO *data));
#ifndef NO_DH
static DH * get_dh512 __P(());
#endif
static RSA * tmp_rsa_cb __P((SSL *, int, int));


static	BIO_METHOD	syslog_methd = {
	BIO_TYPE_SOCKET,
	"syslog",
	snet_write,
	snet_read,
	NULL,	/* sock_puts */
	NULL,	/* sock_gets */
	snet_ctrl,
	snet_new,
	snet_free
};


static int snet_new(b)
BIO *b;
{
	b->init = 0;
	b->num = 0;
	b->ptr = NULL;
	b->flags = 0;
	return 1;
}


static int snet_free(b)
BIO *b;
{
	if (b == NULL)
		return NULL;
	if (b->shutdown) {
		b->flags = 0;
		b->init = 0;
	}
	return 1;
}


static int snet_read(bio, buf, bufsz)
BIO *bio;
char *buf;
int bufsz;
{
	netspec_t *ns;
	htcpcmd_t hc;
	buflist_t *b;
	int len = 0;

	errno = 0;
	ns = fttable[bio->num].ft_ptr;
	if (buf != NULL) {
		b = ns->ns_rlist;
		if (b == NULL)
			return 0;
		if ((ns->ns_state != HTCS_MSG) &&
		    ((ns->ns_ssl->ssl_r == 0) || ((b->bl_ptr == b->bl_nh) &&
		     (ns->ns_ssl->ssl_r < b->bl_rem)))) {
			buf_get(&ns->ns_rlist, (char *)&hc, sizeof(hc));
			ns->ns_ssl->ssl_r = ntohs(hc.hth_op);
			if ((hc.hth_op > 0) && (ns->ns_ssl->ssl_r < b->bl_rem))
				b->bl_nh = b->bl_ptr + hc.hth_op;
			else
				b->bl_nh = NULL;
		}
		if (ns->ns_ssl->ssl_r > buf_size(b))
			return 0;
		len = buf_get(&ns->ns_rlist, buf, bufsz);
		if (ns->ns_ssl->ssl_r > 0)
			ns->ns_ssl->ssl_r -= len;
		BIO_clear_retry_flags(bio);
		if (ns->ns_rlist == NULL)
			ns->ns_rtail = &ns->ns_rlist;
	}
	return len;
}


static int snet_write(bio, buf, bufsz)
BIO *bio;
char *buf;
int bufsz;
{
	int i;

	errno = 0;
	i = htcp_send_excmd(fttable[bio->num].ft_ptr, HTC_SSL|HTC_VARLEN,
			    bufsz, buf, bufsz);
	if (i < 0)
		return i;
	return bufsz;
}



static long snet_ctrl(b, cmd, num, ptr)
BIO *b;
int cmd;
long num;
char *ptr;
{
	long ret;
	int *ip;

	ret = 1;

	switch (cmd)
	{
	case BIO_CTRL_RESET:
		ret = 0;
		break;
	case BIO_CTRL_INFO:
		ret = 0;
		break;
	case BIO_C_SET_FD:
		snet_free(b);
		b->num = *((int *)ptr);
		b->shutdown = (int)num;
		b->init = 1;
		break;
	case BIO_C_GET_FD:
		if (b->init) {
			ip=(int *)ptr;
			if (ip != NULL) *ip=b->num;
			ret=b->num;
		} else
			ret= -1;
		break;
	case BIO_CTRL_GET_CLOSE:
		ret = b->shutdown;
		break;
	case BIO_CTRL_SET_CLOSE:
		b->shutdown = (int)num;
		break;
	case BIO_CTRL_PENDING:
	case BIO_CTRL_WPENDING:
		ret = 0;
		break;
	case BIO_CTRL_DUP:
	case BIO_CTRL_FLUSH:
		ret = 1;
		break;
		break;
	default:
		ret = 0;
		break;
	}
	return ret;
}


static int snet_puts(bp, str)
BIO *bp;
char *str;
{
	int len, ret;

	len = strlen(str);
	ret = snet_write(bp, str, len);
	return ret;
}

extern	BIO	*bio_err;

void
net_sslinit(ns, server)
	netspec_t *ns;
	int server;
{
	SSL_CTX *ctx;
	ssl_t *ssl;
	BIO *bio;
	SSL *s;

	if (options.op_ssl == 0)	/* is SSL turned off ? */
		return;			/* Yes, it is */
	if (options.op_sslcert == NULL)
		return;			/* No certificate set. */
	ssl = (ssl_t *)malloc(sizeof(ssl_t));
	if (ssl != NULL) {
		bzero((char *)ssl, sizeof(*ssl));
		ssl->ssl_rtail = &ssl->ssl_rlist;
		ssl->ssl_cert = options.op_sslcert;
		ssl->ssl_meth = SSLv23_method();
		ctx = SSL_CTX_new(ssl->ssl_meth);

		if (ctx == NULL) {
			if (opts & (OPT_DEBUG|OPT_VERBOSE))
				fprintf(stderr, "SSL_CTX_new failed\n");
			free(ssl);
			return;
		}
		ssl->ssl_ctx = ctx;

		if (!SSL_CTX_set_default_verify_paths(ctx)) {
			if (opts & OPT_VERBOSE)
				fprintf(stderr,
					"SSL_load_verify_locations failed\n");
		}

		s = SSL_new(ssl->ssl_ctx);
		ssl->ssl_ssl = s;
		if (s == NULL) {
			if (opts & (OPT_DEBUG|OPT_VERBOSE))
				fprintf(stderr, "SSL_new failed\n");
			SSL_CTX_free(ssl->ssl_ctx);
			free(ssl);
			return;
		}
		if (ns->ns_type == DEST_TCP)
			SSL_set_session(s, NULL);
		s->debug = 1;

		SSL_use_RSAPrivateKey_file(s, ssl->ssl_cert, SSL_FILETYPE_PEM);
		SSL_use_certificate_file(s, ssl->ssl_cert, SSL_FILETYPE_PEM);

		if (options.op_sslcipher != NULL)
			SSL_CTX_set_cipher_list(ctx, options.op_sslcipher);

#ifndef NO_DH
		{
		DH *dh;

		dh = get_dh512();
		SSL_CTX_set_tmp_dh(ctx, dh);
		DH_free(dh);
		}
#endif
#ifndef NO_RSA
		SSL_CTX_set_tmp_rsa_callback(ctx, tmp_rsa_cb);
#endif
		if (server)
			SSL_set_accept_state(s);
		else
			SSL_set_connect_state(s);
#if 1
		bio = BIO_new(&syslog_methd);
		if (bio == NULL) {
			SSL_CTX_free(ssl->ssl_ctx);
			SSL_free(s);
			free(ssl);
			return;
		}
		BIO_set_fd(bio, ns->ns_fd, BIO_NOCLOSE);
		SSL_set_fd(s, ns->ns_fd);
		SSL_set_bio(s, bio, bio);
#else
		async_fd(ns->ns_fd, 0);
		if (server) {
			if (SSL_accept(s))
				fprintf(stderr, "ssl accept failed: %x\n",
					s->state);
			else
				fprintf(stderr, "ssl accept worked\n");
		} else {
			do {
				if (SSL_connect(s))
					fprintf(stderr,
						"ssl connect failed: %x\n",
						s->state);
			} while (s->state & 0x4000);
		}
		fprintf(stderr, "state=%x\n", s->state);
		async_fd(ns->ns_fd, 1);
#endif
		ssl->ssl_bio = bio;
		ns->ns_ssl = ssl;
	}
}


static int
verify_callback(ok, ctx)
	int ok;
	X509_STORE_CTX *ctx;
{
	char *s, buf[256];

	s = X509_NAME_oneline(X509_get_subject_name(ctx->current_cert),
			      buf, 256);
	if ((s != NULL) && (opts & OPT_VERBOSE)) {
		if (ok)
			fprintf(stderr,"depth=%d %s\n",ctx->error_depth,buf);
		else
			fprintf(stderr,"depth=%d error=%d %s\n",
				ctx->error_depth,ctx->error,buf);
	}

	if (ok == 0)
		switch (ctx->error)
		{
		case X509_V_ERR_CERT_NOT_YET_VALID:
		case X509_V_ERR_CERT_HAS_EXPIRED:
		case X509_V_ERR_DEPTH_ZERO_SELF_SIGNED_CERT:
			ok = 1;
			break;
		}

	return ok;
}

#ifndef NO_DH
static unsigned char dh512_p[] = {
	0xDA, 0x58, 0x3C, 0x16, 0xD9, 0x85, 0x22, 0x89, 0xD0, 0xE4, 0xAF, 0x75,
	0x6F, 0x4C, 0xCA, 0x92, 0xDD, 0x4B, 0xE5, 0x33, 0xB8, 0x04, 0xFB, 0x0F,
	0xED, 0x94, 0xEF, 0x9C, 0x8A, 0x44, 0x03, 0xED, 0x57, 0x46, 0x50, 0xD3,
	0x69, 0x99, 0xDB, 0x29, 0xD7, 0x76, 0x27, 0x6B, 0xA2, 0xD3, 0xD4, 0x12,
	0xE2, 0x18, 0xF4, 0xDD, 0x1E, 0x08, 0x4C, 0xF6, 0xD8, 0x00, 0x3E, 0x7C,
	0x47, 0x74, 0xE8, 0x33,
};
static unsigned char dh512_g[] = {
	0x02,
};

static DH *
get_dh512()
{
	DH *dh=NULL;

	dh = DH_new();
	if (dh == NULL)
		return NULL;
	dh->p = BN_bin2bn(dh512_p, sizeof(dh512_p), NULL);
	dh->g = BN_bin2bn(dh512_g, sizeof(dh512_g), NULL);
	if ((dh->p == NULL) || (dh->g == NULL))
		return NULL;
	return dh;
}
#endif

static RSA *
tmp_rsa_cb(s, export, keylength)
	SSL *s;
	int export;
	int keylength;
{
	static RSA *rsa_tmp = NULL;

	if (rsa_tmp == NULL) {
		if (opts & OPT_VERBOSE)
			fprintf(stderr,"Generating temp (%d bit) RSA key...",
				   keylength);
#ifndef NO_RSA
		rsa_tmp = RSA_generate_key(keylength, RSA_F4, NULL, NULL);
#endif
		if (opts & OPT_VERBOSE)
			fprintf(stderr,"\n");
	}
	return rsa_tmp;
}
#endif /* HAVE_SSL_H */
