#include "h.h"
#include "nsyslog.h"
#include "var.h"
#include "hash.h"


#if !defined(lint)
static const char rcsid[] = "$Id: nmask.c,v 1.6 1999/05/10 18:21:35 darrenr Exp $";
#endif


typedef struct nmaskspec {
	struct nmaskspec *nm_next;
	int nm_cnt;
	char *nm_parm;
	struct in_addr nm_addr;
	struct in_addr nm_mask;
	int nm_not;
} nmaskspec_t;


#define	OK_INET(src, nm)	((((sin_t *)src)->sin_addr.s_addr & \
				  (nm)->nm_mask.s_addr) == (nm)->nm_addr.s_addr)


int nmask_inccnt __P((void *, int));
void nmask_print __P((void *));
void nmask_expire __P((void));
int nmask_match __P((void *, logmsgtype_t *));

static	nmaskspec_t	*nmlist = NULL;
vartype_t	nmasktype = {
	NULL,		/* next */
	NULL,		/* ptr */
	FIL_WRAP,	/* stype */
	0,		/* dtype */
	"netmask",	/* name */
	NULL,		/* init */
	NULL,		/* read */
	NULL,		/* readnext */
	NULL,		/* write */
	NULL,		/* link */
	nmask_match,	/* match */
	NULL,		/* accept */
	NULL,		/* reopen */
	NULL,		/* logmsg */
	nmask_print,	/* print */
	NULL,		/* fd */
	nmask_inccnt,	/* inccnt */
	NULL,		/* drain */
	nmask_expire,	/* expire */
	NULL,		/* expireold */
	NULL,		/* more */
};


nmaskspec_t *
nmask_find(parm)
	char *parm;
{
	nmaskspec_t *nm;

	for (nm = nmlist; nm; nm = nm->nm_next)
		if (!strcmp(nm->nm_parm, parm))
			return nm;
	return NULL;
}


int
nmask_match(ptr, lmp)
	void *ptr;
	logmsgtype_t *lmp;
{
	nmaskspec_t *nm;

	if (VAR_TYPE(lmp->lm_ft->ft_type) != SRC_INET)
		return 1;
	nm = ptr;
	return OK_INET(lmp->lm_src, nm) ^ nm->nm_not;
}


nmaskspec_t *
nmask_new(parm)
	char *parm;
{
	nmaskspec_t *nm;
	char *s;

	if ((nm = nmask_find(parm))) {
		return nm;
	}

	if (!(NEW(nm, NEW_NETMS)))
		return NULL;
	nm->nm_parm = strdup(parm);
	if ((s = strchr(parm, '/'))) {
		*s++ = '\0';
		nm->nm_addr = getipv4addr(parm);
		nm->nm_mask = getipmask(s);
	} else {
		nm->nm_addr = getipv4addr(parm);
		nm->nm_mask.s_addr = inet_addr("255.255.255.255");
	}
	nm->nm_next = nmlist;
	nm->nm_not = get_not();
	nm->nm_cnt = 0;
	nmlist = nm;
	return nm;
}


void
nmask_expire()
{
	nmaskspec_t	*nm;

	while ((nm = nmlist)) {
		assert(nm->nm_cnt == 0);
		nmlist = nm->nm_next;
		free(nm->nm_parm);
		free(nm);
	}
}


int
nmask_inccnt(ptr, inc)
	void *ptr;
	int inc;
{
	nmaskspec_t *nm;

	nm = ptr;
	nm->nm_cnt += inc;
	return nm->nm_cnt;
}


void
nmask_print(ptr)
	void *ptr;
{
	nmaskspec_t *nm;

	nm = ptr;
	print_not(nm->nm_not);
	printf(" netmask %s/", inet_ntoa(nm->nm_addr));
	printf("%s ", inet_ntoa(nm->nm_mask));
}
