#include "h.h"
#include "nsyslog.h"
#include "var.h"
#include "y.tab.h"


#if !defined(lint)
static const char rcsid[] = "$Id: old.c,v 1.32 1999/04/20 04:39:16 darrenr Exp $";
#endif

extern int max_dfd, max_sfd;

static	void	cfline		__P((char *, var_t *));


int
check_comp(str)
	char **str;
{
	char *s = *str;

	if (!strncmp(s, "=", 1)) {
		*str = s + 1;
		return IL_EQ;
	}
	if (!strncmp(s, "!=", 2)) {
		*str = s + 2;
		return IL_NEQ;
	}
	if (!strncmp(s, "<=", 2)) {
		*str = s + 2;
		return IL_LE;
	}
	if (!strncmp(s, ">=", 2)) {
		*str = s + 2;
		return IL_GE;
	}
	if (!strncmp(s, "<", 1)) {
		*str = s + 1;
		return IL_LT;
	}
	if (!strncmp(s, ">", 2)) {
		*str = s + 1;
		return IL_GT;
	}
	return IL_GE;
}


/*
 * Read in a line from the old configuration file format and translate it
 * into the new format.
 */
static void
cfline(line, match)
	char *line;
	var_t *match;
{
	fillist_t *fl = NULL, *fn = NULL;
	char buf[MAXLINE], ebuf[100];
	char *bp, *p, *q, *t;
	var_t *s, *v, *f = NULL;
	int i = 0, pri, comp;

	if (opts & OPT_VERBOSE)
		fprintf(stderr, "cfline(%s)\n", line);

	errno = 0;	/* keep strerror() stuff out of logerror messages */

	/* scan through the list of selectors */
	for (p = line; *p && (*p != '\t') && (*p != ' ');) {

		/* find the end of this facility name list */
		for (q = p; *q && *q != '\t' && *p != ' ' && *q++ != '.'; )
			continue;

		/* collect priority name */
		for (bp = buf; *q && !strchr(" \t,;", *q); )
			*bp++ = *q++;
		*bp = '\0';

		/* skip cruft */
		while (strchr(", ;", *q))
			q++;

		t = buf;
		comp = IL_GE;
		/* decode priority name */
		if (*t == '*')
			pri = LOG_PRIMASK + 1;
		else {
			comp = check_comp(&t);
			pri = pri_findname(t);
			if (pri < 0) {
				snprintf(ebuf, sizeof(ebuf),
				    "unknown priority name \"%s\"", t);
				logerror(ebuf);
				return;
			}
		}

		/* scan facilities */
		i = -1;
		while (*p && !strchr("\t.;", *p)) {
			for (bp = t; *p && !strchr("\t,;.", *p); )
				*bp++ = *p++;
			*bp = '\0';
			if (*t != '*') {
				i = fac_findname(t);
				if (i < 0) {
					snprintf(ebuf, sizeof(ebuf),
					    "unknown facility name \"%s\"", t);
					logerror(ebuf);
					return;
				}
			}
			while (*p == ',' || *p == ' ')
				p++;
		}

		p = q;

		if ((i >= 0) || pri != (LOG_PRIMASK + 1)) {
			if (!(i >= 0))
				f = add_obj(FIL_PRI, prival_new(pri, comp));
			else if (pri == LOG_PRIMASK + 1)
				f = add_obj(FIL_FAC, facval_new(i));
			else
				f = add_obj(FIL_FACPRI,
					    facprival_new(i|pri, comp));
			end_obj();

			if (!fn && !(NEW(fn, NEW_FILLIST)))
				exit(1);
			if (fn->fl_fil) {
				if (!(NEW(fl, NEW_FILLIST)))
					exit(1);
				fl->fl_or = fn->fl_or;
				fn->fl_or = fl;
			} else
				fl = fn;
			f->v_cnt++;
			fl->fl_fil = f;
			fl = NULL;
			f = NULL;
		}

	}

	if (match) {
		if (fl) {
			if (!(NEW(fl->fl_and, NEW_FILLIST)))
				exit(1);
			fl->fl_and->fl_fil = match;
		} else {
			if (!(NEW(fl, NEW_FILLIST)))
				exit(1);
			fl->fl_fil = match;
			match->v_cnt++;
		}
		if (fn) {
			fl->fl_or = fn->fl_or;
			fn->fl_or = fl;
		} else
			fn = fl;
	}

	/* skip to action part */
	while (*p == '\t')
		p++;

	/*
	 * Create a new destination object and give it the correct type
	 * for the member.
	 */
	switch (*p)
	{
	case '@':
		set_raw();
		v = add_obj(DEST_INET, net_new(DEST_INET, NET_UDP,
					       p + 1, "514"));
		break;

	case '/':
		v = add_obj(DEST_FILE, file_new(DEST_FILE, p));
		break;

	case '|':
		v = add_obj(DEST_CMD, cmd_new(p + 1));
		break;

	case '*':
	default:
		while (*p) {
			char ch;

			for (q = p; *q && *q != ','; )
				q++;
			ch = *q;
			*q = '\0';

			v = add_obj(DEST_USER, user_new(p));

			*q = ch;
			while (*q == ',' || *q == ' ')
				q++;
			p = q;
		}
		break;
	}
	end_obj();

	s = get_var("syslog");

	link_sfd(s, fn, v);
}


#ifdef	BIN_M4
int
am_i_loghost()
{
	struct hostent *hp;
	struct in_addr *ips;
	int i, cnt;

	hp = gethostbyname("loghost");
	if (!hp)	/* if no loghost, then I am loghost */
		return 1;

	for (i = 0; hp->h_addr_list[i]; i++)
		;

	ips = (struct in_addr *)malloc(sizeof(*ips) * (i + 1));
	if (ips == NULL)
		return 0;
	for (i = 0; hp->h_addr_list[i]; i++)
		bcopy(hp->h_addr_list[i], (char *)&ips[i], sizeof(ips[i]));

	cnt = i;
	for (i = 0; i < cnt; i++) {
		if (((ntohl(ips[i].s_addr) & 0xff000000) >> 24) ==
		    IN_LOOPBACKNET) {
			free(ips);
			return 1;
		}
	}
	free(ips);
	return 0;
}
#endif


/*
 * read in the configuration file and setup for running.
 */
int
read_oldconfig(file, rfdp)
	char *file;
	fd_set *rfdp;
{
	char cline[512], *s, prog[257], *mb = NULL;
	var_t *m = NULL;
	struct stat sb;
	FILE *fp;
	int i;

	if (opts & OPT_DEBUG)
		fprintf(stderr, "Reading old-configuration from %s\n", file);

	FD_ZERO(rfdp);

	/*
	 * The old syslogd has a number of sources but they are all considered
	 * the same so far as which destinations they may go to.
	 */
	init_internal();
	(void) add_obj(SRC_FILE, file_new(SRC_FILE, "/dev/klog"));
#if SOLARIS
	(void) add_obj(SRC_STREAM, stream_new("/dev/log"));
	(void) add_obj(SRC_DOOR, door_new("/etc/.syslog_door"));
#else
	(void) add_obj(SRC_UNIX, unix_new(SRC_UNIX, SOCK_DGRAM, "/dev/log"));
	/*
	 * If there is a /var/run directory, then it is possible that the
	 * syslog library is expecting a /var/run/log.  Create a symbolic
	 * link to /dev/log for this.
	 */
	if ((stat("/var/run", &sb) == 0) && S_ISDIR(sb.st_mode)) {
		unlink("/var/run/log");
		symlink("/var/run/log", "/dev/log");
	}
#endif
	if ((opts & OPT_OLDNOUDP) == 0)
		(void) add_obj(SRC_INET, net_new(SRC_INET, NET_UDP, "0.0.0.0",
				"514"));
	end_obj();

#ifdef	BIN_M4
	snprintf(prog, sizeof(prog), "echo '%s' | %s - %s",
		am_i_loghost() ? "define(LOGHOST, 1)" : "", BIN_M4, file);

	if (!(fp = popen(prog, "r")))
		return -1;
#else
	if (!(fp = fopen(file, "r")))
		return -1;
#endif

	*prog = '\0';

	while (fgets(cline, sizeof(cline), fp)) {
		for (s = cline; *s && isspace(*s); s++)
			;
		if (!*s)
			continue;
		if (*s == '#') {
			s++;
			if (*s != '!')
				continue;
		}
		if (*s == '!') {
			s++;
			if (mb)
				free(mb);
			while (isspace(*s))
				s++;
			if (!*s)
				*prog = '\0';
			else {
				for (i = 0; i < (sizeof(prog) - 1); i++) {
					if (!isalnum(s[i]))
						break;
					prog[i] = s[i];
				}
				prog[i] = 0;
			}
			m = add_obj(FIL_MATCHPROG, matchprog_new(prog));
			end_obj();
			continue;
		}
		for (s = strchr(cline, '\0'); isspace(*--s);)
			continue;
		*++s = '\0';
		cfline(cline, m);
	}

	fclose(fp);

	/*
	 * Redisplay the configuration file as we see it.
	 */
	if (opts & (OPT_DEBUG))
		dump_all();


	return setup_filetable(rfdp);

}
