#include "h.h"
#include "nsyslog.h"
#include "var.h"


#if !defined(lint)
static const char rcsid[] = "$Id: pipe.c,v 1.7 1999/05/10 18:18:44 darrenr Exp $";
#endif


typedef struct cmdspec {
	struct cmdspec *cm_next;
	int cm_cnt;
	int cm_mode;
	int cm_fd;
	char *cm_path;
} cmdspec_t;


void cmd_reopen __P((void *));
int cmd_logmsg __P((void *, int *, struct iovec *, int, logmsgtype_t *));
int cmd_fd __P((void *));
int cmd_inccnt __P((void *, int));
void cmd_print __P((void *));
void cmd_expire __P((void));

static	cmdspec_t	*cslist = NULL;
vartype_t	cmdtype = {
	NULL,		/* next */
	NULL,		/* ptr */
	0,		/* stype */
	DEST_CMD,	/* dtype */
	"command",	/* name */
	NULL,		/* init */
	NULL,		/* read */
	NULL,		/* readnext */
	NULL,		/* write */
	NULL,		/* link */
	NULL,		/* match */
	NULL,		/* accept */
	cmd_reopen,	/* reopen */
	cmd_logmsg,	/* logmsg */
	cmd_print,	/* print */
	cmd_fd,		/* fd */
	cmd_inccnt,	/* inccnt */
	NULL,		/* drain */
	cmd_expire,	/* expire */
	NULL,		/* expireold */
	NULL,		/* more */
};


cmdspec_t *
cmd_find(parm)
	char *parm;
{
	cmdspec_t *c;

	for (c = cslist; c; c = c->cm_next)
		if (!strcmp(c->cm_path, parm))
			return c;
	return NULL;
}


int
pipe_new(path)
	char *path;
{
	FILE *fp;
	int fd = -1;

	if ((fp = popen(path, "r+"))) {
		fd = fileno(fp);
		fcntl(fd, F_SETFD, fcntl(fd, F_GETFD, 0) | FD_CLOEXEC);
	}
	return fd;
}


cmdspec_t *
cmd_new(parm)
	char *parm;
{
	cmdspec_t *c;

	if ((c = cmd_find(parm))) {
		return c;
	}

	if (!(NEW(c, NEW_CMD)))
		return NULL;
	c->cm_cnt = 0;
	c->cm_path = strdup(parm);
	c->cm_next = cslist;
	c->cm_fd = pipe_new(parm);
	cslist = c;
	return c;
}


void
cmd_expire()
{
	cmdspec_t *cs;

	while ((cs = cslist)) {
		assert(cs->cm_cnt == 0);
		cslist = cs->cm_next;
		free(cs->cm_path);
		free(cs);
	}
}


int
cmd_logmsg(ptr, fdp, iov, nio, lmp)
	void *ptr;
	int *fdp;
	struct iovec *iov;
	int nio;
	logmsgtype_t *lmp;
{
	cmdspec_t *cs;
	int fd, nw;

	cs = ptr;
	fd = cs->cm_fd;
	*fdp = fd;
	nw = 0;

	if (opts & OPT_VERBOSE)
		fprintf(stderr, "msg to command %s\n", cs->cm_path);

	if (fd >= 0)
		nw = writev(fd, iov, nio);
	if (nw < 0)
		syserr("error writing to %s: %m", cs->cm_path);
	return nw;
}


void
cmd_reopen(ptr)
	void *ptr;
{
	cmdspec_t *cs;

	cs = ptr;
	cs->cm_fd = pipe_new(cs->cm_path);
}


int
cmd_inccnt(ptr, inc)
	void *ptr;
	int inc;
{
	cmdspec_t *cs;

	cs = ptr;
	cs->cm_cnt += inc;
	return cs->cm_cnt;
}


int
cmd_fd(ptr)
	void *ptr;
{
	cmdspec_t *cs;

	cs = ptr;
	return cs->cm_fd;
}


void
cmd_print(ptr)
	void *ptr;
{
	cmdspec_t *cs;

	cs = ptr;
	printf(" command %s ", cs->cm_path);
}
