#include "h.h"
#include "nsyslog.h"
#include "var.h"
#include "hash.h"
#include "y.tab.h"


#if !defined(lint)
static const char rcsid[] = "$Id: pri.c,v 1.7 1999/05/10 18:18:43 darrenr Exp $";
#endif


typedef struct prispec {
	struct prispec *ps_next;
	int ps_cnt;
	int ps_pri;
	int ps_not;
	int ps_comp;
} prispec_t;


prispec_t	*prilist = NULL;
void pri_print __P((void *));
int pri_inccnt __P((void *, int ));
void pri_expire __P((void));
int pri_match __P((void *, logmsgtype_t *));

vartype_t	pritype = {
	NULL,		/* next */
	NULL,		/* ptr */
	FIL_PRI,	/* stype */
	0,		/* dtype */
	"pri",		/* name */
	NULL,		/* init */
	NULL,		/* read */
	NULL,		/* readnext */
	NULL,		/* write */
	NULL,		/* link */
	pri_match,	/* match */
	NULL,		/* accept */
	NULL,		/* reopen */
	NULL,		/* logmsg */
	pri_print,	/* print */
	NULL,		/* fd */
	pri_inccnt,	/* inccnt */
	NULL,		/* drain */
	pri_expire,	/* expire */
	NULL,	/* expireold */
	NULL,		/* more */
};


table_t	pris[] = {
	{ "emerg", LOG_EMERG, '0'},	{ "alert", LOG_ALERT, '1' },
	{ "crit", LOG_CRIT, '2'},	{ "err", LOG_ERR, '3' },
	{ "warn", LOG_WARNING, '4'},	{ "notice", LOG_NOTICE, '5' },
	{ "info", LOG_INFO, '6' },	{ "debug", LOG_DEBUG, '7' },
	{ "none", NOPRI },		{ NULL, 0 }
};


table_t *
pri_find(pri)
	int pri;
{
	int	i;

	for (i = 0; pris[i].name; i++)
		if (pri == pris[i].value)
			return pris + i;
	return NULL;
}


/*
 * map a priority name to its number
 */
int
pri_findname(name)
	char *name;
{
	int	i;

	for (i = 0; pris[i].name; i++)
		if (!strcmp(pris[i].name, name))
			return pris[i].value;
	return -1;
}


/*
 * map a priority number to its name
 */
char *
pri_toname(facpri)
	int facpri;
{
	int	i, pri;

	for (i = 0, pri = facpri & LOG_PRIMASK; pris[i].name; i++)
		if (pri == pris[i].value)
			return pris[i].name;
	return NULL;
}


prispec_t *
pri_finduse(pri, comp, not)
	int pri;
	int comp;
	int not;
{
	prispec_t *ps;

	for (ps = prilist; ps; ps = ps->ps_next)
		if (ps->ps_pri == pri && ps->ps_comp == comp &&
		    ps->ps_not == not)
			return ps;
	return NULL;
}


prispec_t *
pri_new(comp, parm)
	int comp;
	char *parm;
{
	prispec_t *ps;
	int pri, not;

	not = get_not();

	if ((pri = pri_findname(parm)) < 0) {
		fprintf(stderr, "unknown priority %s\n", parm);
		return NULL;
	}

	if ((ps = pri_finduse(comp, pri, not)))
		return ps;

	if (!(NEW(ps, NEW_PRIS)))
		return NULL;
	ps->ps_pri = pri;
	ps->ps_next = prilist;
	ps->ps_not = not;
	ps->ps_comp = comp;
	ps->ps_cnt = 0;
	prilist = ps;
	return ps;
}


/*
 * return 1 for match, 0 for non-match
 */
int
pri_matchvals(p1, p2, comp)
	int p1, p2, comp;
{
	p1 &= LOG_PRIMASK;
	p2 &= LOG_PRIMASK;

	switch (comp)
	{
	case IL_EQ :
		return (p1 == p2);
	case IL_NEQ :
		return (p1 != p2);
	case IL_LT :
		return (p1 > p2);
	case IL_LE :
		return (p1 >= p2);
	case IL_GT :
		return (p1 < p2);
	case IL_GE :
		return (p1 <= p2);
	}
	syserr("unknown comparator: %d\n", comp);
	return -1;
}


int
pri_match(ptr, lmp)
	void *ptr;
	logmsgtype_t *lmp;
{
	prispec_t *ps;
	int ok;

	ps = ptr;
	ok = pri_matchvals(lmp->lm_pri, ps->ps_pri, ps->ps_comp);
	ok ^= ps->ps_not;
	return !ok;
}


prispec_t *
prival_new(pri, comp)
	int pri, comp;
{
	prispec_t *ps;
	int not;

	not = get_not();

	if ((ps = pri_finduse(pri, 0, not)))
		return ps;

	if (!(NEW(ps, NEW_PRIS)))
		return NULL;
	ps->ps_pri = pri;
	ps->ps_next = prilist;
	ps->ps_cnt = 0;
	ps->ps_not = not;
	ps->ps_comp = comp;
	prilist = ps;
	return ps;
}


void
pri_expire()
{
	prispec_t	*pri;

	while ((pri = prilist)) {
		assert(pri->ps_cnt == 0);
		prilist = pri->ps_next;
		free(pri);
	}
}


void
pri_print(ptr)
	void *ptr;
{
	prispec_t *ps;

	ps = ptr;
	print_not(ps->ps_not);
	printf(" priority %s%s", comp_toname(ps->ps_comp),
		pri_toname(ps->ps_pri));
}


int
pri_inccnt(ptr, inc)
	void *ptr;
	int inc;
{
	prispec_t *ps;

	ps = ptr;
	ps->ps_cnt += inc;
	return ps->ps_cnt;
}
