#include "h.h"
#include "nsyslog.h"
#include "var.h"
#include "y.tab.h"

#if defined(HAVE_GETMSG) && defined(HAVE_SYS_STRLOG_H)

#include <stropts.h>
#include <sys/strlog.h>


typedef struct streamspec {
	struct streamspec *ss_next;
	int ss_cnt;
	int ss_fd;
	int ss_mode;
	int ss_pri;
	time_t ss_when;
	buflist_t *ss_rlist;
	buflist_t **ss_rtail;
	char *ss_path;
} streamspec_t;


extern	int	do_raw;
streamspec_t	*sslist = NULL;
void *stream_vinit __P((int , arglist_t *));
int stream_read __P((void *, logmsgtype_t *, tcplmsg_t *, char *, size_t));
int stream_readnext __P((void *, logmsgtype_t *, tcplmsg_t *, char *, size_t));
void stream_reopen __P((void *));
void stream_print __P((void *));
int stream_fd __P((void *));
int stream_inccnt __P((void *, int));
void stream_expire __P((void));
void stream_expireold __P((void));
int stream_more __P((void *));

static	streamspec_t	*cur_stream = NULL;
vartype_t	streamtype = {
	NULL,		/* next */
	NULL,		/* ptr */
	SRC_STREAM,	/* stype */
	0,		/* dtype */
	"stream",	/* name */
	stream_vinit,	/* init */
	stream_read,	/* read */
	stream_readnext,/* readnext */
	NULL,		/* write */
	NULL,		/* link */
	NULL,		/* match */
	NULL,		/* accept */
	stream_reopen,	/* reopen */
	NULL,		/* logmsg */
	stream_print,	/* print */
	stream_fd,	/* fd */
	stream_inccnt,	/* inccnt */
	NULL,		/* drain */
	stream_expire,	/* expire */
	stream_expireold,	/* expireold */
	stream_more,		/* more */
};



extern	streamspec_t	*stream_new __P((char *));


void *
stream_vinit(type, args)
	int type;
	arglist_t *args;
{
	int proto;
	char *p1;

	if (args == NULL)
		return NULL;
	p1 = args->a_arg;
	if (args->a_next != NULL)
		return NULL;
	return stream_new(p1);
}


streamspec_t *
stream_find(parm)
	char *parm;
{
	streamspec_t *ss;

	for (ss = sslist; ss; ss = ss->ss_next)
		if (!strcmp(ss->ss_path, parm))
			return ss;
	return NULL;
}


streamspec_t *
stream_new(parm)
	char *parm;
{
	struct strioctl ioc;
	streamspec_t *ss;
	int fd;

	if ((ss = stream_find(parm))) {
		ss->ss_mode &= ~VAR_OLD;
		return ss;
	}

	if (!(NEW(ss, NEW_STREAMS)))
		return NULL;
	ss->ss_rtail = &ss->ss_rlist;

	if (do_raw) {
		ss->ss_mode |= do_raw;
		do_raw = 0;
	}
	ss->ss_next = sslist;
	ss->ss_path = strdup(parm);
	fd = open(parm, O_RDONLY);
	if (fd == -1)
		logerror("error opening streams stream");
	else {
		async_fd(fd, 1);
		fcntl(fd, F_SETFD, fcntl(fd, F_GETFD, 0) | FD_CLOEXEC);

		bzero(&ioc, sizeof(ioc));
		ioc.ic_cmd = I_CONSLOG;
		if (ioctl(fd, I_STR, &ioc) == -1) {
			logerror("cannot enable console logging");
			close(fd);
			fd = -1;
		} else
			ft_set(fd, ss, SRC_STREAM);
	}
	ss->ss_fd = fd;
	ss->ss_cnt = 0;
	sslist = ss;
	return ss;
}


void
stream_free(ptr)
	void *ptr;
{
	streamspec_t *s, **sp, *ss;
	buflist_t *b;

	ss = ptr;
	if (ss->ss_cnt > 0)
		return;

	assert(ss->ss_cnt == 0);
	for (sp = &sslist; (s = *sp); )
		if (s == ss) {
			*sp = s->ss_next;
			break;
		} else
			sp = &s->ss_next;
	close(ss->ss_fd);
	buf_freelist(ss->ss_rlist);
	free(ss->ss_path);
	free(ss);
}


void
stream_expire()
{
	streamspec_t	*ss, **ssp;

	for (ssp = &sslist; (ss = *ssp); ) {
		assert(ss->ss_cnt == 0);
		ss->ss_mode |= VAR_OLD;
		ssp = &ss->ss_next;
		if (opts & OPT_VERBOSE)
			fprintf(stderr, "keep stream: %s (%d)\n",
				ss->ss_path, ss->ss_cnt);
	}
}


void
stream_expireold()
{
	streamspec_t	*ss, **ssp;

	for (ssp = &sslist; (ss = *ssp); ) {
		if (!(ss->ss_mode & VAR_OLD)) {
			ssp = &ss->ss_next;
			continue;
		}
		if (opts & OPT_VERBOSE)
			fprintf(stderr, "close stream: %s (%d)\n",
				ss->ss_path, ss->ss_cnt);
		ssp = &sslist;
		stream_free(ss);
	}
}


void
stream_reopen(ptr)
	void *ptr;
{
	streamspec_t *ss;

	ss = ptr;
	return;
}


static int
stream_recv(ss)
	streamspec_t *ss;
{
	struct strbuf ctl, data;
	int len, error, flags;
	struct log_ctl log;
	buflist_t *bl;

	if (ioctl(ss->ss_fd, FIONREAD, &len) == -1) {
		syserr("ioctl(%d[%s],FIONREAD) error, closing:%m", ss->ss_fd,
			ss->ss_path);
		return -1;
	}
	if (len == 0)
		return 0;
	if (len <= 0) {
		syserr("ioctl error from fd %d[%s], closing:%m", ss->ss_fd,
			ss->ss_path);
		return -1;
	}

	ctl.maxlen = sizeof(log);
	ctl.buf = (char *)&log;
	ctl.len = sizeof(log);

	bl = buf_new(len);
	*ss->ss_rtail = bl;
	ss->ss_rtail = &bl->bl_next;

	data.maxlen = len;
	data.buf = bl->bl_ptr;
	data.len = 0;

	flags = 0;

	error = getmsg(ss->ss_fd, &ctl, NULL, &data, &flags);

	if (error == -1 && (errno == EAGAIN || errno == EWOULDBLOCK ||
			    errno == ENOBUFS || errno == EINTR))
		return 0;
	if (error == -1) {
		syserr("getmsg(stream %s fd%d):%m", ss->ss_path, ss->ss_fd);
		return -1;
	}
	ss->ss_when = log.ttime;
	ss->ss_pri = log.pri;
	while (data.buf[data.len - 1] == '\0')
		data.len--;
	bl->bl_rem += data.len;
	return data.len;
}

int
stream_read(ptr, lmp, tlm, buf, bufsz)
	void *ptr;
	logmsgtype_t *lmp;
	tcplmsg_t *tlm;
	char *buf;
	size_t bufsz;
{
	streamspec_t *ss;
	int len = 0, i;

	ss = ptr;
	len = stream_readnext(ptr, lmp, tlm, buf, bufsz);
	if (len > 0)
		return len;

	switch (stream_recv(ss))
	{
	case -1 :
		return -1;
	case 0 :
		return 0;
	default :
		tlm->tl_pri = ss->ss_pri;
		lmp->lm_lpri = ss->ss_pri;
		tlm->tl_tv = ss->ss_when;
		lmp->lm_lwhen = ss->ss_when;
		lmp->lm_flags = FLG_HAVEPRI;
		len = buf_getmsg(&ss->ss_rlist, NULL, NULL, buf, bufsz);
		if (ss->ss_rlist == NULL)
			ss->ss_rtail = &ss->ss_rlist;
		break;
	}
	return len;
}


int
stream_readnext(ptr, lmp, tlm, buf, bufsz)
	void *ptr;
	logmsgtype_t *lmp;
	tcplmsg_t *tlm;
	char *buf;
	size_t bufsz;
{
	streamspec_t *ss;
	int len = 0;

	ss = ptr;
	if (ss->ss_rlist != NULL) {
		tlm->tl_pri = ss->ss_pri;
		lmp->lm_lpri = ss->ss_pri;
		tlm->tl_tv = ss->ss_when;
		lmp->lm_lwhen = ss->ss_when;
		lmp->lm_flags = FLG_HAVEPRI;
		len = buf_getmsg(&ss->ss_rlist, NULL, NULL, buf, bufsz);
		if (ss->ss_rlist == NULL)
			ss->ss_rtail = &ss->ss_rlist;
	}
	return len;
}


int
stream_more(ptr)
	void *ptr;
{
	streamspec_t *ss;

	ss = ptr;
	return buf_size(ss->ss_rlist);
}


int
stream_inccnt(ptr, inc)
	void *ptr;
	int inc;
{
	streamspec_t *ss;

	ss = ptr;
	ss->ss_cnt += inc;
	return ss->ss_cnt;
}


int
stream_fd(ptr)
	void *ptr;
{
	streamspec_t *ss;

	ss = ptr;
	return ss->ss_fd;
}


void
stream_print(ptr)
	void *ptr;
{
	streamspec_t *ss;

	ss = ptr;
	printf(" stream %s ", ss->ss_path);
}

#endif	/* HAVE_GETMSG */	
