#include "h.h"
#include "nsyslog.h"
#include "var.h"
#include "y.tab.h"

void *unix_vinit __P((int, arglist_t *));
void unix_link __P((void *, void *));
void unix_expire __P((void));
void unix_expireold __P((void));
int unix_read __P((void *, logmsgtype_t *, tcplmsg_t *, char *, size_t));
int unix_readnext __P((void *, logmsgtype_t *, tcplmsg_t *, char *, size_t));
void unix_reopen __P((void *));
void unix_print __P((void *));
int unix_fd __P((void *));
int unix_inccnt __P((void *, int));
int unix_logmsg __P((void *, int *, struct iovec *, int, logmsgtype_t *));
int unix_initsrc __P((int));
int unix_more __P((void *));

static	unixspec_t	*xslist = NULL;
vartype_t	unixtype = {
	NULL,		/* next */
	NULL,		/* ptr */
	SRC_UNIX,	/* stype */
	DEST_UNIX,	/* dtype */
	"unix",		/* name */
	unix_vinit,	/* init */
	unix_read,	/* read */
	unix_readnext,	/* readnext */
	NULL,		/* write */
	NULL,		/* link */
	NULL,		/* match */
	unix_initsrc,	/* accept */
	unix_reopen,	/* reopen */
	unix_logmsg,	/* logmsg */
	unix_print,	/* print */
	unix_fd,	/* fd */
	unix_inccnt,	/* inccnt */
	NULL,		/* drain */
	unix_expire,	/* expire */
	unix_expireold,	/* expireold */
	unix_more,	/* more */
};


extern	int	do_raw;


void *
unix_vinit(type, args)
	int type;
	arglist_t *args;
{
	int proto;
	char *p1;

	if (args == NULL)
		return NULL;
	proto = SOCK_DGRAM;
	p1 = args->a_arg;
	if (!strcasecmp(p1, "stream")) {
		proto = SOCK_STREAM;
		args = args->a_next;
		if (args == NULL)
			return NULL;
		p1 = args->a_arg;
	}
	if (args->a_next)
		return NULL;
	return unix_new(type, proto, p1);
}


void
unix_link(ptr1, ptr2)
	void *ptr1;
	void *ptr2;
{
	unixspec_t *xsp = ptr1;
	unixspec_t *xsc = ptr2;

	xsc->xs_parent = xsp;
}

unixspec_t *
unix_find(proto, parm)
	int proto;
	char *parm;
{
	unixspec_t *xs;

	for (xs = xslist; xs; xs = xs->xs_next)
		if ((xs->xs_type == proto) &&
		    !strncmp(xs->xs_un.sun_path, parm,
			     sizeof(xs->xs_un.sun_path)))
			return xs;
	return NULL;
}


unixspec_t *
unix_new(type, proto, parm)
	int type;
	int proto;
	char *parm;
{
	unixspec_t *xs;

	if (type == SRC_UNIX && proto == IL_STREAM)
		type = SRC_UNIX_LISTEN|VAR_LISTEN;
	if ((xs = unix_find(type, parm))) {
		xs->xs_mode &= ~VAR_OLD;
		ft_set(xs->xs_fd, xs, type);
		return xs;
	}

	if (!(NEW(xs, NEW_UNIXS)))
		return NULL;
	xs->xs_wtail = &xs->xs_wlist;
	xs->xs_rtail = &xs->xs_rlist;
	xs->xs_type = type;
	xs->xs_proto = proto;

	if (do_raw) {
		xs->xs_mode |= do_raw;
		do_raw = 0;
	}
	xs->xs_next = xslist;
	xs->xs_un.sun_family = AF_UNIX;
	strncpy(xs->xs_un.sun_path, parm, sizeof(xs->xs_un.sun_path));
	xs->xs_un.sun_path[sizeof(xs->xs_un.sun_path) - 1] = '\0';
	xs->xs_fd = new_socket(type, xs->xs_proto, (sockaddr_t *)&xs->xs_un,
			       sizeof(xs->xs_un));
	if (xs->xs_fd > 0)
		ft_set(xs->xs_fd, xs, type);
	xs->xs_cnt = 0;
	xslist = xs;
	return xs;
}


void
unix_free(xs)
	unixspec_t *xs;
{
	unixspec_t *x, **xp;
	buflist_t *b;

	if (xs->xs_cnt > 0)
		return;

	assert(xs->xs_cnt == 0);
	if (VAR_TYPE(xs->xs_type) == SRC_UNIX_LISTEN) {
		for (xp = &xslist; (x = *xp); )
			if (x->xs_parent == xs) {
				*xp = x->xs_next;
				unix_free(x);
			} else
				xp = &x->xs_next;
	}
	for (xp = &xslist; (x = *xp); )
		if (x == xs) {
			*xp = x->xs_next;
			break;
		} else
			xp = &x->xs_next;
	if (VAR_MODE(xs->xs_type) == SRC_NONE)
		(void) unlink(xs->xs_un.sun_path);
	close(xs->xs_fd);
	buf_freelist(xs->xs_rlist);
	buf_freelist(xs->xs_wlist);
	free(xs);
}


void
unix_expire()
{
	unixspec_t	*xs, **xsp;

	for (xsp = &xslist; (xs = *xsp); ) {
		assert(xs->xs_cnt == 0);
		if (VAR_TYPE(xs->xs_type) == SRC_UNIX) {
			xs->xs_mode |= VAR_OLD;
			xsp = &xs->xs_next;
			if (opts & OPT_VERBOSE)
				fprintf(stderr, "keep unix: %s (%d)\n",
					xs->xs_un.sun_path, xs->xs_cnt);
			continue;
		}
		if (opts & OPT_VERBOSE)
			fprintf(stderr, "close unix: %s (%d)\n",
				xs->xs_un.sun_path, xs->xs_cnt);
		xsp = &xslist;
		unix_free(xs);
	}
}


void
unix_expireold()
{
	unixspec_t	*xs, **xsp;

	for (xsp = &xslist; (xs = *xsp); ) {
		if (!(xs->xs_mode & VAR_OLD)) {
			xsp = &xs->xs_next;
			continue;
		}
		if (opts & OPT_VERBOSE)
			fprintf(stderr, "close unix: %s (%d)\n",
				xs->xs_un.sun_path, xs->xs_cnt);
		xsp = &xslist;
		unix_free(xs);
	}
}


void
unix_drain(ptr)
	void *ptr;
{
	unixspec_t *xs;

	return;
}


int
unix_logmsg(ptr, fdp, iov, nio, lmp)
	void *ptr;
	int *fdp;
	struct iovec *iov;
	int nio;
	logmsgtype_t *lmp;
{
	struct iovec liov[MAX_NIOV];
	int fd, mode, nw;
	unixspec_t *xs;
	char pri[18];

	xs = ptr;
	mode = xs->xs_mode;
	fd = xs->xs_fd;
	*fdp = fd;
	nw = 0;

	if (opts & OPT_VERBOSE)
		fprintf(stderr, "%smsg to unix %s\n",
			(mode & VAR_RAW) ? "raw " : "", xs->xs_un.sun_path);

	/*
	 * If a destiation is marked as "raw", then send it with the
	 * priority information intact and without the hostname. 
	 */
	if (!(mode & VAR_RAW)) {
		if (fd >= 0)
			nw = writev(fd, iov, nio);
	} else {
		bcopy((char *)iov, (char *)liov, sizeof(liov));
		liov[nio - 3].iov_len = liov[0].iov_len;
		liov[nio - 3].iov_base = liov[0].iov_base;
		snprintf(pri, sizeof(pri) - 1, "<%d>",
			 lmp->lm_pri & (LOG_FACMASK|LOG_PRIMASK));
		liov[nio - 4].iov_base = pri;
		liov[nio - 4].iov_len = strlen(pri);
		if (fd >= 0)
			nw = writev(fd, liov + (nio - 4), 4);
	}
	if (nw < 0)
		syserr("error writing to %s: %m", xs->xs_un.sun_path);
	return nw;
}


void
unix_reopen(ptr)
	void *ptr;
{
	unixspec_t *xs;

	xs->xs_fd = new_socket(xs->xs_type, xs->xs_proto,
			       (sockaddr_t *)&xs->xs_un, sizeof(xs->xs_un));
}


int
unix_recv(ptr, unp, ulen)
	void *ptr;
	struct sockaddr *unp;
	int *ulen;
{
	int len, error;
	unixspec_t *xs;
	buflist_t *bl;

	xs = ptr;
	if (ioctl(xs->xs_fd, FIONREAD, &len) == -1) {
		syserr("ioctl(%d[%s],FIONREAD) error, closing:%m", xs->xs_fd,
			xs->xs_un.sun_path);
		return -1;
	}
	if (len <= 0) {
		syserr("ioctl error from fd %d[%s], closing:%m", xs->xs_fd,
			xs->xs_un.sun_path);
		return -1;
	}

	bl = buf_new(len);
	*xs->xs_rtail = bl;
	xs->xs_rtail = &bl->bl_next;

	error = recvfrom(xs->xs_fd, bl->bl_ptr, len, 0, unp, ulen);
	if (error == -1 && (errno == EAGAIN || errno == EWOULDBLOCK ||
			    errno == ENOBUFS || errno == EINTR))
		return 0;
	if (error <= 0)
		return -1;
	if (*ulen > 0) {
		bl->bl_src = (void *)malloc(*ulen);
		bl->bl_slen = *ulen;
		bcopy ((char *)ulen, bl->bl_src, *ulen);
	}
	bl->bl_rem += error;
	return error;
}


int
unix_read(ptr, lmp, tlm, buf, bufsz)
	void *ptr;
	logmsgtype_t *lmp;
	tcplmsg_t *tlm;
	char *buf;
	size_t bufsz;
{
	unixspec_t *xs;
	int len = 0, i;
	buflist_t *b;

	xs = ptr;
	if (xs->xs_proto == SOCK_STREAM) {
		switch (unix_recv(xs, lmp->lm_src, &lmp->lm_slen))
		{
		case -1 :
			return -1;
		case 0 :
			return 0;
		default :
			len = unix_readnext(ptr, lmp, tlm, buf, bufsz);
			break;
		}
	} else if (xs->xs_proto == SOCK_DGRAM) {
		len = recvfrom(xs->xs_fd, buf, bufsz, 0,
			       lmp->lm_src, &lmp->lm_slen);
		if (len < 0)
			syserr("recvfrom(unix fd%d, %s):%m", xs->xs_fd,
			       ((struct sockaddr_un *)lmp->lm_src)->sun_path);
	}
	lmp->lm_flags = FLG_USER;
	return len;
}


int
unix_readnext(ptr, lmp, tlm, buf, bufsz)
	void *ptr;
	logmsgtype_t *lmp;
	tcplmsg_t *tlm;
	char *buf;
	size_t bufsz;
{
	unixspec_t *xs;
	buflist_t *b;
	int len = 0;

	xs = ptr;
	b = xs->xs_rlist;
	if (b->bl_slen > 0)
		bcopy(b->bl_src, lmp->lm_src,
		      MIN(b->bl_slen, lmp->lm_slen));
	len = buf_getmsg(&xs->xs_rlist, NULL, NULL, buf, bufsz);
	if (len > 0) {
		if (len < bufsz)
			buf[len] = '\0';
		else
			buf[bufsz - 1] = '\0';
	}
	lmp->lm_flags = FLG_USER;
	return len;
}


int
unix_initsrc(lfd)
	int lfd;
{
	unixspec_t *xs;
	int fd, len;

	xs = (unixspec_t *)net_new(SRC_UNIX, NET_CLONE, NULL, NULL);
	if (xs == NULL)
		return -1;

	len = sizeof(xs->xs_un);
	fd = accept(lfd, (struct sockaddr *)&xs->xs_un, &len);
	if (fd == -1) {
		unix_free(xs);
		return -1;
	}

	xs->xs_fd = fd;
	xs->xs_proto = SOCK_STREAM;

	fttable[fd].ft_type = SRC_UNIX;
	fttable[fd].ft_ptr = xs;
	return fd;
}	


int
unix_inccnt(ptr, inc)
	void *ptr;
	int inc;
{
	unixspec_t *xs;

	xs = ptr;
	xs->xs_cnt += inc;
	return xs->xs_cnt;
}


int
unix_fd(ptr)
	void *ptr;
{
	unixspec_t *xs;

	xs = ptr;
	return xs->xs_fd;
}


void
unix_print(ptr)
	void *ptr;
{
	unixspec_t *xs;

	xs = ptr;
	printf(" unix %s ", xs->xs_un.sun_path);
}


int
unix_more(ptr)
	void *ptr;
{
	unixspec_t *xs;

	return buf_size(xs->xs_rlist);
}
