#include "h.h"
#include "nsyslog.h"
#include "var.h"
#include "hash.h"
#include "y.tab.h"


#if !defined(lint)
static const char rcsid[] = "$Id: user.c,v 1.7 1999/05/10 18:18:45 darrenr Exp $";
#endif


typedef struct userspec {
	struct userspec *us_next;
	int us_cnt;
	char us_user[9];
} userspec_t;


int user_logmsg __P((void *, int *, struct iovec *, int, logmsgtype_t *));
void user_expire __P((void));
void user_print __P((void *));
int user_inccnt __P((void *, int));


static	userspec_t	*uslist = NULL;
vartype_t	usertype = {
	NULL,		/* next */
	NULL,		/* ptr */
	0,		/* stype */
	DEST_USER,	/* dtype */
	"user",		/* name */
	NULL,		/* init */
	NULL,		/* read */
	NULL,		/* readnext */
	NULL,		/* write */
	NULL,		/* link */
	NULL,		/* match */
	NULL,		/* accept */
	NULL,		/* reopen */
	user_logmsg,	/* logmsg */
	user_print,	/* print */
	NULL,		/* fd */
	user_inccnt,	/* inccnt */
	NULL,		/* drain */
	user_expire,	/* expire */
	NULL,		/* expireold */
	NULL,		/* more */
};


userspec_t *
user_find(parm)
	char *parm;
{
	userspec_t *us;

	for (us = uslist; us; us = us->us_next)
		if (!strncmp(us->us_user, parm, 9))
			return us;
	return NULL;
}


userspec_t *
user_new(parm)
	char *parm;
{
	userspec_t *us;

	if ((us = user_find(parm))) {
		return us;
	}

	if (!(NEW(us, NEW_USERS)))
		return NULL;
	strncpy(us->us_user, parm, sizeof(us->us_user));
	us->us_user[sizeof(us->us_user) - 1] = '\0';
	us->us_next = uslist;
	us->us_cnt = 0;
	uslist = us;
	return us;
}


void
user_expire()
{
	userspec_t *us;

	while ((us = uslist)) {
		assert(us->us_cnt == 0);
		uslist = us->us_next;
		free(us);
	}
}


/*
 * Send a message to each user logged in as specified by "us"
 */
int
user_logmsg(ptr, fdp, iov, nio, lmp)
	void *ptr;
	int *fdp;
	struct iovec *iov;
	int nio;
	logmsgtype_t *lmp;
{
	userspec_t *us;
#ifdef	_PATH_UTMP
	FILE *fp = fopen(_PATH_UTMP, "r");
#else
	FILE *fp = fopen("/etc/utmp", "r");
#endif
	struct iovec liov[MAX_NIOV], *iovp;
	char line[40], greet[80];
	struct utmp ut;
	int fd;

	us = ptr;
	*fdp = 0;

	if (opts & OPT_VERBOSE)
		fprintf(stderr, "msg to user: %s\n", us->us_user);
	if (!fp)
		return 0;
	bcopy((char *)iov, (char *)liov, sizeof(liov));
	iovp = liov;

	/*
	 * If it's a "WALL" message (to *) then prepend a small message
	 * about where the message is coming from.
	 */
	if (!strcmp("*", us->us_user)) {
		snprintf(greet, sizeof(greet),
			 "\r\n\7Message from syslogd@%.16s at %.24s ...\r\n",
			 (char *)liov[2].iov_base, ctime(&lmp->lm_when));
		iovp++;
		nio--;
		iovp[0].iov_base = greet;
		iovp[0].iov_len = strlen(greet);
	}

	/*
	 * Scan utmp file for people who are logged in.
	 */
	while (fread(&ut, 1, sizeof(ut), fp)) {
		if (!*ut.ut_name || !*ut.ut_line)
			continue;
		if (strncmp(ut.ut_name, us->us_user, sizeof(us->us_user) - 1) &&
		    strncmp("*", us->us_user, sizeof(us->us_user) - 1))
			continue;
		snprintf(line, sizeof(line), "/dev/%-32.32s", ut.ut_line);
		/*
		 * Attempt to prevent it hanging on opening a tty.
		 */
		alarm(1);
		fd = open(line, O_WRONLY|O_APPEND);
		alarm(0);
		if (fd == -1)
			continue;
		alarm(1);
		(void) writev(fd, iovp, nio);
		alarm(0);
		close(fd);
	}
	fclose(fp);
	return 0;
}

int
user_inccnt(ptr, inc)
	void *ptr;
	int inc;
{
	userspec_t *us;

	us = ptr;
	us->us_cnt += inc;
	return us->us_cnt;
}


void
user_print(ptr)
	void *ptr;
{
	userspec_t *us;

	us = ptr;
	printf(" user %s ", us->us_user);
}
