#include "h.h"
#include "nsyslog.h"
#include "var.h"
#include "hash.h"
#include "y.tab.h"


#if !defined(lint)
static const char rcsid[] = "$Id: var.c,v 1.61 1999/05/10 18:22:47 darrenr Exp $";
#endif

var_t		*vlist = NULL;
sdlist_t	*sdlist = NULL, *sdtail = NULL;
options_t	options = { 0,			/* op_flags */
			    MARKFREQ,		/* op_markfreq */
			    SYNCFREQ,		/* op_syncfreq */
			    -1,
			    0,
			    NSYSLOG_LOGDIR,	/* op_logdir */
			    NSYSLOG_DIR,	/* op_sysdir */
			    NULL,		/* op_chroot */
			    HASH_NONE,		/* op_hashtype */
			    NULL,		/* op_hashdir */
			    NULL,		/* op_hashext */
			    1,			/* op_shrepeat */
			    NULL,		/* op_pkgconfs */
			    0,			/* op_shfacpri */
			    0,			/* op_ssl */
			    NULL,		/* op_sslcert */
			    NULL,		/* op_sslcipher */
};

static	var_t		*current = NULL;
static	void		inc_filter __P((fillist_t *));
int	not_match = 0, save_not = 0;
int	do_raw = 0;


int
get_yesno(str)
	char *str;
{
	if (!strcasecmp(str, "yes") ||
	    ((str[1] == '\0') && (*str == 'y' || *str == 'Y')))
		return 1;
	else if (!strcasecmp(str, "no") ||
	    ((str[1] == '\0') && (*str == 'n' || *str == 'N')))
		return 0;
	return -1;
}


void
set_not(val)
	int val;
{
	save_not = not_match;
	not_match = val;
}


int
get_not()
{
	int ret = not_match;

	not_match = save_not;
	return ret;
}


/*
 * set the do_raw flag so that the next network socket created picks it
 * up.
 */
void
set_raw()
{
	do_raw = VAR_RAW;
}


/*
 * set a global option.
 */
void
set_option(opt, parm)
	int opt;
	char *parm;
{
	int yn;

	if (opt == IL_MARK)
		options.op_markfreq = strtol(parm, NULL, 0);
	else if (opt == IL_SYNC)
		options.op_syncfreq = strtol(parm, NULL, 0);
	else if (opt == IL_SYNCTIME)
		options.op_synctime = strtol(parm, NULL, 0);
	else if (opt == IL_SYNCPRIORITY) {
		if (isalpha(*parm))
			options.op_syncpri = pri_findname(parm);
		else
			options.op_syncpri = strtol(parm, NULL, 0);
	} else if (opt == IL_LOGDIR) {
		if (options.op_flags & OPF_LOGDIR)
			free(options.op_logdir);
		options.op_logdir = parm;
		options.op_flags |= OPF_LOGDIR;
	} else if (opt == IL_DIRECTORY) {
		if (options.op_flags & OPF_SYSDIR)
			free(options.op_sysdir);
		options.op_sysdir = parm;
		options.op_flags |= OPF_LOGDIR;
	} else if (opt == IL_REPEATED) {
		if ((yn = get_yesno(parm)) >= 0)
			options.op_shrepeat = yn;
	} else if (opt == IL_PACKAGES) {
		if (options.op_flags & OPF_PKGCONFS)
			free(options.op_pkgconfs);
		options.op_pkgconfs = parm;
		options.op_flags |= OPF_PKGCONFS;
	} else if (opt == IL_SHOWFACPRI) {
		if ((yn = get_yesno(parm)) >= 0)
			options.op_shfacpri = yn;
	}
}


void
set_sslopt(opt, parm)
	int opt;
	char *parm;
{
	options.op_ssl = 1;

	if (opt == IL_CERTIFICATE)
		options.op_sslcert = parm;
	else if (opt == IL_CIPHER)
		options.op_sslcipher = parm;
}

/*
 * set a hash option.
 */
void
set_hoption(opt, parm)
	int opt;
	char *parm;
{
	struct stat sb;
	htable_t *ht;

	if (opt == IL_HASHALGO) {
		ht = hash_lookup(parm);
		if (ht) {
			options.op_hashtype = ht->ht_type;
			options.op_hashext = ht->ht_ext;
		} else
			syserr("unknown hash algorithm %s", parm);
	} else if (opt == IL_DIRECTORY) {
		if (options.op_hashdir)
			free(options.op_hashdir);
		if (stat(parm, &sb) == -1 && errno == ENOENT) {
			if (mkdir(parm, 0700) == -1)
				syserr("failed to make hash directory %s",
					parm);
			else if (opts & OPT_VERBOSE)
				fprintf(stderr, "created hash directory %s\n",
					parm);
		}
		options.op_hashdir = parm;
	}
}


/*
 * get a pointer to a named variable.
 */
var_t *
get_var(name)
	char *name;
{
	var_t *v;

	for (v = vlist; v; v = v->v_next)
		if (!strcmp(v->v_name, name))
			return v;
	return NULL;
}


/*
 * create a new named variable and return its type.  if one already exists
 * with that name, return its type anyway.
 */
int
new_var(name, type)
char *name;
	int type;
{
	var_t *v;

	if ((v = get_var(name)) != 0L)
		return v->v_type;
	if (!(NEW(v, NEW_VARS)))
		return -1;
	v->v_name = (char *)strdup(name);
	v->v_list = new_filetype(type);
	v->v_type = type;
	v->v_next = vlist;
	vlist = v;
	return 0;
}


/*
 * create a new anonymous variable.  An anonymous variable isn't defined in
 * the configuration file and is an artifact of creating something without
 * a name.
 */
var_t *
new_anon(type)
	int type;
{
	static int anonvars = 0;
	char sbuf[20];
	var_t *v;

	if (!(NEW(v, NEW_VARS)))
		return NULL;
	snprintf(sbuf, sizeof(sbuf), "<%x>", anonvars++);
	v->v_type = type;
	v->v_name = (char *)strdup(sbuf);
	v->v_list = new_filetype(type);
	v->v_next = vlist;
	vlist = v;
	return v;
}


/*
 * set the working object to a known variable.
 */
void
new_obj(name)
	char *name;
{
	var_t *v;

	v = get_var(name);
	current = v;
}


/*
 * associate an internal file type structure with a known variable.
 */
var_t *
add_obj(type, ptr)
	int type;
	void *ptr;
{
	var_t *v;
	int i;

	if (ptr == NULL)
		return NULL;

	if (VAR_TYPE(type) == SRC_INET)
		type = ((netspec_t *)ptr)->ns_type;
	if (!current)
		switch (VAR_MODE(type))
		{
		case DEST_NONE :
			current = new_anon(VAR_DST);
			break;
		case SRC_NONE :
			current = new_anon(VAR_SRC);
			break;
		case FIL_NONE :
			current = new_anon(VAR_FIL);
			break;
		default:
			return NULL;
			break;
		}

	if (!(v = current))
		return NULL;

	if (v->v_num > 0) {
		v->v_num++;
		v->v_list = (filetype_t *)realloc(v->v_list,
						sizeof(*v->v_list) * v->v_num);
	} else {
		v->v_num = 1;
	}
	i = v->v_num - 1;
	v->v_list[i].ft_ptr = ptr;
	v->v_list[i].ft_type = type;
	v->v_list[i].ft_parent = v;
	v->v_list[i].ft_vt = type_find(type);
	return v;
}


/*
 * stop working with a variable.
 */
var_t *
end_obj()
{
	var_t *v;

	v = current;
	current = NULL;
	return v;
}


int
var_ismember(var, ft)
	var_t *var;
	filetype_t *ft;
{
	void *obj;
	int i;

	if (ft->ft_parent == var)
		return 1;
	obj = ft->ft_ptr;
	for (i = var->v_num - 1; i >= 0; i--)
		if (var->v_list[i].ft_ptr == obj)
			return 1;
	return 0;
}


void
link_sfd(src, fil, dst)
	var_t *src;
	fillist_t *fil;
	var_t *dst;
{
	sdlist_t *sd;
	int i;

	if (!(NEW(sd, NEW_SDLIST)))
		return;
	sd->sd_src = src;
	src->v_cnt++;
	for (i = src->v_num - 1; i >= 0; i--)
		VAR_INCTYPE(src->v_list + i, 1);
	sd->sd_dst = dst;

	dst->v_cnt++;
	for (i = dst->v_num - 1; i >= 0; i--)
		VAR_INCTYPE(dst->v_list + i, 1);
	sd->sd_fil = fil;
	inc_filter(fil);
	if (!sdlist)
		sdlist = sd;
	else
		sdtail->sd_next = sd;
	sdtail = sd;
}


fillist_t *
link_filters(name, tail)
	char *name;
	fillist_t *tail;
{
	fillist_t *fl;
	var_t *fil;

	fil = get_var(name);
	if (!fil || !(NEW(fl, NEW_FILLIST)))
		return NULL;
	fl->fl_fil = fil;
	fl->fl_comp = get_not();
	fil->v_cnt++;
	fl->fl_and = tail;
	return fl;
}


fillist_t *
add_filter(f1, f2)
	fillist_t *f1, *f2;
{
	f1->fl_or = f2;
	return f1;
}


void
print_not(isnot)
	int isnot;
{
	if (isnot)
		printf(" !");
}


table_t	comparators[] = {
	{ "= ",		IL_EQ },
	{ "!= ",	IL_NEQ },
	{ "<= ",	IL_LE },
	{ "< ",		IL_LT },
	{ "",		IL_GE },
	{ "> ",		IL_GT },
	{ NULL, 0 }
};


char *
comp_toname(comp)
	int comp;
{
	int i;

	for (i = 0; comparators[i].name; i++)
		if (comparators[i].value == comp)
			return comparators[i].name;
	return NULL;
}


void
print_type(ft)
	filetype_t *ft;
{
	(* ft->ft_vt->vt_print)(ft->ft_ptr);
}


void
dump_all()
{
	fillist_t *f1, *f2;
	filetype_t *ft;
	sdlist_t *s;
	var_t *v;
	int i;

	for (v = vlist; v; v = v->v_next) {
		if (v->v_type == VAR_SRC || v->v_type == IL_SRCNAME)
			printf("source ");
		else if (v->v_type == VAR_DST || v->v_type == IL_DESTNAME)
			printf("destination ");
		else if (v->v_type == VAR_FIL || v->v_type == IL_FILNAME)
			printf("filter ");
		printf("%s {", v->v_name);
		for (i = 0; i < v->v_num; i++) {
			print_type(v->v_list + i);
			printf(";");
		}
		printf("}; # %d, %d\n", v->v_num, v->v_cnt);
	}

	for (s = sdlist; s; s = s->sd_next) {
		printf("log { source %s;\n", s->sd_src->v_name);
		for (f1 = s->sd_fil; f1; f1 = f1->fl_or) {
			printf("\tfilter ");
			for (f2 = f1; f2; f2 = f2->fl_and) {
				printf("%s%s%c", f2 == f1 ? "" : "\t\t",
					f2->fl_fil->v_name,
					f2->fl_and ? ',' : ';');
				printf(" # %d\n", f2->fl_fil->v_cnt);
			}
		}
		printf("\tdestination %s; # %d\n};\n", s->sd_dst->v_name,
			s->sd_dst->v_cnt);
	}
}


int
next_sfd(ptr, cnt)
	void **ptr;
	int *cnt;
{
	sdlist_t *s = NULL;
	int i = 0, fd = -1;
	filetype_t *ft;

	if (ptr)
		s = *((sdlist_t **)ptr);

	if (!s)
		s = sdlist;
	else {
		if (*cnt >= s->sd_src->v_num)
			s = s->sd_next;
		else
			i = *cnt;
	}
	*ptr = s;

	if (!s)
		return -2;

	ft = s->sd_src->v_list + i;
	fd = (*ft->ft_vt->vt_fd)(ft->ft_ptr);
	i++;
	*cnt = i;
	return fd;
}


int
next_dfd(ptr, cnt)
	void **ptr;
	int *cnt;
{
	sdlist_t *s = NULL;
	int i = 0, fd = -1;
	filetype_t *ft;

	if (ptr)
		s = *((sdlist_t **)ptr);

	if (!s)
		s = sdlist;
	else {
		if (*cnt >= s->sd_dst->v_num)
			s = s->sd_next;
		else
			i = *cnt;
	}
	*ptr = s;

	if (!s)
		return -2;

	ft = s->sd_dst->v_list + i;
	/*
	 * Users are a valid destination that don't have an fd associated
	 * with them, hence no function returning it.
	 */
	if (ft->ft_vt->vt_fd != NULL)
		fd = (*ft->ft_vt->vt_fd)(ft->ft_ptr);
	else
		fd = -1;
	i++;
	*cnt = i;
	return fd;
}


int
regerr(err)
	int err;
{
	fprintf(stderr, "Regular expression error: ");
	if (err == 11)
		fprintf(stderr, "Range endpoint too large.\n");
	else if (err == 16)
		fprintf(stderr, "Bad number.\n");
	else if (err == 25)
		fprintf(stderr, "``\\ digit'' out of range.\n");
	else if (err == 36)
		fprintf(stderr, "Illegal or missing delimiter.\n");
	else if (err == 41)
		fprintf(stderr, "No remembered search string.\n");
	else if (err == 42)
		fprintf(stderr, "\\( \\) imbalance.\n");
	else if (err == 43)
		fprintf(stderr, "Too many \\(.\n");
	else if (err == 44)
		fprintf(stderr, "More than 2 numbers  given  in \\{ \\}.\n");
	else if (err == 45)
		fprintf(stderr, "} expected after \\.\n");
	else if (err == 46)
		fprintf(stderr, "First number exceeds second in \\{ \\}.\n");
	else if (err == 49)
		fprintf(stderr, "[] imbalance.\n");
	else if (err == 50)
		fprintf(stderr, "Regular expression too long.\n");
	else
		fprintf(stderr, "unknown error %d\n", err);
	exit(1);
}


void
free_var(v)
	var_t *v;
{
	var_t **vp;
	int i;
	for (vp = &vlist; (*vp); ) {
		if (*vp == v) {
			v->v_cnt--;
			assert(v->v_cnt >= 0);
			for (i = v->v_num - 1; i >= 0; i--)
				VAR_INCTYPE(v->v_list + i, -1);
			if (v->v_cnt > 0) {
				v = NULL;
				break;
			}
			*vp = v->v_next;
			free(v->v_list);
			free(v->v_name);
			free(v);
			v = NULL;
			break;
		} else
			vp = &(*vp)->v_next;
	}
	if (v && (opts & OPT_DEBUG))
		fprintf(stderr, "Didn't find %p [%s]\n", v, v->v_name);
}


/*
 * this filter is being used another time, to increase all the reference counts
 * for it and its component objects.
 */
static void
inc_filter(fil)
	fillist_t *fil;
{
	fillist_t *fl;
	var_t *v;
	int i;

	while ((fl = fil)) {
		fil = fl->fl_and;
		v = fl->fl_fil;
		for (i = v->v_num - 1; i >= 0; i--)
			VAR_INCTYPE(v->v_list + i, 1);
		if (fl->fl_or)
			inc_filter(fl->fl_or);
	}
}


void
free_filter(fil)
	fillist_t *fil;
{
	fillist_t *fl;

	while ((fl = fil)) {
		fil = fl->fl_and;
		free_var(fl->fl_fil);
		if (fl->fl_or)
			free_filter(fl->fl_or);
		free(fl);
	}
}


void
expire_all()
{
	sdlist_t *sd;
	vartype_t *v;
	void *ptr;
	int i;

	while ((sd = sdlist)) {
		sdlist = sd->sd_next;
		free_var(sd->sd_src);
		free_var(sd->sd_dst);
		free_filter(sd->sd_fil);
		free(sd);
	}
	sdtail = sdlist;
	assert(sdlist == NULL);

	ptr = NULL;
	for (v = NULL; v = type_next(&ptr); )
		if (v->vt_expire != NULL)
			(*v->vt_expire)();

	options.op_markfreq = MARKFREQ;
	options.op_syncfreq = SYNCFREQ;
}


void
expire_old()
{
	vartype_t *v;
	void *ptr;

	ptr = NULL;
	for (v = NULL; v = type_next(&ptr); )
		if (v->vt_expireold != NULL)
			(*v->vt_expireold)();
}
