#if !defined(lint)
static const char var_rcsid[] = "$Id: var.h,v 1.55 1999/05/10 18:25:41 darrenr Exp $";
#endif

typedef	struct	sockaddr	sockaddr_t;
typedef	struct	sockaddr_in	sin_t;
typedef	struct	sockaddr_un	sun_t;


typedef	struct options	{
	int	op_flags;	/* which have been changed */
	time_t	op_markfreq;	/* how long between marks */
	int	op_syncfreq;	/* how many messages between fsync's */
	int	op_syncpri;	/* fsync on set priority or higher */
	int	op_synctime;	/* fsync every X seconds */
	char	*op_logdir;	/* default dir. for log files */
	char	*op_sysdir;	/* default dir. for syslog config files */
	char	*op_chroot;	/* directory to chroot to */
	int	op_hashtype;
	char	*op_hashdir;
	char	*op_hashext;
	int	op_shrepeat;	/* show repeats */
	char	*op_pkgconfs;	/* mask for finding package confs */
	int	op_shfacpri;	/* put the facility & priority into log msg */
	int	op_ssl;		/* is ssl enabled ? */
	char	*op_sslcert;	/* ssl certificate file */
	char	*op_sslcipher;	/* ssl cipher selection */
} options_t;

/* for op_flags */
#define	OPF_LOGDIR	0x001
#define	OPF_SYSDIR	0x002
#define	OPF_CHROOT	0x004
#define	OPF_HASHDIR	0x008
#define	OPF_HASHEXT	0x010
#define	OPF_PKGCONFS	0x020

#define	MARKFREQ	20 * 60 /* seconds between marks */
#define	SYNCFREQ	1


typedef	struct	logmsg	{
	char 	*lg_line;
	char	*lg_host;
	char	*lg_prog;
	int	lg_len;
	int	lg_ref;
	int	lg_pri;
	u_long	lg_id;
	time_t	lg_time;
} logmsg_t;


typedef	struct	tcplmsg	{
	u_char	tl_v;		/* version */
	u_char	tl_zero[3];	/* reserved */
	struct	in_addr	tl_src;	/* original source IP address */
	u_int	tl_len;		/* total length of message */
	u_int	tl_mlen;	/* length of log message */
	u_int	tl_hlen;	/* length of hash after message */
	u_int	tl_pri;		/* facility|priority of message */
	time_t	tl_tv;		/* time message was received */
} tcplmsg_t;


typedef	struct	nonce {
	char	*no_buf;
	size_t	no_sz;
} nonce_t;

#define	INIT_NONCE(x)	{(x)->no_buf = (char *)malloc(NSYSLOG_NONCE_SZ); \
			 (x)->no_sz = NSYSLOG_NONCE_SZ; }


typedef	struct	filetype	{
	int ft_type;
	void *ft_ptr;
	void *ft_var;
	void *ft_parent;
	struct vartype *ft_vt;
	struct logmsg *ft_log;
	time_t ft_lasttime;
	time_t ft_closed;
	int ft_mcount;
} filetype_t;

#define	ft_line	ft_log->lg_line
#define	ft_host	ft_log->lg_host
#define	ft_prog	ft_log->lg_prog
#define	ft_len	ft_log->lg_len
#define	ft_pri	ft_log->lg_pri
#define	ft_firsttime	ft_log->lg_time
#define	ft_logid	ft_log->lg_id

#define	DEST_NONE	0x20
#define	DEST_FILE	0x21
#define	DEST_INET	0x22
#define	DEST_UNIX	0x23
#define	DEST_CMD	0x24
#define	DEST_USER	0x25
#define	DEST_TCP	0x26
#define	SRC_NONE	0x40
#define	SRC_FILE	0x41
#define	SRC_INET	0x42
#define	SRC_UNIX	0x43
#define	SRC_INET_LISTEN	0x44
#define	SRC_TCP		0x45
#define	SRC_UNIX_LISTEN	0x46
#define	SRC_STREAM	0x47
#define	SRC_DOOR	0x48
#define	FIL_NONE	0x80
#define	FIL_FAC		0x81
#define	FIL_PRI		0x82
#define	FIL_FACPRI	0x83
#define	FIL_WRAP	0x84
#define	FIL_MATCH	0x85
#define	FIL_MATCHFILE	0x86
#define	FIL_MATCHPROG	0x87
#define	FIL_CMD		0x88
#define	VAR_TMSK	(DEST_NONE|SRC_NONE|FIL_NONE)
#define	VAR_WHICH	0x1f
#define	VAR_MODE(x)	((x) & VAR_TMSK)
#define	VAR_DSF		(VAR_TMSK|VAR_WHICH)
#define	VAR_TYPE(x)	((x) & VAR_DSF)
#define	VAR_OLD		0x100
#define	VAR_RAW		0x200
#define	VAR_HASH	0x400
#define	VAR_LISTEN	0x800
#define	VAR_INTERNAL	0x1000
#define	VAR_SSL		0x2000

#define	TCP_LISTEN	(SRC_INET_LISTEN|VAR_LISTEN)
#define	UNIX_LISTEN	(SRC_UNIX_LISTEN|VAR_LISTEN)

#define	VAR_DOHASH(x)	((x) & VAR_HASH)

#define	VAR_INCTYPE(x, y)	(*(x)->ft_vt->vt_inccnt)((x)->ft_ptr, (y))

typedef struct logmsgtype	{
	struct filetype *lm_ft;	/* source of message */
	int lm_flags;
	struct logmsg *lm_log;
	char *lm_time;		/* timestamp of origin */
	void *lm_src;		/* source of message (socket) */
	int lm_slen;		/* size of source (socket length) */
	int lm_lpri;		/* local priority: this and the next field */
	time_t lm_lwhen;	/* local time: are for when lm_log is null  */
} logmsgtype_t;

#define	lm_line	lm_log->lg_line
#define	lm_host	lm_log->lg_host
#define	lm_prog	lm_log->lg_prog
#define	lm_len	lm_log->lg_len	/* length of message */
#define	lm_pri	lm_log->lg_pri	/* combination of facility and priority */
#define	lm_when	lm_log->lg_time	/* when it was received */
#define	lm_logid	lm_log->lg_id


typedef	struct	table	{
	char *name;
	int value;
	char letter;
} table_t;

extern	table_t	*find_fac __P((int));
extern	table_t	*find_pri __P((int));


typedef struct var {
	struct var *v_next;
	char *v_name;
	int v_type;
	int v_cnt;
	int v_num;
	struct filetype *v_list;
} var_t;

#define	VAR_SRC	1
#define	VAR_FIL	2
#define	VAR_DST	3


typedef struct filespec {
	struct filespec *fs_next;
	u_long fs_cnts[3];	/* hash file counters */
	int fs_cnt;	/* reference count */
	int fs_mode;	/* opening mode of the file. */
	int fs_fd;	/* fd to the open file */
	int fs_hfd;	/* fd to the open hash file */
	int fs_written;	/* how many entries written since last fsync ? */
	char *fs_path;	/* path to the log file */
	int fs_syncfreq;	/* fsync every X messages */
	int fs_synctime;	/* fsync every X seconds */
	int fs_syncpri;		/* fsync priority X or higher */
	time_t fs_lastsync;
	int fs_markfreq;
	void *fs_dbuf;	/* where to put hashing private buffer */
	void *fs_hptr;	/* pointer to struct with info. about hashing */
} filespec_t;

#define	fs_hcnt	fs_cnts[0]	/* hash count */
#define	fs_lcnt	fs_cnts[1]	/* line count (lines written) */
#define	fs_fcnt	fs_cnts[2]	/* file count */

extern	filespec_t	*fslist;

#define	FILE_DOHASH(x)	VAR_DOHASH((x)->fs_mode)

typedef struct unixspec {
	struct unixspec *xs_next;
	struct unixspec *xs_parent;
	int xs_cnt;
	int xs_fd;
	int xs_mode;
	int xs_type;
	int xs_proto;
	buflist_t *xs_rlist;
	buflist_t **xs_rtail;
	buflist_t *xs_wlist;
	buflist_t **xs_wtail;
	struct sockaddr_un xs_un;
} unixspec_t;

#define	UNSZ	sizeof(((struct sockaddr_un *)0)->sun_path)

extern	unixspec_t	*xslist;

#ifdef	HAVE_SSL_H
typedef	struct	ssl	{
	char	*ssl_cert;
	SSL	*ssl_ssl;
	SSL_CTX	*ssl_ctx;
	SSL_METHOD *ssl_meth;
	int	ssl_r;
	BIO	*ssl_bio;
	buflist_t *ssl_rlist;
	buflist_t **ssl_rtail;
} ssl_t;
#endif


typedef struct netspec {
	struct netspec *ns_next;
	struct netspec *ns_parent;	/* for accepted connections */
	int ns_fd;		/* fd for open connection */
	int ns_mode;
	int ns_cnt;		/* usage count */
	int ns_type;		/* type of network connection */
	int ns_proto;		/* network protocol */
	int ns_state;		/* current state of TCP connection */
	int ns_nstate;		/* saved state of TCP connection */
	sin_t ns_sin;		/* remote end of connection for TCP/UDP */
	int ns_hash;		/* hashing function selected */
	int ns_crypt;		/* encryption function selected */
	void *ns_secret;	/* per-host secret */
	size_t ns_slen;		/* length of ns_secret */
	u_long ns_lcnt;		/* line count in file */
	int ns_hfd;		/* fd for the file to store hash entries */
	int ns_lfd;		/* fd for the file to store log messages */
	char *ns_path;		/* where to find H0, hash and log files */
	void *ns_hptr;		/* pointer to hashing info */
	void *ns_dbuf;		/* used to store H0 and updated thereafter */
	char *ns_hmac;		/* used to store hash MAC to send */
	nonce_t ns_local;	/* locally generated nonce */
	nonce_t ns_remote;	/* remote generated nonce */
	nonce_t ns_digest;
	buflist_t **ns_rtail;	/* pointer to where to add a buffer */
	buflist_t *ns_rlist;	/* list of buffers with data to read */
	buflist_t **ns_wtail;
	buflist_t *ns_wlist;	/* list of buffers with data to write */
#ifdef	HAVE_SSL_H
	ssl_t *ns_ssl;
#endif
} netspec_t;

extern	netspec_t	*nslist;
#define	ns_nlen		ns_local.no_sz
#define	ns_nonce	ns_local.no_buf
#define	ns_rnlen	ns_remote.no_sz
#define	ns_rnonce	ns_remote.no_buf
#define	ns_diglen	ns_digest.no_sz
#define	ns_dig		ns_digest.no_buf

#define	NET_TCP	IPPROTO_TCP
#define	NET_UDP	IPPROTO_UDP
#define	NET_CLONE	0


typedef struct sdlist {
	struct sdlist *sd_next;
	var_t *sd_src;
	var_t *sd_dst;
	struct fillist *sd_fil;
} sdlist_t;

extern sdlist_t *sdlist;


typedef struct fillist {
	struct fillist *fl_and;
	struct fillist *fl_or;
	var_t *fl_fil;
	int fl_comp;
} fillist_t;


#define	REGCOMP_FLAGS	(REG_EXTENDED|REG_ICASE|REG_NOSUB)
#define	NEW(x, y)	(x) = (void *)calloc(1, sizeof(*x))
#define	NEW_FT			1
#define	NEW_DOOR		2
#define	NEW_FACS		3
#define	NEW_FACPRIS		4
#define	NEW_FILES		5
#define	NEW_HASH		6
#define	NEW_MATCHS		7
#define	NEW_MATCHFS		8
#define	NEW_MATCHPS		9
#define	NEW_NETS		10
#define	NEW_NETMS		11
#define	NEW_FILLIST		12
#define	NEW_CMD			13
#define	NEW_PRIS		14
#define	NEW_STREAMS		15
#define	NEW_UNIXS		16
#define	NEW_VARS		17
#define	NEW_SDLIST		18
#define	NEW_USERS		19


#include "type.h"

extern	filetype_t	*fttable;
extern	u_int		fttsize;
extern	options_t	options;
extern	var_t		*vlist;
extern	int		token;

extern	var_t		*get_var __P((char *));
extern	int		new_var __P((char *, int));
extern	var_t		*new_anon __P((int));
extern	void		new_obj __P((char *));
extern	var_t 		*add_obj __P((int, void *));
extern	var_t		*end_obj __P((void));
extern	void		end_file __P((void));

extern	unixspec_t	*unix_new __P((int, int, char *));
extern	unixspec_t	*unix_find __P((int, char *));

extern	filespec_t	*file_new __P((int, char *));
extern	filespec_t	*file_find __P((int, char *));
extern	int		open_file __P((int, char *));

extern	netspec_t	*net_new __P((int, int, char *, char *));
extern	void		net_free __P((netspec_t *));

extern	struct facspec	*fac_new __P((char *));
extern	struct facprispec *facpri_new __P((char *, int, char *));
extern	struct prispec *pri_new __P((int, char *));
extern	struct matchspec *match_new __P((char *));
extern	struct matchfilespec *matchfile_new __P((char *));
extern	struct matchprogspec *matchprog_new __P((char *));
extern	struct nmaskspec *nmask_new __P((char *));
extern	struct cmdspec *cmd_new __P((char *));
extern	struct prispec *prival_new __P((int, int));
extern	struct facspec *facval_new __P((int));
extern	struct facprispec *facprival_new __P((int, int));
extern	struct table *pri_find __P((int));
extern	struct table *fac_find __P((int));
extern	struct userspec *user_new __P((char *));
extern	struct userspec *user_find __P((char *));


extern	int		new_socket __P((int, int, sockaddr_t *, int));

extern	void		link_sfd __P((var_t *, fillist_t *, var_t *));
extern	fillist_t	*link_filters __P((char *, fillist_t *));
extern	fillist_t	*add_filter __P((fillist_t *, fillist_t *));

#ifndef	REGEX_OLD
extern	int		regerr __P((int));
#endif

extern	void		init_config __P((void));
extern	int		read_config __P((char *, fd_set *));
extern	int		read_oldconfig __P((char *, fd_set *));
extern	int		load_config __P((void));
extern	int		setup_filetable __P((fd_set *));
extern	void		init_lex __P((void));

extern	filetype_t	 *new_filetype __P((int));

extern	struct	in_addr	getipv4addr __P((char *));
extern	struct	in_addr getipmask __P((char *));
extern	u_short		getportnum __P((char *, char *));
extern	char		*get_shorthostname __P((void));
extern	char		*get_fullhostname __P((void));

extern	void		set_sync __P((int));
extern	void		set_mark __P((int));
extern	void		set_raw __P((void));
extern	void		set_option __P((int, char *));
extern	void		set_not __P((int));
extern	int		get_not __P((void));

extern	void		expire_all __P((void));
extern	void		expire_old __P((void));

extern	void		free_filter __P((fillist_t *));
extern	void		free_type __P((filetype_t *));
extern	void		free_var __P((var_t *));

extern	void		dump_all __P((void));
extern	void		print_type __P((filetype_t *ft));
extern	void		print_not __P((int));

extern	int		next_dfd __P((void **, int *));
extern	int		next_sfd __P((void **, int *));

extern	void		logmsg __P((logmsgtype_t *));
extern	void		checkmarks __P((void));
extern	void		checksync __P((void));
extern	void		writelogmsg __P((logmsgtype_t *));


extern	void		close_all __P((void));
extern	int		reload __P((void));

extern	void		logerror __P((char *));

extern	logmsg_t	*new_logmsg __P((char *, int));
extern	void		free_logmsg __P((logmsg_t **));

#if defined(HAVE_GETMSG) && defined(HAVE_SYS_STRLOG_H)
extern	struct streamspec 	*stream_new __P((char *));
# if SOLARIS
extern	struct doorspec	*door_new __P((char *));
# endif
#endif
