%{
/*
 * Copyright (C) 1997 by Darren Reed.
 *
 * $Id: yac.y,v 1.36 1999/05/10 18:23:11 darrenr Exp $
 */
#include "h.h"
#include "nsyslog.h"
#include "var.h"

#if !defined(lint)
static const char rcsid[] = "$Id: yac.y,v 1.36 1999/05/10 18:23:11 darrenr Exp $";
#endif

extern	int	opts;
extern	int	state, state, lineNum, token;
extern	int	yylineno;
extern	char	yytext[];
extern	FILE	*yyin;

void	yyerror	__P((char *));
int	yylex	__P((void));
#define	YYDEBUG 1
#if !defined(ultrix) && !defined(hpux)
int	yydebug = 0;
#else
extern	int	yydebug;
#endif

int	yyparse __P((void));
%}
%union {
	void	*vd;
	var_t	*var;
	char	*str;
	int	num;
	void	*mspec;
	void	*nspec;
	void	*fspec;
	void	*uspec;
	void	*xspec;
	void	*cspec;
	void	*nmask;
	struct	vartype	*vptr;
	struct	arglist	*aptr;
	void	*fac;
	void	*pri;
	void	*facpri;
	void	*rot;
}
%type	<mspec>	regex regfile matchprog
%type	<fspec>	ofile dfile sfile
%type	<uspec> user
%type	<vptr> srctype stype dtype dsttype
%type	<aptr> arglist
%type	<num> sync
%type	<cspec> command
%type	<nmask> wrapper
%type	<fac> facility
%type	<pri> priority pri
%type	<facpri> facpri
%type	<vd> actsrc actdest dofilter actfil filnamelist
%type	<var> dest destlist destobj src srclist srcobj fil mfil fillist filobj
%type	<str> filname
%type	<num> compare
%token	<str> IL_TOKEN
%token	<str> IL_DESTNAME IL_SRCNAME IL_FILNAME
%token	IL_SOURCE IL_FILE IL_UNIX IL_USER IL_TCP IL_UDP IL_FILTER IL_FACILITY
%token	IL_PRIORITY IL_FACPRI IL_MATCH IL_MATCHFILE IL_MATCHPROG
%token	IL_DESTINATION IL_RAW IL_COMMAND
%token	IL_LOG IL_NOT
%token	<num> IL_LT IL_LE IL_GT IL_GE IL_NEQ IL_EQ IL_DGRAM IL_STREAM
%token	IL_OUTPUT IL_NETMASK
%token	IL_OPTIONS IL_MARK IL_HASH IL_DIRECTORY IL_HASHALGO IL_LOGDIR
%token	IL_REPEATED IL_PACKAGES IL_SHOWFACPRI IL_INTERNAL
%token	IL_SSL IL_CERTIFICATE IL_CIPHER
%token	<num> IL_SYNC IL_SYNCTIME IL_SYNCPRIORITY
%%
file:	file line ';'
	| line ';'
	;

line:	options
	| source
	| filter
	| destination
	| action
	;

options:
	IL_OPTIONS '{' optlist '}'
	;

optlist:
	option
	| option optlist
	;

option:	syncopt ';'
	| markopt ';'
	| hash ';'
	| diropt ';'
	| logdiropt ';'
	| repeats ';'
	| pkgconfs ';'
	| showfacpri ';'
	| ssl ';'
	;

markopt:
	IL_MARK IL_TOKEN		{ set_option(IL_MARK, $2); free($2); }
	;

syncopt:
	IL_SYNC IL_TOKEN		{ set_option($1, $2); free($2); }
	| IL_SYNCTIME IL_TOKEN		{ set_option($1, $2); free($2); }
	| IL_SYNCPRIORITY IL_TOKEN	{ set_option($1, $2); free($2); }
	;

diropt:	IL_DIRECTORY IL_TOKEN		{ set_option(IL_DIRECTORY, $2); }
	;

logdiropt:
	IL_LOGDIR IL_TOKEN		{ set_option(IL_LOGDIR, $2); }
	;

hash:	IL_HASH '{' hashlist '}'
	;

repeats:
	IL_REPEATED IL_TOKEN		{ set_option(IL_REPEATED, $2); }
	;

showfacpri:
	IL_SHOWFACPRI IL_TOKEN		{ set_option(IL_SHOWFACPRI, $2); }
	;

pkgconfs:
	IL_PACKAGES IL_TOKEN		{ set_option(IL_PACKAGES, $2); }
	;

hashlist:
	hashopt
	| hashopt hashlist
	;

hashopt:
	hashdir ';'
	| algo ';'
	;

hashdir:
	IL_DIRECTORY IL_TOKEN		{ set_hoption(IL_DIRECTORY, $2); }
	;

algo:	IL_HASHALGO IL_TOKEN		{ set_hoption(IL_HASHALGO, $2);
					  free($2);}
	;

ssl:	IL_SSL '{' ssllist '}'
	;

ssllist:
	sslopt
	| sslopt ssllist
	;

sslopt:
	sslcert ';'
	| sslcipher ';'
	;

sslcert:
	IL_CERTIFICATE IL_TOKEN		{ set_sslopt(IL_CERTIFICATE, $2); }
	;

sslcipher:
	IL_CIPHER IL_TOKEN		{ set_sslopt(IL_CIPHER, $2); }
	;

source:	newsrc srcobj
	;

newsrc:	IL_SOURCE IL_SRCNAME		{ new_obj($2); free($2); }
	;

srcobj:	'{' srclist '}'			{ $$ = $2; end_obj(); }
	;

srclist:
	src				{ $$ = $1; }
	| src srclist			{ $$ = $1; }
	;

src:	sfile ';'		{ $$ = add_obj(SRC_FILE, $1); }
	| stype ';'		{ $$ = add_obj($1->vt_stype, $1->vt_ptr); }
	| IL_INTERNAL ';'	{ $$ = add_obj(SRC_FILE,
					file_find(SRC_FILE, "/dev/null")); }
	;

stype:	srctype arglist	{ $1->vt_ptr = (*$1->vt_init)($1->vt_stype, $2);
			  arg_free($2); $$ = $1; }
	| srctype	{ $1->vt_ptr = (*$1->vt_init)($1->vt_stype, NULL);
			  $$ = $1; }
	;
srctype:
	IL_TOKEN		{ $$ = type_lookup($1, SRC_NONE);
				  if ($$ == NULL) {
					yyerror("unknown destination type");
					YYABORT;
				  }
				}
	;


dtype:	dsttype arglist	{ $1->vt_ptr = (*$1->vt_init)($1->vt_dtype, $2);
			  arg_free($2); $$ = $1; }
	| dsttype	{ $1->vt_ptr = (*$1->vt_init)($1->vt_dtype, NULL);
			  $$ = $1; }
	;

dsttype:
	IL_TOKEN		{ $$ = type_lookup($1, DEST_NONE);
				  if ($$ == NULL) {
					yyerror("unknown destination type");
					YYABORT;
				  }
				}
	;

arglist:
	IL_TOKEN ',' arglist		{ $$ = arg_add($3, $1); }
	| IL_TOKEN			{ $$ = arg_add(NULL, $1); }
	;

sfile:	IL_FILE IL_TOKEN		{ $$ = file_new(SRC_FILE, $2);
					  free($2); }
	;

filter:	newfil filobj
	;

newfil:	IL_FILTER IL_FILNAME		{ new_obj($2); free($2); }
	;

filobj:	'{' fillist '}' 		{ $$ = $2; end_obj(); }
	;

fillist:
	mfil				{ $$ = $1; }
	| mfil fillist			{ $$ = $1; }
	;

mfil:	fil				{ $$ = $1; }
	| not fil			{ $$ = $2; }
	;

not:	IL_NOT				{ set_not(1); }
	;

fil:	facility ';'			{ $$ = add_obj(FIL_FAC, $1); }
	| priority ';'			{ $$ = add_obj(FIL_PRI, $1); }
	| facpri ';'			{ $$ = add_obj(FIL_FACPRI,$1);}
	| regex ';'			{ $$ = add_obj(FIL_MATCH, $1);}
	| regfile ';'			{ $$ = add_obj(FIL_MATCHFILE, $1);}
	| matchprog ';'			{ $$ = add_obj(FIL_MATCHPROG, $1);}
	| command ';'			{ $$ = add_obj(FIL_CMD, $1); }
	| wrapper ';'			{ $$ = add_obj(FIL_WRAP, $1); }
	;

facility:
	IL_FACILITY IL_TOKEN		{ $$ = fac_new($2); free($2); }
	;

priority:
	IL_PRIORITY pri 			{ $$ = $2; }
	;

facpri:	IL_FACPRI IL_TOKEN ',' IL_TOKEN { $$ = facpri_new($2, IL_GE, $4);
					  free($2); free($4); }
	| IL_FACPRI IL_TOKEN ',' compare IL_TOKEN
					{ $$ = facpri_new($2, $4, $5);
					  free($2); free($5); }
	;

pri:	IL_TOKEN			{ $$ = pri_new(IL_GE, $1); free($1); }
	| compare IL_TOKEN		{ $$ = pri_new($1, $2); free($2); }
	;

compare:
	IL_EQ				{ $$ = $1; }
	| IL_NEQ			{ $$ = $1; }
	| IL_LT				{ $$ = $1; }
	| IL_LE				{ $$ = $1; }
	| IL_GT				{ $$ = $1; }
	| IL_GE				{ $$ = $1; }
	;

regex:	IL_MATCH IL_TOKEN 		{ $$ = match_new($2); free($2); }
	;

regfile:
	IL_MATCHFILE IL_TOKEN 		{ $$ = matchfile_new($2); free($2); }
	;

matchprog:
	IL_MATCHPROG IL_TOKEN		{ $$ = matchprog_new($2); free($2); }
	;

wrapper:
	IL_NETMASK IL_TOKEN 		{ $$ = nmask_new($2); free($2); }
	;

command:
	IL_COMMAND '|' IL_TOKEN		{ $$ = cmd_new($3); free($3); }
	;

destination:
	newdest destobj
	;

newdest:
	IL_DESTINATION IL_DESTNAME	{ new_obj($2); free($2); }
	;

destobj:
	'{' destlist '}' 		{ $$ = $2; end_obj(); }
	;

destlist:
	dest				{ $$ = $1; }
	| dest destlist			{ $$ = $1; }
	;

dest:	dfile ';'			{ $$ = add_obj(DEST_FILE, $1);
					  file_end(); }
	| dtype ';'			{$$ = add_obj($1->vt_dtype,$1->vt_ptr);}
	| command ';'			{ $$ = add_obj(DEST_CMD, $1); }
	| user ';'			{ $$ = add_obj(DEST_USER, $1); }
	;

dfile:	ofile
	| ofile sync ';'		{ set_sync($2); }
	;

ofile:	IL_FILE IL_TOKEN		{ $$ = file_new(DEST_FILE, $2);
					  free($2); }
	;

user:	IL_USER IL_TOKEN		{ $$ = user_new($2); free($2); }
	;

sync:	IL_SYNC IL_TOKEN		{ $$ = strtol($2, NULL, 0); free($2); }
	;

action:	IL_LOG '{' logrule '}'
	;

logrule:
	actsrc ';' actdest ';' 			{ link_sfd($1, NULL, $3); }
	| actsrc ';' dofilter actdest ';'	{ link_sfd($1, $3, $4); }
	;

actsrc:	IL_SOURCE IL_SRCNAME			{ $$ = get_var($2); free($2); }
	| srcobj				{ $$ = $1; }
	;

actdest:
	IL_DESTINATION IL_DESTNAME 		{ $$ = get_var($2); free($2); }
	| destobj 				{ $$ = $1; }
	;

dofilter:
	actfil ';'				{ $$ = add_filter($1, NULL); }
	| dofilter actfil ';'			{ $$ = add_filter($1, $2); }
	;

actfil:	IL_FILTER filnamelist			{ $$ = $2; }
	;

filnamelist:
	filnamelist ',' filname		{ $$ = link_filters($3, $1);
					  free($3); }
	| filname			{ $$ = link_filters($1, NULL);
					  free($1); }
	;

filname:
	IL_FILNAME				{ $$ = $1; }
	| not IL_FILNAME			{ $$ = $2; }
	| filobj				{ $$ = strdup($1->v_name); }
	;
%%
struct in_addr
getipv4addr(arg)
	char *arg;
{
	struct hostent *hp;
	struct in_addr in;

	in.s_addr = 0xffffffff;

	if ((hp = gethostbyname(arg)))
		bcopy(hp->h_addr, &in.s_addr, sizeof(struct in_addr));
	else
		in.s_addr = inet_addr(arg);
	return in;
}


u_short
getportnum(pr, name)
	char *pr, *name;
{
	struct servent *sp;

	if (!(sp = getservbyname(name, pr)))
		return htons(atoi(name));
	return sp->s_port;
}


struct in_addr
getipmask(arg)
	char *arg;
{
	struct in_addr in;
	int val;

	if (!strchr(arg, '.')) {
		val = atoi(arg);

		if (isdigit(*arg) && val >=0 && val <= 32) {
			in.s_addr = 0;
			for (; val; val--) {
				in.s_addr >>= 1;
				in.s_addr |= 0x8000000;
			}
			in.s_addr = htonl(in.s_addr);
		} else
			in.s_addr = htonl((u_long)strtol(arg, NULL, 0));
	} else
		in.s_addr = inet_addr(arg);
	return in;
}


void
yyerror(msg)
	char	*msg;
{
	if (opts & (OPT_DEBUG|OPT_VERBOSE))
		fprintf(stderr, "%s error at \"%s\", line %d\n", msg, yytext,
			lineNum + 1);
}
